/**
 * @file person.c
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on July 16, 2015, 18:34 PM
 * @brief  Simulate activiti of the person.
 *
 *
 * @section LICENSE
 *
 * SmartSlog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SmartSlog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * <b>Simulate activity of the person.</b> <br />
 * Person is a part of the GPS example: one KP publishes inforamtion about a person and
 * a gps-data. Another gets persons using subscription to classes. <br />
 * This part inserts and removes the Person object while you are not pressed
 * CTRL+C.
 *
 * This file is part of the SmartSlog project.
 *
 * Copyright (C) 2015 by PetrSU SmartSlog team.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#include "gps_foaf.h"

// Default parameters of the KP
#define KP_SS_NAME "X"
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"     // Local SIB
#define KP_SS_PORT 10010

#define KP_NAME "PersonKP"

// Using to generate and manipulate with a random number.
#define KP_MAX_SLEEPING_TIME 15 // Seconds
#define KP_RND_MAX_VALUE 10000
#define KP_RND_MAX_VALUE_LEN 4
#define KP_RND_MAX_STRING_LEN 255

// Name and surname of the person.
#define KP_PERSON_URI_DEFAULT "person1"
#define KP_PERSON_NAME_PREFIX "Person_"
#define KP_PERSON_SURNAME_PREFIX "Hrushkin_"

// Maximum length for strings.
#define KP_SS_ADDRESS_MAX_LEN 17
#define KP_SS_NAME_MAX_LEN 255
#define KP_PERSON_URI_MAX_LEN 255
#define KP_NAME_MAX_LEN 255

// Using to manipulate with parameters of the propgram.
#define KP_PARAM_MAX_LEN 10
#define KP_PARAM_KP_NAME "--kpname"
#define KP_PARAM_SS_NAME "--name"
#define KP_PARAM_SS_ADDRESS "--addr"
#define KP_PARAM_SS_PORT "--port"
#define KP_PARAM_HELP "--help"

/**
 * @brief Contains data of the KP.
 */
typedef struct kp_data_s {
    int ss_port;
    char ss_name[KP_SS_NAME_MAX_LEN];
    char kp_name[KP_NAME_MAX_LEN];
    char ss_address[KP_SS_ADDRESS_MAX_LEN];
    char person_uri[KP_PERSON_URI_MAX_LEN];
} kp_data_t;

/**
 *  @brief Flag to stop the activity (publish and leave)
 *
 * This flag is used when CTR+C is pressed.
 */
static bool is_activity_stopped = false;



/******************************************************************************/
/*************************** Static functions list ****************************/
static void init_kp_data(int argc, char **argv, kp_data_t *data);
static void init_random();

static void simulate_activity(sslog_individual_t *person, kp_data_t *data);
static bool join_and_publish(sslog_node_t *node, sslog_individual_t *person, kp_data_t *data);
static bool clean_and_leave(sslog_node_t *node, sslog_individual_t *person);
static void populte_locally(sslog_individual_t * person);
static void populate_names(sslog_individual_t * person);
static void populate_coordinates(sslog_individual_t * person);

static void signal_callback_handler(int signum);

static void print_kp_data(kp_data_t *data);
static void print_help();

static int next_random(int max_value);
static char *get_random_string(const char *prefix);
/******************************************************************************/

/**
 * @brief Starts the KP.
 *
 * Parameters: <br />
 *  --name: name of the smart space; <br />
 *  --addr: IP address of the SIB; <br />
 *  --port: port of the SIB. <br />
 *  --kpname: name of the KP.<br />
 *
 * @param[in] argc number of parameters.
 * @param[in] argv parameter's array.
 *
 * @return 0.
 */
int main(int argc, char **argv)
{

    // Print help if it is needed.
    if (argc == 2  && strncmp(KP_PARAM_HELP, argv[1], KP_PARAM_MAX_LEN) == 0) {
        print_help();
        return 0;
    }

    // Sets handler for CTR+C
    signal(SIGINT, &signal_callback_handler);

    kp_data_t kp_data = {0, "\0", "\0", "\0", "\0"};

    sslog_init();

    register_ontology();

    init_kp_data(argc, argv, &kp_data);

    print_kp_data(&kp_data);

    init_random();

    sslog_individual_t *person = sslog_new_individual(CLASS_PERSON, kp_data.person_uri);
    sslog_individual_t *point = sslog_new_individual(CLASS_POINT, NULL);

    sslog_insert_property(person, PROPERTY_LOCATION, point);

    simulate_activity(person, &kp_data);

    sslog_shutdown();

    return 0;
}



/******************************************************************************/
/***************************** Static functions *******************************/
/**
 * @brief Initializes KP data using parameters of default values.
 * 
 * @param[in] argc Number of parameters
 * @param[in] argv parameter's array.
 * @param[in] data kp data to initialize.
 */
static void init_kp_data(int argc, char **argv, kp_data_t *data)
{
    if (data == NULL) {
        return;
    }

    // Set default values:
    data->ss_port = KP_SS_PORT;
    strncpy(data->ss_name, KP_SS_NAME, KP_SS_NAME_MAX_LEN);
    strncpy(data->ss_address, KP_SS_ADDRESS, KP_SS_ADDRESS_MAX_LEN);
    strncpy(data->kp_name, KP_NAME, KP_NAME_MAX_LEN);

    char *uri = sslog_generate_uri(CLASS_PERSON);
    if (uri != NULL) {
        strncpy(data->person_uri, uri, KP_PERSON_URI_MAX_LEN);
        free(uri);
    }
    else {
        strncpy(data->person_uri, KP_PERSON_URI_DEFAULT, KP_PERSON_URI_MAX_LEN);
    }

    if (argc < 2 || argv == NULL) {
        return;
    }

    // Start checking parameters after program name.
    for (int i = 1; i + 1 < argc; i += 2) {
        if (argv[i] == NULL) {
            continue;
        }

        if (strncmp(KP_PARAM_KP_NAME, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->kp_name, argv[i + 1], KP_NAME_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_NAME, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->ss_name, argv[i + 1], KP_SS_NAME_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_ADDRESS, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->ss_address, argv[i + 1], KP_SS_ADDRESS_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_PORT, argv[i], KP_PARAM_MAX_LEN) == 0) {
            data->ss_port = (int) strtol(argv[i + 1], NULL, 10);
        }
    }
}


/** @brief Initializes random generator. */
static void init_random()
{
    time_t t;
    srand((unsigned) time(&t));
}


/**
 * @brief Simulate activity of the person.
 *
 * @param person individual with who manipulate.
 * @param data information about the KP and the smart space.
 */
static void simulate_activity(sslog_individual_t *person, kp_data_t *data)
{
    if (person == NULL) {
        return;
    }

    sslog_node_t *node = sslog_new_node(data->kp_name, data->ss_name, data->ss_address, data->ss_port);

    while (is_activity_stopped == false) {

        bool result = join_and_publish(node, person, data);

        if (result == true) {
            int rnd = next_random(KP_MAX_SLEEPING_TIME);
            sleep((unsigned int) rnd);

            clean_and_leave(node, person);
        } else {
            printf("\nCan't publish person's data. Waiting...");
        }

        int rnd = next_random(KP_MAX_SLEEPING_TIME);
        sleep((unsigned int) rnd);
    }
}


/**
 * @brief Joins to the smart space and publishes data.
 *
 * @param person individual to insert.
 * @param data information about the KP and the smart space.
 *
 * @return true on success, or false if it can't joins.
 */
static bool join_and_publish(sslog_node_t *node, sslog_individual_t *person, kp_data_t *data)
{
    // In the smart space the name of the KP must be unique, if you try to
    // join without leave, then you received an error.
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        return false;
    }

    // Evry time we populate the person to change GPS-coordinates.
    // On the first time a name and a surname will be set.
    populte_locally(person);

    // Insert two individuals: point with gps-coordinates and person
    sslog_individual_t *location_ind = (sslog_individual_t *) sslog_get_property(person, PROPERTY_LOCATION);

    if (location_ind != NULL) {
        sslog_node_insert_individual(node, location_ind);
    }

    sslog_node_insert_individual(node, person);

    printf("\n\nPerson with URI = '%s' is added to the smart space. ", sslog_entity_get_uri(person));

    return true;
}

/**
 * @brief Cleans the smart space and leaves it.
 *
 * @param person individual to remove.
 * @return true on success ot false otherwise.
 */
static bool clean_and_leave(sslog_node_t *node, sslog_individual_t *person)
{
    if (person == NULL) {
        return false;
    }

    // Remove person and coordinates from the smart space.
    sslog_individual_t *location_ind = (sslog_individual_t *) sslog_get_property(person, PROPERTY_LOCATION);

    if (location_ind != NULL) {
        sslog_node_remove_individual(node, location_ind);
    }

    sslog_node_remove_individual(node, person);

    sslog_insert_property(person, PROPERTY_LOCATION, location_ind);


    printf("\n\nPerson is removed from the smart space...");

    sleep(1);

    sslog_node_leave(node);

    return true;
}

/**
 * @brief Populates a persons.
 *
 * It sets names and GPS-coordinates.
 *
 * @param person individual to populate.
 */
static void populte_locally(sslog_individual_t * person)
{
      populate_names(person);

    populate_coordinates(person);
}

/**
 * @brief Creates names for a person and sets properties.
 *
 * It creates two random names (first name and surname) and set them to
 * given individual.
 *
 * @param person individual to set names.
 */
static void populate_names(sslog_individual_t * person)
{
    if (person == NULL) {
        return;
    }

    // In the first time set name and surname to the person.
    sslog_individual_t *location_ind = (sslog_individual_t *) sslog_get_property(person, PROPERTY_FIRSTNAME);

    if (location_ind == NULL) {
        char *value = get_random_string(KP_PERSON_NAME_PREFIX);
        sslog_insert_property(person, PROPERTY_FIRSTNAME, value);
        free(value);

        value = get_random_string(KP_PERSON_SURNAME_PREFIX);
        sslog_insert_property(person, PROPERTY_FIRSTNAME, value);
        free(value);
    }
}

/**
 * @brief Creates GPS-coordinates for a point and sets properties.
 *
 * Gets location property from the given person and sets random GPS-coordinates
 * for the location (Point).
 *
 * @param person individial to set location.
 */
static void populate_coordinates(sslog_individual_t * person)
{
    if (person == NULL) {
        return;
    }

    // Coordinates are stored as with object property ('Location')
    // They are represented with 'Point' class.
    sslog_individual_t *point = (sslog_individual_t *) sslog_get_property(person, PROPERTY_LOCATION);

    if (point == NULL) {

    }

    // Update all properties that are stored GPS coordinates with random values.
    char *value = get_random_string(NULL);
    //sslog_update_property(person, PROPERTY_ALT->name, value);
    sslog_update_property(point, PROPERTY_ALT, NULL, value);
    //sslog_remove_properties(point, PROPERTY_ALT);
    //sslog_insert_property(point, PROPERTY_ALT, value);
    free(value);

    value = get_random_string(NULL);
    //sslog_update_property(person, PROPERTY_LAT->name, value);
    sslog_remove_properties(point, PROPERTY_LAT);
    sslog_insert_property(point, PROPERTY_LAT, value);
    free(value);

    value = get_random_string(NULL);
    //sslog_update_property(person, PROPERTY_LONG->name, value);
    sslog_remove_properties(point, PROPERTY_LONG);
    sslog_insert_property(point, PROPERTY_LONG, value);
    free(value);

    printf("Coordinates are set.\n");
}

/**
 * Function to handle ctrl+c (SIGINT) signal.
 *
 * @param[in] signum signal identifier number.
 */
void signal_callback_handler(int signum)
{
    if (is_activity_stopped == false) {
        printf("Please waiting, i'm going to shutdown the program...");
        is_activity_stopped = true;
    } else {
        printf("OK, am exiting...");
        exit(signum);
    }
}

/** @brief Prints information about KP. */
static void print_kp_data(kp_data_t *data)
{

    if (data == NULL) {
        return;
    }

    printf("\n\nKP data:\n\tKP NAME: %s\n\tSS NAME: %s\n\tSS ADDR: %s" \
        "\n\tSS PORT: %i\n\tPerson's URI: %s\n\n",
            data->kp_name, data->ss_name, data->ss_address, data->ss_port,
            data->person_uri);

}

/** @brief Prints information about parameters and how to use program. */
static void print_help()
{
    printf("\nUsage: person [OPTION...]\n\n" \
            " --name     name of the smart space;\n"
            " --addr     IP address of the SIB;\n" \
            " --port     port of the SIB;\n" \
            " --kpname   name of the KP;\n" \
            " --help     print this.\n\n");
}


/**
 * @brief Returns nex random number.
 *
 * @param max_value maximum value of the random number.
 *
 * @return new random number.
 */
static int next_random(int max_value)
{
    return rand() % max_value;
}

/**
 * @brief Returns string with random number.
 *
 * It generates a random number and translate it to string.
 * If there is a prefix then it adds the prefix to the result string before
 * the random number.
 *
 * @param prefix prefix of the random number.
 *
 * @return string with prefix (if given) and ramndom number.
 */
static char *get_random_string(const char *prefix)
{
    size_t len = KP_RND_MAX_VALUE_LEN + 1;

    if (prefix != NULL) {
        len += strlen(prefix);
    }

    char *string = (char *) malloc(sizeof (char) * len);

    if (string == NULL) {
        return NULL;
    }

    int rnd = next_random(KP_RND_MAX_VALUE);

    if (prefix == NULL) {
        sprintf(string, "%i", rnd);
    } else {
        sprintf(string, "%s%i", prefix, rnd);
    }

    return string;
}


