/**
 * @file registrator.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on May 9, 2012, 18:34 PM
 * @brief  Gets infromation about individuals using subscription.
 *
 *
 * @section LICENSE
 *
 * SmartSlog KP Library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SmartSlog KP Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * registrator.c -
 * <b>Gets infromation about individuals using subscription.</b> <br />
 * A part of the GPS example: one KP publishes inforamtion about a person and
 * a gps-data. Another gets persons using subscription to classes. <br />
 * This part uses subscription to classes and gets information about
 * individuals that were inserted or removed individuals. <br />
 *
 * <b>Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.</b> <br />
 * Mail-list: smartslog@cs.karelia.ru
 */



#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include "gps_foaf.h"


// Default parameters of the KP
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"     // Local SIB
#define KP_SS_PORT 10010

#define KP_NAME "PersonRegistratorKP"

// Maximum length for strings.
#define KP_SS_ADDRESS_MAX_LEN 17
#define KP_SS_NAME_MAX_LEN 255
#define KP_PERSON_UUID_MAX_LEN 255
#define KP_NAME_MAX_LEN 255

// Using to manipulate with parameters of the propgram.
#define KP_PARAM_MAX_LEN 10
#define KP_PARAM_KP_NAME "--kpname"
#define KP_PARAM_SS_NAME "--name"
#define KP_PARAM_SS_ADDRESS "--addr"
#define KP_PARAM_SS_PORT "--port"
#define KP_PARAM_HELP "--help"

#define KP_PERSON_UUID_DEFAULT "person1"

/**
 * @brief Contains data of the KP.
 */
typedef struct kp_data_s {
    int ss_port;
    char ss_name[KP_SS_NAME_MAX_LEN];
    char kp_name[KP_NAME_MAX_LEN];
    char ss_address[KP_SS_ADDRESS_MAX_LEN];
    char person_uuid[KP_PERSON_UUID_MAX_LEN];
} kp_data_t;

/**
 *  @brief Flag to stop the activity.
 *
 * This flag is used when CTR+C is pressed.
 */
static bool is_activity_stopped = false;


/******************************************************************************/
/*************************** Static functions list ****************************/
static void init_kp_data(int argc, char **argv, kp_data_t *data);

static void process_subscription(sslog_node_t *node, sslog_subscription_t *subscription);
static void signal_callback_handler(int signum);

static void print_persons(sslog_node_t *node, const list_t *persons);
static void print_help();

/******************************************************************************/

/**
 * @brief Starts the KP.
 *
 * Parameters: <br />
 *  --name: name of the smart space; <br />
 *  --addr: IP address of the SIB; <br />
 *  --port: port of the SIB. <br />
 *  --kpname: name of the KP.<br />
 *
 * @param[in] argc number of parameters.
 * @param[in] argv parameter's array.
 *
 * @return 0.
 */
int main(int argc, char **argv)
{
    // Print help if it is needed.
    if (argc == 2 && strncmp(KP_PARAM_HELP, argv[1], KP_PARAM_MAX_LEN) == 0) {
        print_help();
        return 0;
    }

    // Sets handler for CTR+C
    signal(SIGINT, &signal_callback_handler);

    kp_data_t kp_data = {0, "\0", "\0", "\0", "\0"};

    sslog_init();

    register_ontology();

    init_kp_data(argc, argv, &kp_data);

    sslog_node_t *node = sslog_new_node(kp_data.kp_name, kp_data.ss_name, kp_data.ss_address, kp_data.ss_port);


    // In the smart space the name of the KP must be unique, if you try to
    // join without leave, then you received an error.
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        return 0;
    }

    // Create a new subscription and add class to it.
    sslog_subscription_t *subscription = sslog_new_subscription(node, false);

    sslog_sbcr_add_class(subscription, CLASS_PERSON);

    // Try to subscribe, after this we will subscribe to class
    // and will get notifications about inserting or removing
    // individuals of the subscribed class in the smart space.
    if (sslog_sbcr_subscribe(subscription) != SSLOG_ERROR_NO) {
        printf("\nCan't subscribe.");
        sslog_shutdown();
        return 0;
    }

    // Checking information about subscription.
    process_subscription(node, subscription);

    // This function is used to show how to usubscribe all subscription for node.
    // In this demo it is not so useful and sslog_free_subscription can be used.
    sslog_sbcr_unsubscribe_all(node, true);

    sslog_node_leave(node);

    sslog_shutdown();

    return 0;
}



/******************************************************************************/
/***************************** Static functions *******************************/

/**
 * @brief Initializes KP data using parameters of default values.
 *
 * @param[in] argc Number of parameters
 * @param[in] argv parameter's array.
 * @param[in] data kp data to initialize.
 */
static void init_kp_data(int argc, char **argv, kp_data_t *data)
{

    if (data == NULL) {
        return;
    }

    // Set default values:
    data->ss_port = KP_SS_PORT;
    strncpy(data->ss_name, KP_SS_NAME, KP_SS_NAME_MAX_LEN);
    strncpy(data->ss_address, KP_SS_ADDRESS, KP_SS_ADDRESS_MAX_LEN);
    strncpy(data->kp_name, KP_NAME, KP_NAME_MAX_LEN);

    char *uri = sslog_generate_uri(CLASS_PERSON);
     if (uri != NULL) {
        strncpy(data->person_uuid, uri, KP_PERSON_UUID_MAX_LEN);
        free(uri);
    }
    else {
        strncpy(data->person_uuid, KP_PERSON_UUID_DEFAULT, KP_PERSON_UUID_MAX_LEN);
    }

    if (argc < 2 || argv == NULL) {
        return;
    }

    // Start checking parameters after program name.
    for (int i = 1; i + 1 < argc; i += 2) {
        if (argv[i] == NULL) {
            continue;
        }

        if (strncmp(KP_PARAM_KP_NAME, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->kp_name, argv[i + 1], KP_NAME_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_NAME, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->ss_name, argv[i + 1], KP_SS_NAME_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_ADDRESS, argv[i], KP_PARAM_MAX_LEN) == 0) {
            strncpy(data->ss_address, argv[i + 1], KP_SS_ADDRESS_MAX_LEN);
        } else if (strncmp(KP_PARAM_SS_PORT, argv[i], KP_PARAM_MAX_LEN) == 0) {
            data->ss_port = (int) strtol(argv[i + 1], NULL, 10);
        }
    }
}

/**
 * @brief Waits subscription and print information about last changes.
 *
 * @param[in] subscription subscription to check.
 */
static void process_subscription(sslog_node_t *node, sslog_subscription_t *subscription)
{
    if (subscription == NULL) {
        return;
    }

    while (sslog_sbcr_is_active(subscription) == true
            && is_activity_stopped == false) {

        if (sslog_sbcr_wait(subscription) != SSLOG_ERROR_NO) {
            continue;
        }

        sslog_sbcr_changes_t *changes =
                sslog_sbcr_get_changes_last(subscription);

        sslog_sbcr_ch_print(changes);

        const list_t *inserted_ind =
                sslog_sbcr_ch_get_individual_by_action(changes, SSLOG_ACTION_INSERT);

        const list_t *removed_ind =
                sslog_sbcr_ch_get_individual_by_action(changes, SSLOG_ACTION_REMOVE);

        printf("\n---- Inserted individuals:\n");
        print_persons(node, inserted_ind);

        printf("\n---- Removed individuals:\n");
        print_persons(node, removed_ind);
    }
}

/**
 * @brief Print information about persons and their locations.
 *
 * @param[in] persons list with persons.
 */
static void print_persons(sslog_node_t *node, const list_t *persons)
{
    if (persons == NULL) {
        return;
    }

    list_head_t *list_walker = NULL;

    list_for_each(list_walker, &persons->links) {
        list_t *list_node = list_entry(list_walker, list_t, links);
        char *uri = (char *) list_node->data;

        // After subscription new individuals will be created, and here
        // get individual from the repository.
        sslog_individual_t *ind = sslog_get_individual(uri);

        if (ind == NULL) {
            printf ("No individual in the local store.\n");
            continue;
        }

        printf("Indiviual URI: %s\n", sslog_entity_get_uri(ind));

        sslog_individual_t *point = (sslog_individual_t *) sslog_node_get_property(node, ind, PROPERTY_LOCATION);

        if (point == NULL) {
            printf ("Not location property.\n");
            continue;
        }

        sslog_node_populate(node, point);

        char *lat = (char *) sslog_get_property(point, PROPERTY_LAT);
        char *longitude = (char *) sslog_get_property(point, PROPERTY_LONG);
        char *alt = (char *) sslog_get_property(point, PROPERTY_ALT);

        if (point == NULL) {
            printf(" - No information about location.\n\n");
        } else {
            printf(" - Location infomation: lat: %s; long: %s; alt: %s;\n\n", lat, longitude, alt);
        }

        sslog_remove_individual(point);
    }
}

/**
 * Function to handle ctrl+c (SIGINT) signal.
 *
 * @param[in] signum signal identifier number.
 */
void signal_callback_handler(int signum)
{
    if (is_activity_stopped == false) {
        printf("Please waiting, i'm going to shutdown the program...");
        sslog_sbcr_stop_all();
        printf("\n\nERROR %s", sslog_error_get_last_text());
        is_activity_stopped = true;
    } else {
        printf("OK, am exiting...");
        exit(signum);
    }
}

/** @brief Prints information about parameters and how to use program. */
static void print_help()
{
    printf("\nUsage: person [OPTION...]\n\n" \
            " --name     name of the smart space;\n"
            " --addr     IP address of the SIB;\n" \
            " --port     port of the SIB;\n" \
            " --kpname   name of the KP;\n" \
            " --help     print this.\n\n");
}
