/**
 * @file consumer_kp.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 3, 2012
 * @brief  Cunsumer of the example 'Hello World' without subscription.
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello World' example without subscription.
 * This KP makes aqueries to triples using <KP - says - *> triple-pattern.
 * It gets triples from the smart space and writes them. If the one of triples 
 * has an 'exit'  object, thenthis KP leaves the smart space and ends.
 * To stop the KP use CTRL+C 
 *
 * Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */

#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <smartslog.h>

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"   // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "HelloConsumer"

// Parts of the triple for the query.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"
#define KP_EXIT_WORD_LEN 4

// Time to sleep between cheking(query) a triple.
#define KP_WAITING_TIME 4   // For sleep() function.


/****************** Static functions list **********************/
static void checking_triple(sslog_node_t *node);
/***************************************************************/
        

/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    sslog_init();

    // Crteate a node to connect to the smart space.
    sslog_node_t *node = sslog_new_node(KP_NAME, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);

   
    // In the smart space the name of the KP must be unique, if you try to 
    // join without leave, then you received an error.
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        sslog_shutdown();
        return 0;
    }

    printf("KP joins to SS.\n");
    
    // Start to chek triples: create a triple for the query 
    // and make query op[eration to the smart space.
    checking_triple(node);

    sslog_node_leave(node);

    printf("\nKP leave SS...\n");

    sslog_shutdown();

    return 0;
}

/**
 * @brief Checks triples in the SS.
 *
 * It queries triples and write information about them.
 * If it receives a KP_EXIT_WORD as value of the object in some triple, then 
 * it stops checking.
 *
 * @param[in] info information about the smart space..
 */
static void checking_triple(sslog_node_t *node)
{  
    if (node == NULL) {
        return;
    }

    // Create the triple for a query operation.
    sslog_triple_t *req_triple = sslog_new_triple_detached(KP_SUBJECT, KP_PREDICATE, SS_RDF_SIB_ANY,
                SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
        
    if (req_triple == NULL) {
        printf("Error: %s", sslog_error_get_last_text());
        return;
    }
    
    bool is_checking = true;
    
    while (is_checking == true) {
    
        // Remove local triples and get new from smart space.
        sslog_remove_triple(req_triple);
        list_t *result_triples = sslog_node_query_triple(node, req_triple);

        if (sslog_error_get_last_code() != SSLOG_ERROR_NO) {
           printf("Query operation fails: %s", sslog_error_get_last_text());
           break;            
        }
    
        // Print information using triples that were receievng by the query.
        list_head_t *list_walker = NULL;
        list_for_each(list_walker, &result_triples->links) {
            list_t *node = list_entry(list_walker, list_t, links);
            sslog_triple_t *result_triple = (sslog_triple_t *) node->data;

            printf("KP says: '%s'.\n", result_triple->object);

            // If there is the KP_EXIT_WORD as object, then stop checking.
            if (strncmp(KP_EXIT_WORD, result_triple->object, KP_EXIT_WORD_LEN) == 0) {
                printf("It's time to leave...\n");
                is_checking = false;
                break;
            }

        }

        list_free_with_nodes(result_triples, NULL);

        // Sleep for a some time between cheking.
        sleep(KP_WAITING_TIME);       
    }
    
    sslog_free_triple(req_triple);
}
