/**
 * @file publisher_kp.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 3, 2012
 * @brief  Publisher of the example 'Hello World' without subscription.
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello World' example without subscription.
 * This KP updates a triple <KP - says - text>, where 'text' is a input from 
 * the user. 
 *
 * Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <smartslog.h>

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"  // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "HelloPublisher"

// Parts of the triple for the inserting.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT  KP_NS_URI"KP"
#define KP_PREDICATE  KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"
#define KP_EXIT_WORD_LEN 4

// After KP has publushed 'exit' keyword, it waits for a some time, 
// before cleaning the smart space. This time gives a chance to 
// KP-consumers, that a work without subscription receiving 'exit' keyword and 
// ends their work.
#define KP_WAITING_TIME_CLEANING 5   // For sleep() function


/****************** Static functions list **********************/
static void process_publish(sslog_node_t *node);
static void remove_newline(char *str);         
/*******************************************************/


/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    // Initialize the smart slog and node.
    sslog_init();

    // Crteate a node to connect to the smart space.
    sslog_node_t *node = sslog_new_node(KP_NAME, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);

    // In the smart space the name of the KP must be unique, if you try to 
    // join without leave, then you received an error.   
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        sslog_shutdown();
        return 0;
    }

    printf("KP join to SS.\n");
    
    // Prepare triples to work, get a text string from the user
    // and update information in the smart space.
    process_publish(node);

    sslog_node_leave(node);

    printf("\nKP leaves SS...\n");

    sslog_shutdown();

    return 0;
}

/**
 * @brief Publish information to the smart space.
 *
 * Creates triples for removing and adding. 
 * The object of the inserting triple is reading from the sdin.
 * If the user input 'exit' keyword then it publishes it, waits for a some time, 
 * cleans the smart space and ends working.
 *
 * @param[in] info information about the smart space..
 */
static void process_publish(sslog_node_t *node)
{  
    if (node == NULL) {
        return;
    }
    
    // Create triple for removing data, this triple is not changing while 
    // publishing process working.    
    sslog_triple_t *rem_triple = sslog_new_triple_detached(KP_SUBJECT, KP_PREDICATE, SS_RDF_SIB_ANY,
                SS_RDF_TYPE_URI, SS_RDF_TYPE_URI);
        
    if (rem_triple == NULL) {
        printf("Error: %s\n", sslog_error_get_last_text());
        return;
    }

    // Read a string from the stdin and create a triple using this line as 
    // a value for the object. 
    // Update triple in the SS. 
    // If object equals KP_EXIT_WORD constant, then it waiting for a some time
    // and cleaning the SS.
    while (true) {
        char object_text[SS_OBJECT_MAX_LEN];

        printf("\nSay something or write '%s':\n", KP_EXIT_WORD);

        if (fgets(object_text, SS_OBJECT_MAX_LEN, stdin) == NULL) {
            printf("Something wrong with the input!\n");
            break;
        }    
        
        // Remove the '\n' character.
        remove_newline(object_text);
          
        // Create triple for inserting.
        sslog_triple_t *ins_triple = sslog_new_triple_detached(KP_SUBJECT, KP_PREDICATE, object_text,
                SS_RDF_TYPE_URI, SS_RDF_TYPE_LIT);
 
        if (ins_triple == NULL) {
            printf("Error: %s.\n", sslog_error_get_last_text());
            break;
        }
                
        printf("Publishing: %s\n", ins_triple->object);

        // Removes old value and insert a new.
        int result = sslog_node_update_triple(node, rem_triple, ins_triple);

        sslog_free_triple(ins_triple);
               
        if (result != 0) {
            printf("Can't update triples: %s.\n", sslog_error_get_last_text());
            break;
        } 
        
        // If the user typed KP_EXIT_WORD, then gives consumer KPs time to 
        // receiveng data and cleaning the SS.
        if (strncmp(KP_EXIT_WORD, object_text, KP_EXIT_WORD_LEN) == 0) {
            printf("Waiting before cleaning...\n");
            sleep(KP_WAITING_TIME_CLEANING);
            
            printf("Cleaning...\n");
            sslog_node_remove_triple(node, rem_triple);
            break;
        }               
    }
    
    sslog_free_triple(rem_triple);
}

/**
 *@brief Finds '\n' character in the end of the line and chenges it to '\0';
 * 
 * @param[in] str string to remove endline character.
 */
static void remove_newline(char *str) 
{
    if (str == NULL) {
        return;
    }
    
    size_t len = strlen(str);
    
    if (str[len - 1] == '\n') {
        str[len - 1] = '\0';
    }
}

