/**
 * @file consumer_kp.c
 * @Author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 19, 2012
 * @brief  Cunsumer of the example 'Hello World' with subscription.
 *
 * This is file is a part of the C KPI library project and it has the same licence. 
 *
 * @section LICENSE
 * 
 * C KPI library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * C KPI Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 *
 * @section DESCRIPTION
 *
 * This file is a part of the 'Hello World' example with subscription.
 * This KP make a subscription to triples using <KP - says - *> triple-pattern.
 * It gets information about chnages and output old and new values. 
 * To stop the KP use CTRL+C 
 *
 * Copyright (C) SmartSlog Team (Aleksandr A. Lomov).
 * All rights reserved.
 * Mail-list: smartslog@cs.karelia.ru
 */


#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "smartslog.h"

// ---Constants---
// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"  // Local SIB
#define KP_SS_PORT 10010
//#define KP_SS_PORT 10011

// Name of the KP.
#define KP_NAME "HelloConsumerSparqlSub"

// Parts of the triple for the subscription.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

// The word(object of the triple) for the exiting.
#define KP_EXIT_WORD "exit"
#define KP_EXIT_WORD_LEN 4


/**
 *  @brief Flag to stop the program activity.
 *
 * This flag is used when CTR+C is pressed 
 * or when subscription indication is a error.
 */
static sig_atomic_t exit_flag = 0;


/****************** Static functions list **********************/
static void signal_callback_handler(int signum);

static sslog_subscription_t *make_subscription(sslog_node_t *node);
static void handle_subscription(sslog_subscription_t *subscription);

static void print_sparql_result(sslog_sparql_result_t *result);
/*******************************************************/        


/**
 * @brief Starts the KP, initializes KP state and calls all necessary functions.
 *
 * @return 0.
 */
int main()
{
    // Sets handler for CTR+C
    signal(SIGINT, &signal_callback_handler);

    sslog_init();

    sslog_node_t *node = sslog_new_node(KP_NAME, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);
    
    // In the smart space the name of the KP must be unique, if you try to
    // join without leave, then you received an error.
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        sslog_shutdown();
        return 0;
    }

    printf("KP join to SS.\n");
    
    // Create a subscription to <KP-says-*> triples.
    sslog_subscription_t *subscription = make_subscription(node);

    if (subscription == NULL) {
        printf("Can't subscribe.\n");

        sslog_node_leave(node);
        sslog_shutdown();

        return 0;
    }
    
    printf("\nWaiting new data.\n");
    
    // Start subscription processing, it is ends:
    // - when CTRL+C is pressed;
    // - when unsubscription indication is received from the SS;
    // - on error.
    handle_subscription(subscription);
    
    // Before leave we need to unsubscribe our subscription.
    sslog_sbcr_unsubscribe(subscription);

   // free(subscription);
    sslog_free_subscription(subscription);
        
    sslog_node_leave(node);

    printf("\nKP leave SS...\n");

    sslog_shutdown();

    return 0;
}



/******************* Static functions ************************/

/**
 * @brief Function to handle ctrl+c (SIGINT) signal. If you press 
 * CTRL+C twice, then program will be closed immediately.
 *
 * @param[in] signum signal identifier number.
 */
static void signal_callback_handler(int signum)
{
    if (exit_flag == 0) {
        printf("Please waiting, i'm going to shutdown the program...");
        sslog_sbcr_stop_all();
        printf("\n\nERROR %s", sslog_error_get_last_text());
        exit_flag = true;
    } else {
        printf("OK, am exiting...");
        exit(signum);
    }
}

/**
 * @brief Makes a subscription to triples.
 *
 * @param[in] info information about the smart space.
 * @param[in] subs subscription.
 *
 * @return 0 on success or -1 otherwise.
 */
static sslog_subscription_t* make_subscription(sslog_node_t *node)
{
    // SPARQL for a subscription operation.
    // Now BIND does't work with subscription.
//    const char query[] = "SELECT ?sub ?pred ?obj WHERE { <" KP_SUBJECT "> <" KP_PREDICATE "> ?obj "
//                                 "BIND (<" KP_SUBJECT "name" "> AS ?sub) "
//                                 "BIND (<" KP_PREDICATE "surname" "> AS ?pred)}";

    // Use other query.
    const char query[] = "SELECT ?obj ?sub ?pred  "
                         "WHERE { <" KP_SUBJECT "> <" KP_PREDICATE "> ?obj. "
                                 "<" KP_SUBJECT "> ?pred ?obj."
                                 "?sub <" KP_PREDICATE "> ?obj }";

    sslog_subscription_t *subscription = sslog_new_subscription(node, false);

    sslog_sbcr_add_sparql_select(subscription, query, "sub pred obj");
    //sslog_sbcr_add_sparql_select(subscription, query, NULL);

    if (sslog_sbcr_subscribe(subscription) != SSLOG_ERROR_NO) {
        printf("Failed to subscribe: %s", sslog_error_get_last_text());

        sslog_free_subscription(subscription);
        subscription = NULL;
    }

    sslog_sbcr_changes_t *changes = sslog_sbcr_get_changes_last(subscription);
    print_sparql_result(sslog_sbcr_ch_get_new_sparql_select(changes));
    
    return subscription;
 }

/**
 * @brief Checks notifications from the smart space. 
 * It works while exit_flag is 0 or while error/unsubscription indication is not 
 * received.   
 *
 * @param subscription. Subscription to handle.
 */
 static void handle_subscription(sslog_subscription_t *subscription)
{  
    sslog_triple_t *triple_template = sslog_new_triple_detached(KP_SUBJECT, KP_PREDICATE, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    while (exit_flag == 0) {

        int result = sslog_sbcr_wait(subscription);

        if (result != SSLOG_ERROR_NO) {
            printf("Error: %s.", sslog_error_get_last_text());
            break;
        }

        if (sslog_sbcr_is_active(subscription) == false) {
            printf("Subscription is inactive state: unsubscription message was received.");
            break;
        }

        sslog_sbcr_changes_t *changes = sslog_sbcr_get_changes_last(subscription);

        sslog_sparql_result_t *new_result = sslog_sbcr_ch_get_new_sparql_select(changes);
        sslog_sparql_result_t *old_result = sslog_sbcr_ch_get_old_sparql_select(changes);

        if (new_result != NULL) {
            printf("New values: \n");
            print_sparql_result(new_result);
        }

        if (old_result != NULL) {
            printf("Old values: \n");
            print_sparql_result(old_result);
        }

        list_t *stored_triples = sslog_query_triple(triple_template);

        if (list_is_null_or_empty(stored_triples) == 0) {
            sslog_triple_t *stored_triple = (sslog_triple_t *) list_get_first_prev_node(stored_triples)->data;

            // It's safe to use this function to stored triples.
            // Here it is also possible to free list with nodes: list_free_with_node(list, NULL);
            sslog_free_triples(stored_triples);
            stored_triples = NULL;

            printf("Triple: %s - %s - %s (o-type: %i)\n",
                   stored_triple->subject, stored_triple->predicate, stored_triple->object, stored_triple->object_type);

            if (strncmp(KP_EXIT_WORD, stored_triple->object, KP_EXIT_WORD_LEN) == 0) {
                break;
            }
        }

        sslog_free_triples(stored_triples);
    }

    sslog_free_triple(triple_template);
}

/**
 * @brief Prints SPARQL SELECT result.
 * @param result Result to print.
 */
static void print_sparql_result(sslog_sparql_result_t *result)
{
    if (result == NULL || result->rows_count == 0) {
        printf("No result to print.\n");
        return;
    }

    for (int i = 0; i < result->rows_count; ++i) {
        printf("Row N %i\n", i + 1);

        for (int j = 0; j < result->bindings_count; ++j) {
                printf("Name: %s; Value: %s; Type: %i;\n", result->names[j], result->rows[i]->values[j], result->rows[i]->types[j]);
        }
    }
}


