/**
 * @file sparql_kp.c
 * @author Aleksandr A. Lomov <lomov@cs.karelia.ru>
 * @date   Created on August 10, 2015
 * @brief  KP that allows to execute SPARQL queries.
 *
 * @section LICENSE
 *
 * SmartSlog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * SmartSlog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SmartSlog KP Library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 *
 * @section DESCRIPTION
 *
 * KP waits a query from user, executes it and prints results.
 *
 * This file is part of the SmartSlog project.
 *
 * Copyright (C) 2015 by PetrSU SmartSlog team.
 */

#include "smartslog.h"
#include <stdlib.h>


/******************************** Definitions ********************************/
/******************* Defines, structures, constatnts and etc.*****************/

// Describes connection to the SS.
#define KP_SS_NAME "X"
#define KP_SS_ADDRESS "194.85.173.9"  // Public PetrSU SIB
//#define KP_SS_ADDRESS "127.0.0.1"     // Local SIB
#define KP_SS_PORT 10010

// Name of the KP.
#define KP_NAME "SparqlExecutorKP"

// Parts of the triple for the query.
#define KP_NS_URI "http://X/Hello/#"
#define KP_SUBJECT KP_NS_URI"KP"
#define KP_PREDICATE KP_NS_URI"says"

enum QUERY_TYPE {
    SPARQL_CONSTRUCT = 1,
    SPARQL_ASK,
    SPARQL_SELECT
};

/*****************************************************************************/


/******************************* Definitions *********************************/
/**************************** Static functions *******************************/
/**
 * @brief Process SPARQL query.
 * @param[in] node Node to work with smart space.
 * @param[in] query_type Query form: CONSTRUCT, ASK, SELECT.
 * @param[in] query Query to process.
 */
static void process_query(sslog_node_t *node, int query_type, const char *query);

/**
 * @brief Print triples.
 * @param triples Triples to print.
 */
static void print_triples(list_t *triples);
/*****************************************************************************/

        

/**
 * @brief Starts the KP, initializes state and cals all necessary functions.
 *
 * @return 0.
 */
int main()
{
    sslog_init();

    sslog_node_t *node = sslog_new_node(KP_NAME, KP_SS_NAME, KP_SS_ADDRESS, KP_SS_PORT);

   // In the smart space the name of the KP must be unique, if you try to
   // join without leave, then you received an error.
    if (sslog_node_join(node) != SSLOG_ERROR_NO) {
        printf("Can't join to SS: %s\n", sslog_error_get_last_text());
        return 0;
    }

    printf("KP join to SS.\n");

    char input[2000] = { 0 };
    int choose;

    do {
        printf("Choose query type or press 4 to exit:\n1 - SPARQL CONSTRUCT\n2 - SPARQL ASK\n3 - SPARQL SELECT\n");
        if (fgets(input, 2000, stdin) == NULL) {
            printf("Error or eof.\n");
        } else {
            choose = atoi(input);
        }

        if (choose == 4) {
         break;
        }

        printf("Enter query:\n");
        if (fgets(input, 2000, stdin) == NULL) {
            printf("Error or eof.\n");
        } else {
            process_query(node, choose, input);
        }

    } while (choose != 4);


    sslog_node_leave(node);

    printf("KP leave SS...\n");

    sslog_shutdown();

    return 0;
}




static void process_query(sslog_node_t *node, int query_type, const char *query)
{
    switch (query_type) {

    case SPARQL_CONSTRUCT: {

        list_t *triples = sslog_node_sparql_construct(node, query);

        if (triples == NULL) {
            printf("Error: %s\n", sslog_error_get_last_text());
            break;
        }

        print_triples(triples);
        sslog_free_triples(triples);
        break;
    }

    case SPARQL_ASK: {

        bool result = sslog_node_sparql_ask(node, query);

        if (sslog_error_get_last_code() != SSLOG_ERROR_NO) {
            printf("Error: %s\n", sslog_error_get_last_text());
            break;
        } else {
            printf("Result: %s\n", (result == true) ? "true" : "false");
        }
        break;
    }

    case SPARQL_SELECT: {

        sslog_sparql_result_t *result = sslog_node_sparql_select(node, query, NULL);

        if (result == NULL) {
            printf("Error: %s\n", sslog_error_get_last_text());
            break;
        } else {

            for (int i = 0; i < result->rows_count; ++i) {
                for (int j = 0; j < result->bindings_count; ++j) {
                    printf("%i - name: %s type: %d value: %s\n", i, result->names[j], result->rows[i]->types[j], result->rows[i]->values[j]);
                }
            } // for - rows
        }

        sslog_free_sparql_result(result);

        break;
    }
    default:
        printf("Unknown query type: %i\n", query_type);
        break;
    }
}


static void print_triples(list_t *triples)
{
    if (list_is_null_or_empty(triples) == true) {
        printf("No triples to print.\n");
        return;
    }

    int counter = 1;

    list_head_t *list_walker = NULL;
    list_for_each(list_walker, &triples->links) {
        list_t *node = list_entry(list_walker, list_t, links);

        sslog_triple_t *triple = (sslog_triple_t *) node->data;

        printf("%i triple: '%s - %s - %s (o-type: %i)'\n", counter, triple->subject, triple->predicate, triple->object, triple->object_type);
        ++counter;
    }
}
