
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>

#include "include/Basic.h"

#include "kpi_api.h"
#include "session.h"

#include "utils/errors_internal.h"
#include "utils/errors_decl.h"
#include "utils/util_func.h"


#include "tests.h"

sslog_node_t *node;

/* The suite initialization function.
 * Opens the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int init_suite_nodes(void)
{

    int result = sslog_init();

    if (result != SSLOG_ERROR_NO) {
        return result;
    }

    node = sslog_new_node("testNodeId_kpibaseoperations",
                          SSLOG_TESTS_SS_ID, SSLOG_TESTS_SS_ADDRESS, SSLOG_TESTS_SS_PORT);

    if (node == NULL || sslog_error_get_last_code() != SSLOG_ERROR_NO) {
        return sslog_error_get_last_code();
    }

    return sslog_node_join(node);
}

/* The suite cleanup function.
 * Closes the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int clean_suite_nodes(void)
{
    sslog_free_node(node);
    sslog_shutdown();

    return 0;
}


/* Checks insert_triple function with NULL parameters. */
void test_null_triple_insert()
{
    sslog_triple_t *triple = sslog_new_triple(
                SSLOG_TEST_URI "subject", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    int result = sslog_node_insert_triple(NULL, NULL);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);

    result = sslog_node_insert_triple(node, NULL);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);

    result = sslog_node_insert_triple(NULL, triple);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);
}

/* Checks insert_triples function with NULL parameters. */
void test_null_triples_insert()
{
    sslog_triple_t *triple = sslog_new_triple(
                SSLOG_TEST_URI "subject", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *triples = list_new();
    list_add_data(triples, triple);

    int result = sslog_node_insert_triples(NULL, NULL);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);

    result = sslog_node_insert_triples(node, NULL);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);

    result = sslog_node_insert_triples(NULL, triples);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NULL_ARGUMENT);

    list_free_with_nodes(triples, NULL);
}


/* Checks insert/query/remove_triple functions.
 * 1 - Insert one triples (checks operation results).
 * 2 - Query inserted triple .
 * 3 - Remove insered triple.
 * 4 - Query inserted triple (the triple must be removed from smart space).
 */
void test_uri_triple_insert_query_delete()
{
    sslog_triple_t *triple = sslog_new_triple(
                SSLOG_TEST_URI "subject", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    int result = sslog_node_insert_triple(node, triple);
    CU_ASSERT_EQUAL_FATAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);


    sslog_triple_t *query_triple = sslog_new_triple_detached(SSLOG_TEST_URI "subject", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object",
                                                    SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *triples = sslog_node_query_triple(node, query_triple);
    CU_ASSERT_EQUAL_FATAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1);

    sslog_free_triple(query_triple);

    list_t *triple_node = list_get_first_prev_node(triples);

    sslog_triple_t *inserted_triple = (sslog_triple_t*) triple_node->data;
    CU_ASSERT_TRUE(sslog_equal_triples(triple, inserted_triple));

    list_free_with_nodes(triples, SSLOG_FREE_FUNC sslog_free_triple);

    result = sslog_node_remove_triple(node, triple);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);

    triples = sslog_node_query_triple(node, triple);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(list_count(triples), 0);

    list_free_with_nodes(triples, NULL);
}


/* Checks insert/query/remove_triples functions.
 * 1 - Insert 3 triples (checks operation results).
 * 2 - Query inserted triple .
 * 3 - Remove insered triple.
 * 4 - Query inserted triple (the triple must be removed from smart space).
 */
void test_uri_triples_insert_query_delete()
{
       sslog_triple_t *triple1 = sslog_new_triple(
                SSLOG_TEST_URI "subject1", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object1",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_triple_t *triple2 = sslog_new_triple(
                SSLOG_TEST_URI "subject2", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object2",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_triple_t *triple3 = sslog_new_triple(
                SSLOG_TEST_URI "subject3", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object1",
                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *insert_triples = list_new();
    list_add_data(insert_triples, triple1);
    list_add_data(insert_triples, triple2);
    list_add_data(insert_triples, triple3);

    int result = sslog_node_insert_triples(node, insert_triples);

    list_free_with_nodes(insert_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    list_t *query_triples = list_new();
    sslog_triple_t *query_triple = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, triple1->predicate, triple1->object,
                                                             triple1->subject_type, triple1->object_type);
    list_add_data(query_triples, query_triple);

    query_triple = sslog_new_triple_detached(triple2->subject, triple2->predicate, triple2->object,
                                             triple2->subject_type, triple2->object_type);
    list_add_data(query_triples, query_triple);


    list_t *triples = sslog_node_query_triples(node, query_triples);

    list_free_with_nodes(query_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);
    list_free_with_nodes(triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);







//    sslog_triple_t *triple1 = sslog_new_triple(
//                SSLOG_TEST_URI "subject1", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object1",
//                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

//    sslog_triple_t *triple2 = sslog_new_triple(
//                SSLOG_TEST_URI "subject2", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object2",
//                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

//    sslog_triple_t *triple3 = sslog_new_triple(
//                SSLOG_TEST_URI "subject3", SSLOG_TEST_URI "predicate", SSLOG_TEST_URI "object1",
//                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

//    list_t *insert_triples = list_new();
//    list_add_data(insert_triples, triple1);
//    list_add_data(insert_triples, triple2);
//    list_add_data(insert_triples, triple3);

//    int result = sslog_node_insert_triples(node, insert_triples);
//    CU_ASSERT_EQUAL_FATAL(result, SSLOG_ERROR_NO);
//    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
//    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);

//    list_free_with_nodes(insert_triples, SSLOG_FREE_FUNC sslog_free_triple);

//    list_t *query_triples = list_new();
//    sslog_triple_t *query_triple = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, triple1->predicate, triple1->object,
//                                                             triple1->subject_type, triple1->object_type);
//    list_add_data(query_triples, query_triple);

//    /*query_triple = sslog_new_triple_detached(triple2->subject, triple2->predicate, triple2->object,
//                                             triple2->subject_type, triple2->object_type);
//    list_add_data(query_triples, query_triple);*/


//    list_t *triples = sslog_node_query_triples(node, query_triples);

//    list_free_with_nodes(query_triples, SSLOG_FREE_FUNC sslog_free_triple);
//    list_free_with_nodes(triples, SSLOG_FREE_FUNC sslog_free_triple);
        return;
    CU_ASSERT_EQUAL_FATAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3);

    list_del_and_free_nodes(query_triples, SSLOG_FREE_FUNC sslog_free_triple);

    list_head_t *list_walker = NULL;
    list_for_each(list_walker, &triples->links) {
        list_t *node = list_entry(list_walker, list_t, links);
        sslog_triple_t *triple = (sslog_triple_t *) node->data;

        if (sslog_equal_triples(triple, triple1) == true
                || sslog_equal_triples(triple, triple2) == true
                || sslog_equal_triples(triple, triple3) == true) {
            continue;
        }

        CU_FAIL("Query triples failed.");
    }

    list_free_with_nodes(triples, NULL);

    query_triple = sslog_new_triple_detached(triple1->subject, triple1->predicate, triple1->object,
                                             triple1->subject_type, triple1->object_type);
    list_add_data(query_triples, query_triple);

    query_triple = sslog_new_triple_detached(triple2->subject, triple2->predicate, triple2->object,
                                             triple2->subject_type, triple2->object_type);
    list_add_data(query_triples, query_triple);

    query_triple = sslog_new_triple_detached(triple3->subject, triple3->predicate, triple3->object,
                                             triple3->subject_type, triple3->object_type);
    list_add_data(query_triples, query_triple);



    triples = sslog_node_query_triples(node, query_triples);
    CU_ASSERT_EQUAL_FATAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3);

    list_for_each(list_walker, &triples->links) {
        list_t *node = list_entry(list_walker, list_t, links);
        sslog_triple_t *triple = (sslog_triple_t *) node->data;

        if (sslog_equal_triples(triple, triple1) == true
                || sslog_equal_triples(triple, triple2) == true
                || sslog_equal_triples(triple, triple3) == true) {
            continue;
        }

        CU_FAIL("Query triples failed.");
    }

    list_free_with_nodes(triples, NULL);


    result = sslog_node_remove_triples(node, query_triples);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);

    triples = sslog_node_query_triples(node, query_triples);
    CU_ASSERT_EQUAL(result, SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(sslog_error_get_code(node), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL(list_count(triples), 0);

    list_free_with_nodes(triples, NULL);
    list_free_with_nodes(query_triples, SSLOG_FREE_FUNC sslog_free_triple);
}



/* The main() function for setting up and running the tests.
 * Returns a CUE_SUCCESS on successful running, another
 * CUnit error code on failure.
 */
int main()
{
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("Suite: session and nodes", init_suite_nodes, clean_suite_nodes);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }



    /* add the tests to the suite */

    if (//(NULL == CU_add_test(pSuite, "null triple insert", test_null_triple_insert)) ||
        //(NULL == CU_add_test(pSuite, "null triple list insert", test_null_triples_insert))
            //|| (NULL == CU_add_test(pSuite, "insert/query/delete triple", test_uri_triple_insert_query_delete))
             (NULL == CU_add_test(pSuite, "insert/query/delete triples", test_uri_triples_insert_query_delete))
            )
    {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();

    int failed_tests = CU_get_number_of_tests_failed();

    CU_cleanup_registry();

    return failed_tests;
}
