
#include "tests.h"


#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>


/* The suite initialization function.
 * Opens the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int init_suite_nodes(void)
{
    if (sslog_init() == SSLOG_ERROR_NO) {
        return 0;
    } else {
        return -1;
    }
}

/* The suite cleanup function.
 * Closes the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int clean_suite_nodes(void)
{
    sslog_shutdown();

    return 0;
}


void clean_local_store()
{
    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}



void test_sslog_generate_uri()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 SSLOG_TEST_URI "Class1");

    char *uri1 = sslog_generate_uri(ont_class1);
    char *uri2 = sslog_generate_uri(ont_class1);

    CU_ASSERT_STRING_NOT_EQUAL_FATAL(uri1, uri2);

    free(uri1);
    free(uri2);
}



void test_subentity()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_class_t *ont_class2 = sslog_new_class(SSLOG_TEST_URI "Class2");
    sslog_property_t *ont_prop1 = sslog_new_property(SSLOG_TEST_URI "Prop1", SSLOG_PROPERTY_TYPE_DATA);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 SSLOG_TEST_URI "Class1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class2)),
                                 SSLOG_TEST_URI "Class2");

    sslog_individual_t *ind1 = sslog_new_individual(ont_class1, NULL);
    sslog_individual_t *ind2 = sslog_new_individual(ont_class2, sslog_entity_get_uri(sslog_individual_as_entity(ind1)));

    sslog_insert_property(ind1, ont_prop1, "test");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_get_property(ind2, ont_prop1), "test");

}


/**
 * @brief Tests URI of new individuals.
 * Create a class and 3 individuals.
 * Each individual must have uniq URI.
 */
void test_sslog_new_individual_uri_generation()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");

    CU_ASSERT_PTR_NOT_NULL_FATAL(ont_class1);

    sslog_individual_t *ind1 = sslog_new_individual(ont_class1, NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ind1);

    sslog_individual_t *ind2 = sslog_new_individual(ont_class1, NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ind2);

    sslog_individual_t *ind3 = sslog_new_individual(ont_class1, NULL);
    CU_ASSERT_PTR_NOT_NULL_FATAL(ind3);

    CU_ASSERT_STRING_NOT_EQUAL_FATAL(sslog_entity_get_uri(ind1), sslog_entity_get_uri(ind2));
    CU_ASSERT_STRING_NOT_EQUAL_FATAL(sslog_entity_get_uri(ind1), sslog_entity_get_uri(ind3));
    CU_ASSERT_STRING_NOT_EQUAL_FATAL(sslog_entity_get_uri(ind2), sslog_entity_get_uri(ind3));
}




void test_new_entities_and_triple_converting()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 SSLOG_TEST_URI "Class1");

    sslog_class_t *ont_class2 = sslog_new_class(SSLOG_TEST_URI "Class2");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class2)),
                                 SSLOG_TEST_URI "Class2");

    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_data)),
                                 SSLOG_TEST_URI "Property1D");

    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_object)),
                                 SSLOG_TEST_URI "Property2O");


    sslog_individual_t *ind1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind1)),
                                 SSLOG_TEST_URI "Intd1");

    sslog_individual_t *ind2 = sslog_new_individual(ont_class2, SSLOG_TEST_URI "Intd2");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind2)),
                                 SSLOG_TEST_URI "Intd2");


    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 sslog_class_to_triple(ont_class1)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class2)),
                                 sslog_class_to_triple(ont_class2)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_data)),
                                 sslog_property_to_triple(property_data)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_object)),
                                 sslog_property_to_triple(property_object)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind1)),
                                 sslog_individual_to_triple(ind1)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind2)),
                                 sslog_individual_to_triple(ind2)->subject);


    sslog_triple_t *triple = sslog_entity_to_triple(sslog_class_as_entity(ont_class1));
    sslog_entity_t *entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, ont_class1);

    triple = sslog_entity_to_triple(sslog_individual_as_entity(ind1));
    sslog_class_t *ont_class = sslog_class_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(ont_class, ont_class1);


    triple = sslog_entity_to_triple(sslog_object_as_entity(property_data));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, property_data);


    triple = sslog_entity_to_triple(sslog_object_as_entity(property_object));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, property_object);

    triple = sslog_entity_to_triple(sslog_object_as_entity(ind1));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, ind1);


    sslog_triple_t *prop_value_triples = sslog_new_triple(sslog_entity_get_uri(ind1), sslog_entity_get_uri(property_data), "value1", SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);

    // From triples for property value owner must be returned.
    entity = sslog_entity_from_triple(prop_value_triples);
    CU_ASSERT_PTR_EQUAL_FATAL(entity, ind1);

    sslog_individual_t *test_ind = sslog_individual_from_triple(prop_value_triples);
    CU_ASSERT_PTR_EQUAL_FATAL(test_ind, ind1);

    sslog_property_t *test_prop = sslog_property_from_triple(prop_value_triples);
    CU_ASSERT_PTR_EQUAL_FATAL(test_prop, property_data);

    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}

/**
 * @brief Testing removing property.
 * Sets data and object properties for individual.
 * Removes properties and check count of triples and properties.
 */
void test_remove_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value2");
    sslog_insert_property(individual1, property_data, "value3");
    sslog_insert_property(individual1, property_data, "value4");

    sslog_insert_property(individual1, property_object, individual2);
    sslog_insert_property(individual1, property_object, individual3);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),4);   // 4 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_property(individual1, property_data, "value2");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 6); // 5 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),3);   // 3 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_property(individual1, property_data, "value1");
    sslog_remove_property(individual1, property_data, "value2");
    sslog_remove_property(individual1, property_data, "value3");
    sslog_remove_property(individual1, property_object, individual2);


    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property value
    list_free_with_nodes(values, NULL);

    sslog_remove_property(individual1, property_data, "value1");
    sslog_remove_property(individual1, property_data, "value2");
    sslog_remove_property(individual1, property_data, "value3");
    sslog_remove_property(individual1, property_data, "value4");
    sslog_remove_property(individual1, property_object, individual2);
    sslog_remove_property(individual1, property_object, individual3);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 0 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property value
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);


    triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}



/**
 * @brief Testing removing properties.
 * Sets data and object properties for individual.
 * Removes properties and check count of triples and properties.
 */
void test_remove_properties()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value2");
    sslog_insert_property(individual1, property_data, "value3");
    sslog_insert_property(individual1, property_data, "value4");

    sslog_insert_property(individual1, property_object, individual2);
    sslog_insert_property(individual1, property_object, individual3);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),4);   // 4 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_properties(individual1, property_data);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_properties(individual1, property_object);


    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 1 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 1 object property value
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);

    clean_local_store();
}



/**
 * @brief Testing removing properties.
 * Update data and object properties for individual and check coun of properties
 * and their values.
 */
void test_update_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_update_property(individual1, property_object, individual2, NULL);
    sslog_update_property(individual1, property_data, "value1", NULL);


    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 0 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property values
    list_free_with_nodes(values, NULL);


    sslog_update_property(individual1, property_object, NULL, individual2);
    sslog_update_property(individual1, property_data, NULL, "value1");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    const void *value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value1");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual2);


    sslog_update_property(individual1, property_object, NULL, NULL);
    sslog_update_property(individual1, property_data, NULL, NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value1");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual2);


    sslog_update_property(individual1, property_object, individual2, individual3);
    sslog_update_property(individual1, property_data, "value1", "value2");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value2");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual3);


    sslog_update_property(individual1, property_object, individual2, NULL);
    sslog_update_property(individual1, property_data, "value1", NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value2");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual3);


    sslog_update_property(individual1, property_object, individual3, NULL);
    sslog_update_property(individual1, property_data, "value2", NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property values
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);

    clean_local_store();
}



/**
 * @brief Testing getting individual.
 * Creates individual by new-function and by creating triples and tryes to get them.
 */
void test_get_individual()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_class_t *ont_class2 = sslog_new_class(SSLOG_TEST_URI "Class2");
    sslog_class_t *ont_class3 = sslog_new_class(SSLOG_TEST_URI "Class3");

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class2, SSLOG_TEST_URI "Intd2");

    sslog_individual_t *check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd1");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual1);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual2);


    sslog_triple_t *ind_triple = sslog_new_triple(SSLOG_TEST_URI "Intd3", SSLOG_TRIPLE_RDF_TYPE, sslog_entity_get_uri(ont_class3),
                     SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd3");
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(check_ind), SSLOG_TEST_URI "Intd3");

    sslog_individual_t *individual3 = sslog_new_individual(ont_class3, SSLOG_TEST_URI "Intd3");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual3);


    sslog_remove_individual(individual1);
    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd1");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);


    sslog_remove_triple(ind_triple);
    sslog_free_triple(ind_triple);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd3");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);


    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TEST_URI "Intd2", SSLOG_TRIPLE_RDF_TYPE, sslog_entity_get_uri(ont_class2),
                              SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);


    sslog_remove_triple(triple_template);
    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);

    sslog_free_triple(triple_template);

    clean_local_store();
}



/**
 * @brief Testing removing induvudaul from local store.
 * Creates individual and add some properties to it.
 * Gets triples that describes individual and properties before removing and after removing.
 */
void test_remove_individual()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value2");
    sslog_insert_property(individual1, property_data, "value3");
    sslog_insert_property(individual1, property_data, "value4");

    sslog_insert_property(individual1, property_object, individual2);
    sslog_insert_property(individual1, property_object, individual3);


    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);

    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 property valuer + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    sslog_remove_individual(individual1);

    triples = sslog_query_triple(triple_template);

    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);
    list_free_with_nodes(triples, NULL);

    sslog_free_triple(triple_template);

    triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}


void test_add_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value1");

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                sslog_entity_get_uri(property_data), SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);

    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1);

    list_free_with_nodes(triples, NULL);

    sslog_free_triple(triple_template);


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                    sslog_entity_get_uri(property_data), "value1",
                                                                    SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    triples = sslog_query_triple(triple_template);

    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);

    list_free_with_nodes(triples, NULL);


    triple_template->object_type = SSLOG_RDF_TYPE_LIT;

    triples = sslog_query_triple(triple_template);

    sslog_triple_t *triple = (sslog_triple_t *) list_get_first_prev_node(triples)->data;

    list_free_with_nodes(triples, NULL);

    CU_ASSERT_PTR_NOT_NULL_FATAL(triple);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(individual1),
                                 triple->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(property_data),
                                 triple->predicate);

    CU_ASSERT_STRING_EQUAL_FATAL("value1",
                                 triple->object);

    sslog_free_triple(triple_template);

    triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}


/* The main() function for setting up and running the tests.
 * Returns a CUE_SUCCESS on successful running, another
 * CUnit error code on failure.
 */
int main()
{
    CU_pSuite pSuite = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();

    /* add a suite to the registry */
    pSuite = CU_add_suite("Suite: entities", init_suite_nodes, clean_suite_nodes);
    if (NULL == pSuite) {
        CU_cleanup_registry();
        return CU_get_error();
    }



    /* add the tests to the suite */
    if ((NULL == CU_add_test(pSuite, "Creating entities, converting", test_new_entities_and_triple_converting))
            || (NULL == CU_add_test(pSuite, "Generating URI", test_sslog_generate_uri))
            || (NULL == CU_add_test(pSuite, "Sub entity properties", test_subentity))
            || (NULL == CU_add_test(pSuite, "Generation of individuals URI", test_sslog_new_individual_uri_generation))
            || (NULL == CU_add_test(pSuite, "Adding property", test_add_property))
            || (NULL == CU_add_test(pSuite, "Removing property", test_remove_property))
            || (NULL == CU_add_test(pSuite, "Removing properties", test_remove_properties))
            || (NULL == CU_add_test(pSuite, "Updating properties", test_update_property))
            || (NULL == CU_add_test(pSuite, "Getting individual", test_get_individual))
            || (NULL == CU_add_test(pSuite, "Removing individual", test_remove_individual)))
    {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();

    int failed_tests = CU_get_number_of_tests_failed();

    CU_cleanup_registry();

    return failed_tests;
}
