
#include "tests.h"

#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>


sslog_node_t *node;

void clean_ss();

/* The suite initialization function.
 * Opens the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int init_suite_nodes(void)
{
    if (sslog_init() != SSLOG_ERROR_NO) {
        return -1;
    }

    node = sslog_new_node("testNodeId_low_api",
                          SSLOG_TESTS_SS_ID, SSLOG_TESTS_SS_ADDRESS, SSLOG_TESTS_SS_PORT);

    if (node == NULL || sslog_error_get_last_code() != SSLOG_ERROR_NO) {
        return sslog_error_get_last_code();
    }

    if (sslog_node_join(node) == SSLOG_ERROR_NO){
            clean_ss(node);
    } else {
        return -1;
    }

    return 0;
}

/* The suite cleanup function.
 * Closes the temporary file used by the tests.
 * Returns zero on success, non-zero otherwise.
 */
int clean_suite_nodes(void)
{

    sslog_node_leave(node);
    sslog_shutdown();

    return 0;
}


void clean_local_store()
{
    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}

void clean_ss()
{
    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_node_remove_triple(node, triple_template);

    sslog_free_triple(triple_template);
}



/**
 * @brief Testing SPARQL ASK query (simple form).
 * 1) Checks existence of triples.
 * 2) Inserts these triples to smart space.
 * 3) Checks existence again.
 * 4) Removes triples.
 * 5) Checks existence again.
 */
void test_node_sparql_ask()
{
    const char query[] = "ASK { ?x <" SSLOG_TEST_URI "name>  \"Alice\" ;"
             "<" SSLOG_TEST_URI "parent>  <" SSLOG_TEST_URI "Batka1>}\0";

    CU_ASSERT_EQUAL_FATAL(sslog_node_sparql_ask(node, query), false);
    CU_ASSERT_EQUAL_FATAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);

    list_t *triples = list_new();
    sslog_triple_t *triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "name", "Alice",
                                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "parent", SSLOG_TEST_URI "Batka1",
                                                           SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);


    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_triples(node, triples), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_sparql_ask(node, query), true);


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_triples(node, triples), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_sparql_ask(node, query), false);
    CU_ASSERT_EQUAL_FATAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);

    sslog_free_triples(triples);

    clean_local_store();
    clean_ss();
}


/**
 * @brief Testing SPARQL ASK query (simple form).
 * 1) Inserts some triples to smart space.
 * 2) Makes construct query.
 * 3) Checks constructed triples and stored triples after query.
 */
void test_node_sparql_construct()
{
    list_t *triples = list_new();
    sslog_triple_t *triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "name", "Alice",
                                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "surname", SSLOG_TEST_URI "Pupkina",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "parent", SSLOG_TEST_URI "Batka1",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);

    const char query[] = "construct { ?x <http://sslog.ru/test#firstname> ?name } "
                         "where { ?x <http://sslog.ru/test#name> ?name."
                         "?x <http://sslog.ru/test#parent> <http://sslog.ru/test#Batka1> }";

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_triples(node, triples), SSLOG_ERROR_NO);

    sslog_free_triples(triples);

    list_t *constructed_triples = sslog_node_sparql_construct(node, query);
    CU_ASSERT_EQUAL_FATAL(list_count(constructed_triples), 1);

    sslog_triple_t *constructed_triple = (sslog_triple_t *) list_get_first_prev_node(constructed_triples)->data;

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "firstname", SSLOG_TRIPLE_ANY,
                              SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *stored_triples = sslog_query_triple(triple);
    CU_ASSERT_EQUAL_FATAL(list_count(stored_triples), 1);

    sslog_free_triple(triple);

    sslog_triple_t *stored_triple = (sslog_triple_t *) list_get_first_prev_node(stored_triples)->data;

    CU_ASSERT_PTR_EQUAL(stored_triple, constructed_triple);
    CU_ASSERT_STRING_EQUAL_FATAL(stored_triple->predicate, SSLOG_TEST_URI "firstname");

    sslog_free_triples(stored_triples);
    sslog_free_triples(constructed_triples);

    clean_local_store();
    clean_ss();
}


/**
 * @brief Testing SPARQL SELECT query.
 * 1) Inserts some triples to smart space.
 * 2) Makes select query.
 * 3) Checks sparql results and and stored triples after query.
 */
void test_node_sparql_select()
{
    list_t *triples = list_new();
    sslog_triple_t *triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "name", "Alice",
                                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "surname", "Pupkina",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TEST_URI "parent", SSLOG_TEST_URI "Batka1",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Boy1", SSLOG_TEST_URI "name", "Bob",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Boy1", SSLOG_TEST_URI "surname", "Pupkin",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Boy1", SSLOG_TEST_URI "parent", SSLOG_TEST_URI "Batka1",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);


    const char query[] = "select ?a ?nameprop ?surnameprop ?name ?surname "
                         "where { ?a <" SSLOG_TEST_URI "parent" "> <" SSLOG_TEST_URI "Batka1" "> . "
                                 "?a <" SSLOG_TEST_URI "name" "> ?name . "
                                 "?a <" SSLOG_TEST_URI "surname" "> ?surname "
                                 "BIND (<" SSLOG_TEST_URI "name" "> AS ?nameprop) "
                                 "BIND (<" SSLOG_TEST_URI "surname" "> AS ?surnameprop)}";

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_triples(node, triples), SSLOG_ERROR_NO);

    // Remove triples from local store.
    CU_ASSERT_EQUAL_FATAL(sslog_remove_triples(triples), SSLOG_ERROR_NO);

    sslog_free_triples(triples);


    sslog_sparql_result_t *result = sslog_node_sparql_select(node, query, "a nameprop name  .     a surnameprop    surname");
    CU_ASSERT_EQUAL_FATAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(result->rows_count, 2);
    CU_ASSERT_EQUAL_FATAL(result->bindings_count, 5);

    CU_ASSERT_STRING_EQUAL(result->names[0], "a");
    CU_ASSERT_STRING_EQUAL(result->names[1], "nameprop");
    CU_ASSERT_STRING_EQUAL(result->names[2], "surnameprop");
    CU_ASSERT_STRING_EQUAL(result->names[3], "name");
    CU_ASSERT_STRING_EQUAL(result->names[4], "surname");

    CU_ASSERT_STRING_EQUAL(result->rows[0]->values[1],  SSLOG_TEST_URI "name");
    CU_ASSERT_STRING_EQUAL(result->rows[1]->values[1],  SSLOG_TEST_URI "name");

    CU_ASSERT_STRING_EQUAL(result->rows[0]->values[2],  SSLOG_TEST_URI "surname");
    CU_ASSERT_STRING_EQUAL(result->rows[1]->values[2],  SSLOG_TEST_URI "surname");

    if (strncmp(result->rows[0]->values[0], SSLOG_TEST_URI "Boy1", SSLOG_TRIPLE_URI_LEN) == 0) {
        CU_ASSERT_STRING_EQUAL(result->rows[0]->values[3],  "Bob");
        CU_ASSERT_STRING_EQUAL(result->rows[0]->values[4],  "Pupkin");

        CU_ASSERT_STRING_EQUAL(result->rows[1]->values[3],  "Alice");
        CU_ASSERT_STRING_EQUAL(result->rows[1]->values[4],  "Pupkina");
    } else {
        CU_ASSERT_STRING_EQUAL(result->rows[1]->values[3],  "Bob");
        CU_ASSERT_STRING_EQUAL(result->rows[1]->values[4],  "Pupkin");

        CU_ASSERT_STRING_EQUAL(result->rows[0]->values[3],  "Alice");
        CU_ASSERT_STRING_EQUAL(result->rows[0]->values[4],  "Pupkina");
    }



    sslog_free_sparql_result(result);

    triples = list_new();
    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Boy1", SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Girl1", SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_add_data(triples, triple);

    list_t *stored_triples = sslog_query_triples(triples);
    CU_ASSERT_EQUAL_FATAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(list_count(stored_triples), 4);

    sslog_free_triples(stored_triples);
    sslog_free_triples(triples);

    clean_local_store();
    clean_ss();
}


/**
 * @brief Testing SPARQL-endpoint SELECT query.
 * 1) Inserts some triples to smart space.
 * 2) Makes select query.
 * 3) Checks sparql results and and stored triples after query.
 */
void test_node_sparql_enpoint_select()
{
    static const char *endpoint_address = "http://dbpedia.org/sparql";
    static const char *endpoint_query = "select distinct ?Concept where {[] a ?Concept} LIMIT 5";
    static const char *endpoint_extra = "&format=application/xml";


    sslog_sparql_result_t *result = sslog_node_sparql_enpoint_select(node,  endpoint_address, endpoint_query, endpoint_extra, NULL);


    CU_ASSERT_EQUAL_FATAL(sslog_error_get_last_code(), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(result->rows_count, 5);
    CU_ASSERT_EQUAL_FATAL(result->bindings_count, 1);


    sslog_free_sparql_result(result);

    clean_local_store();
    clean_ss();
}


/**
 * @brief Testing inserting property.
 * Creates individual,
 * adds properties to it,
 * locally getting properties,
 * query triples from smart space and checks.
 */
void test_node_insert_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data, "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data, "value2"), SSLOG_ERROR_NO);

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object, individual3), SSLOG_ERROR_NO);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);

    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);

    list_free_with_nodes(values, NULL);


    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                        SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 5);   // 4 properties + 1 rdf:type


    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    clean_local_store();
    clean_ss();
}

/**
 * @brief Testing removing property with node.
 * Creates individual,
 * adds properties to it,
 * locally and remotley getting properties,
 * query triples from smart space and checks.
 */
void test_node_remove_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data1, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data1, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data1, "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data2, "value2.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_data2, "value2.2"), SSLOG_ERROR_NO);

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object1, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object1, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object2, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_property(node, individual1, property_object2, individual3), SSLOG_ERROR_NO);

    list_t *values = sslog_get_properties(individual1, property_data1);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_data2);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object1);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object2);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);
    list_free_with_nodes(values, NULL);


    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_data1), SSLOG_TRIPLE_ANY,
                                                        SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 2);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_data2), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 2);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object1), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, NULL);


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object2), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 2);
    sslog_free_triple(triple_template);
    sslog_free_triples(result_triples);


    clean_local_store();
    clean_ss();

    return;


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data1, "value2"), SSLOG_ERROR_NO);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_get_property(individual1, property_data1), "value1");


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_data1), SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);
    sslog_triple_t *result_triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;
    CU_ASSERT_STRING_EQUAL_FATAL(result_triple->object, "value1");
    sslog_free_triples(result_triples);


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data1, "value1"), SSLOG_ERROR_NO);

    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_data1));

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data2, "value2.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data2, "value2.2"), SSLOG_ERROR_NO);

    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_data2));


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_data2), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, NULL);


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_object1, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_object2, individual3), SSLOG_ERROR_NO);

    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_object1));
    CU_ASSERT_PTR_EQUAL_FATAL(sslog_get_property(individual1, property_object2), individual2);

    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object1), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, NULL);


    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object2), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, NULL);



    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_object2, individual2), SSLOG_ERROR_NO);

    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_object2));

    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object2), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    sslog_free_triple(triple_template);
    list_free_with_nodes(result_triples, NULL);


    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data1, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_data2, "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_object1, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_property(node, individual1, property_object2, individual3), SSLOG_ERROR_NO);

    clean_local_store();
    clean_ss();
}



/**
 * @brief Testing updating property with node.
 * Creates individual,
 * update properties,
 * check triples in the store an smart space.
 */
void test_node_update_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, NULL, NULL), SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, NULL, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, NULL, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_get_property(individual1, property_data1), "value1");

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_data1), SSLOG_TRIPLE_ANY,
                                                        SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    sslog_triple_t *triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;
    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, "value1");
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;

    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, "value1");
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);



    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, "value1", "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, "value1", "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, "value2", "value2"), SSLOG_ERROR_NO);
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_get_property(individual1, property_data1), "value2");

    result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;
    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, "value2");
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;

    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, "value2");
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);


    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, "value2", NULL), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_data1, "value2", NULL), SSLOG_ERROR_NO);
    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_data1));

    result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    sslog_free_triple(triple_template);


    // Object property
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, NULL, NULL), SSLOG_ERROR_NULL_ARGUMENT);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, NULL, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, NULL, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_PTR_EQUAL_FATAL(sslog_get_property(individual1, property_object1), individual2);

    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), sslog_entity_get_uri(property_object1), SSLOG_TRIPLE_ANY,
                                                            SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;
    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, sslog_entity_get_uri(individual2));
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;

    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, sslog_entity_get_uri(individual2));
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);



    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, individual2, individual3), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, individual2, individual3), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, individual3, individual3), SSLOG_ERROR_NO);
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_get_property(individual1, property_object1), individual3);

    result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;
    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, sslog_entity_get_uri(individual3));
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 1);

    triple = (sslog_triple_t *) list_get_first_prev_node(result_triples)->data;

    CU_ASSERT_STRING_EQUAL_FATAL(triple->object, sslog_entity_get_uri(individual3));
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);


    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, individual3, NULL), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_node_update_property(node, individual1, property_object1, individual3, NULL), SSLOG_ERROR_NO);
    CU_ASSERT_PTR_NULL_FATAL(sslog_get_property(individual1, property_object1));

    result_triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    result_triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(result_triples), 0);
    list_free_with_nodes(result_triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    sslog_free_triple(triple_template);

    clean_local_store();
    clean_ss();
}





/**
 * @brief Testing populating of individual.
 * Create an local individual,
 * insert some triples to the smart space,
 * removes local triples and populates an individual.
 */
void test_node_populate()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");

    list_t *triples = list_new();

    sslog_triple_t *triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                       sslog_entity_get_uri(property_data1),
                                                       "value1",
                                                       SSLOG_RDF_TYPE_URI,
                                                       SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                           sslog_entity_get_uri(property_data2),
                                                           "value2.1",
                                                           SSLOG_RDF_TYPE_URI,
                                                           SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                           sslog_entity_get_uri(property_data2),
                                                           "value2.2",
                                                           SSLOG_RDF_TYPE_URI,
                                                           SSLOG_RDF_TYPE_LIT);
    list_add_data(triples, triple);

    // For object property.
    triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                       sslog_entity_get_uri(property_object1),
                                       SSLOG_TEST_URI "Intd2",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);

    // For individual.
    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Intd2",
                                       SSLOG_TRIPLE_RDF_TYPE,
                                       sslog_entity_get_uri(ont_class1),
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);


    // For object property.
    triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                       sslog_entity_get_uri(property_object2),
                                       SSLOG_TEST_URI "Intd2.1",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);

    // For individual.
    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Intd2.1",
                                       SSLOG_TRIPLE_RDF_TYPE,
                                       sslog_entity_get_uri(ont_class1),
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);


    // For object property.
    triple = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                       sslog_entity_get_uri(property_object2),
                                       SSLOG_TEST_URI "Intd2.2",
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);

    // For individual.
    triple = sslog_new_triple_detached(SSLOG_TEST_URI "Intd2.2",
                                       SSLOG_TRIPLE_RDF_TYPE,
                                       sslog_entity_get_uri(ont_class1),
                                       SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_add_data(triples, triple);

    sslog_node_insert_triples(node, triples);
    sslog_remove_triples(triples);


    CU_ASSERT_EQUAL_FATAL(sslog_node_populate(node, individual1), SSLOG_ERROR_NO);

    char *data_value = (char *) sslog_get_property(individual1, property_data1);
    CU_ASSERT_STRING_EQUAL_FATAL(data_value, "value1");

    list_t *data_values = sslog_get_properties(individual1, property_data2);
    CU_ASSERT_EQUAL(list_count(data_values), 2);

    list_free_with_nodes(data_values, NULL);

    sslog_individual_t *object_value = (sslog_individual_t *) sslog_get_property(individual1, property_object1);
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(object_value), SSLOG_TEST_URI "Intd2");

    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_EQUAL_FATAL(object_value, individual2);

    list_t *object_values = sslog_get_properties(individual1, property_object2);
    CU_ASSERT_EQUAL(list_count(object_values), 2);

    list_free_with_nodes(object_values, NULL);

    list_free_with_nodes(triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    clean_local_store();
    clean_ss();
}


/**
 * @brief Testing inserting of individual.
 * Creates a local individual,
 * sets properties,
 * inserts an individual,
 * gets triples from smart space and checks it.
 */
void test_node_insert()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");
    sslog_individual_t *individual4 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd4");

    sslog_insert_property(individual1, property_data1, "value1");
    sslog_insert_property(individual1, property_data2, "value2.1");
    sslog_insert_property(individual1, property_data2, "value2.2");
    sslog_insert_property(individual1, property_data2, "value2.2"); // Ignore value

    sslog_insert_property(individual1, property_object1, individual2);
    sslog_insert_property(individual1, property_object2, individual3);
    sslog_insert_property(individual1, property_object2, individual4);
    sslog_insert_property(individual1, property_object2, individual4); // Ignore value

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_individual(node, individual1), SSLOG_ERROR_NO);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), SSLOG_TRIPLE_ANY,
                                                                SSLOG_TRIPLE_ANY, SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    list_t *triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 properties + rdf:type
    sslog_free_triple(triple_template);

    individual1 = sslog_get_individual(SSLOG_TEST_URI "Intd1");

    char *data_value = (char *) sslog_get_property(individual1, property_data1);
    CU_ASSERT_STRING_EQUAL_FATAL(data_value, "value1");

    list_t *data_values = sslog_get_properties(individual1, property_data2);
    CU_ASSERT_EQUAL(list_count(data_values), 2);

    list_free_with_nodes(data_values, NULL);

    sslog_individual_t *object_value = (sslog_individual_t *) sslog_get_property(individual1, property_object1);
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(object_value), SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_EQUAL_FATAL(object_value, individual2);

    list_t *object_values = sslog_get_properties(individual1, property_object2);
    CU_ASSERT_EQUAL(list_count(object_values), 2);

    list_free_with_nodes(object_values, NULL);

    list_free_with_nodes(triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    clean_local_store();
    clean_ss();
}



/**
 * @brief Testing removing of individual.
 * Creates a local individual,
 * sets properties,
 * inserts an individual,
 * removes some properties localy,
 * removes individual,
 * gets triples from smart space and checks it.
 */
void test_node_remove()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");
    sslog_individual_t *individual4 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd4");

    sslog_insert_property(individual1, property_data1, "value1");
    sslog_insert_property(individual1, property_data2, "value2.1");
    sslog_insert_property(individual1, property_data2, "value2.2");
    sslog_insert_property(individual1, property_data2, "value2.2"); // Ignore value

    sslog_insert_property(individual1, property_object1, individual2);
    sslog_insert_property(individual1, property_object2, individual3);
    sslog_insert_property(individual1, property_object2, individual4);
    sslog_insert_property(individual1, property_object2, individual4); // Ignore value

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_individual(node, individual1), SSLOG_ERROR_NO);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), SSLOG_TRIPLE_ANY,
                                                                SSLOG_TRIPLE_ANY, SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    // Checks: only rdf:type triple is in store
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7);  // 6 properties + rdf:type
    list_free_with_nodes(triples, NULL);

    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_individual(node, individual1), SSLOG_ERROR_NO);

    // Checks: only rdf:type triple is in store
    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1);

    sslog_triple_t *rdftype = (sslog_triple_t *) list_get_first_prev_node(triples)->data;
    CU_ASSERT_EQUAL(sslog_triple_is_type(rdftype), true);

    list_free_with_nodes(triples, NULL);

    // Checks triple in the smart space: no triples.
    triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);

    sslog_free_triple(triple_template);

    list_free_with_nodes(triples, LIST_CAST_TO_FREE_FUNC sslog_free_triple);

    clean_local_store();
    clean_ss();
}



/**
 * @brief Testing removing of individual in the smart space ind locally.
 * Creates a local individual,
 * sets properties,
 * inserts an individual,
 * removes individual,
 * gets triples from local store and from smart space and checks it.
 */
void test_node_remove_with_local()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");
    sslog_individual_t *individual4 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd4");

    sslog_insert_property(individual1, property_data1, "value1");
    sslog_insert_property(individual1, property_data2, "value2.1");
    sslog_insert_property(individual1, property_data2, "value2.2");
    sslog_insert_property(individual1, property_data2, "value2.2"); // Ignore value

    sslog_insert_property(individual1, property_object1, individual2);
    sslog_insert_property(individual1, property_object2, individual3);
    sslog_insert_property(individual1, property_object2, individual4);
    sslog_insert_property(individual1, property_object2, individual4); // Ignore value

    CU_ASSERT_EQUAL_FATAL(sslog_node_insert_individual(node, individual1), SSLOG_ERROR_NO);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1), SSLOG_TRIPLE_ANY,
                                                                SSLOG_TRIPLE_ANY, SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    // Checks: only rdf:type triple is in store
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7);  // 6 properties + rdf:type
    list_free_with_nodes(triples, NULL);

    CU_ASSERT_EQUAL_FATAL(sslog_node_remove_individual_with_local(node, individual1), SSLOG_ERROR_NO);

    // Checks: only rdf:type triple is in store
    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);

    sslog_free_triples(triples);

    // Checks triple in the smart space: no triples.
    triples = sslog_node_query_triple(node, triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);

    sslog_free_triple(triple_template);

    sslog_free_triples(triples);

    clean_local_store();
    clean_ss();
}




void test_new_entities_and_triple_converting()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 SSLOG_TEST_URI "Class1");

    sslog_class_t *ont_class2 = sslog_new_class(SSLOG_TEST_URI "Class2");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class2)),
                                 SSLOG_TEST_URI "Class2");

    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_data)),
                                 SSLOG_TEST_URI "Property1D");

    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_object)),
                                 SSLOG_TEST_URI "Property2O");


    sslog_individual_t *ind1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind1)),
                                 SSLOG_TEST_URI "Intd1");

    sslog_individual_t *ind2 = sslog_new_individual(ont_class2, SSLOG_TEST_URI "Intd2");

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind2)),
                                 SSLOG_TEST_URI "Intd2");


    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class1)),
                                 sslog_class_to_triple(ont_class1)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ont_class2)),
                                 sslog_class_to_triple(ont_class2)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_data)),
                                 sslog_property_to_triple(property_data)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(property_object)),
                                 sslog_property_to_triple(property_object)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind1)),
                                 sslog_individual_to_triple(ind1)->subject);

    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(sslog_class_as_entity(ind2)),
                                 sslog_individual_to_triple(ind2)->subject);


    sslog_triple_t *triple = sslog_entity_to_triple(sslog_class_as_entity(ont_class1));
    sslog_entity_t *entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, ont_class1);

    triple = sslog_entity_to_triple(sslog_object_as_entity(property_data));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, property_data);


    triple = sslog_entity_to_triple(sslog_object_as_entity(property_object));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, property_object);

    triple = sslog_entity_to_triple(sslog_object_as_entity(ind1));
    entity = sslog_entity_from_triple(triple);

    CU_ASSERT_PTR_EQUAL_FATAL(entity, ind1);


    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}

/**
 * @brief Testing removing property.
 * Sets data and object properties for individual.
 * Removes properties and check count of triples and properties.
 */
void test_remove_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value2");
    sslog_insert_property(individual1, property_data, "value3");
    sslog_insert_property(individual1, property_data, "value4");

    sslog_insert_property(individual1, property_object, individual2);
    sslog_insert_property(individual1, property_object, individual3);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),4);   // 4 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_property(individual1, property_data, "value2");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 6); // 5 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),3);   // 3 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_property(individual1, property_data, "value1");
    sslog_remove_property(individual1, property_data, "value2");
    sslog_remove_property(individual1, property_data, "value3");
    sslog_remove_property(individual1, property_object, individual2);


    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property value
    list_free_with_nodes(values, NULL);

    sslog_remove_property(individual1, property_data, "value1");
    sslog_remove_property(individual1, property_data, "value2");
    sslog_remove_property(individual1, property_data, "value3");
    sslog_remove_property(individual1, property_data, "value4");
    sslog_remove_property(individual1, property_object, individual2);
    sslog_remove_property(individual1, property_object, individual3);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 0 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property value
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);


    triple_template = sslog_new_triple_detached(SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    sslog_remove_triple(triple_template);

    sslog_free_triple(triple_template);
}



/**
 * @brief Testing removing properties.
 * Sets data and object properties for individual.
 * Removes properties and check count of triples and properties.
 */
void test_remove_properties()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_insert_property(individual1, property_data, "value1");
    sslog_insert_property(individual1, property_data, "value2");
    sslog_insert_property(individual1, property_data, "value3");
    sslog_insert_property(individual1, property_data, "value4");

    sslog_insert_property(individual1, property_object, individual2);
    sslog_insert_property(individual1, property_object, individual3);

    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 7); // 6 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values),4);   // 4 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_properties(individual1, property_data);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 2);   // 2 object property values
    list_free_with_nodes(values, NULL);


    sslog_remove_properties(individual1, property_object);


    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 1 data property value
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 1 object property value
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);

    clean_local_store();
}



/**
 * @brief Testing removing properties.
 * Update data and object properties for individual and check coun of properties
 * and their values.
 */
void test_update_property()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_property_t *property_data = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd3");

    sslog_update_property(individual1, property_object, individual2, NULL);
    sslog_update_property(individual1, property_data, "value1", NULL);


    sslog_triple_t *triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual1),
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);
    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 0 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    list_t *values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property values
    list_free_with_nodes(values, NULL);


    sslog_update_property(individual1, property_object, NULL, individual2);
    sslog_update_property(individual1, property_data, NULL, "value1");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    const void *value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value1");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual2);


    sslog_update_property(individual1, property_object, NULL, NULL);
    sslog_update_property(individual1, property_data, NULL, NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value1");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual2);


    sslog_update_property(individual1, property_object, individual2, individual3);
    sslog_update_property(individual1, property_data, "value1", "value2");

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value2");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual3);


    sslog_update_property(individual1, property_object, individual2, NULL);
    sslog_update_property(individual1, property_data, "value1", NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 3); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 1);   // 1 object property values
    list_free_with_nodes(values, NULL);

    value = sslog_get_property(individual1, property_data);
    CU_ASSERT_STRING_EQUAL_FATAL(value, "value2");

    value = sslog_get_property(individual1, property_object);
    CU_ASSERT_PTR_EQUAL(value, individual3);


    sslog_update_property(individual1, property_object, individual3, NULL);
    sslog_update_property(individual1, property_data, "value2", NULL);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 1); // 2 property values + 1 individual rdf:type
    list_free_with_nodes(triples, NULL);

    values = sslog_get_properties(individual1, property_data);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 data property values
    list_free_with_nodes(values, NULL);

    values = sslog_get_properties(individual1, property_object);
    CU_ASSERT_EQUAL_FATAL(list_count(values), 0);   // 0 object property values
    list_free_with_nodes(values, NULL);

    sslog_free_triple(triple_template);

    clean_local_store();
}



/**
 * @brief Testing getting individual.
 * Creates individual by new-function and by creating triples and tryes to get them.
 */
void test_get_individual()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");
    sslog_class_t *ont_class2 = sslog_new_class(SSLOG_TEST_URI "Class2");
    sslog_class_t *ont_class3 = sslog_new_class(SSLOG_TEST_URI "Class3");

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Intd1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class2, SSLOG_TEST_URI "Intd2");

    sslog_individual_t *check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd1");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual1);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual2);


    sslog_triple_t *ind_triple = sslog_new_triple(SSLOG_TEST_URI "Intd3", SSLOG_TRIPLE_RDF_TYPE, sslog_entity_get_uri(ont_class3),
                     SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd3");
    CU_ASSERT_STRING_EQUAL_FATAL(sslog_entity_get_uri(check_ind), SSLOG_TEST_URI "Intd3");

    sslog_individual_t *individual3 = sslog_new_individual(ont_class3, SSLOG_TEST_URI "Intd3");
    CU_ASSERT_PTR_EQUAL_FATAL(check_ind, individual3);


    sslog_remove_individual(individual1);
    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd1");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);


    sslog_remove_triple(ind_triple);
    sslog_free_triple(ind_triple);

    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd3");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);


    sslog_triple_t *triple_template = sslog_new_triple_detached(SSLOG_TEST_URI "Intd2", SSLOG_TRIPLE_RDF_TYPE, sslog_entity_get_uri(ont_class2),
                              SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);


    sslog_remove_triple(triple_template);
    check_ind = sslog_get_individual(SSLOG_TEST_URI "Intd2");
    CU_ASSERT_PTR_NULL_FATAL(check_ind);

    sslog_free_triple(triple_template);

    clean_local_store();
}

/**
 * @brief Testing local removing of individual.
 * Create an individual an set different properties to it,
 * removes the individual and checks local store.
 */
void test_local_remove_individual()
{
    sslog_class_t *ont_class1 = sslog_new_class(SSLOG_TEST_URI "Class1");

    sslog_property_t *property_data1 = sslog_new_property(SSLOG_TEST_URI "Property1D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data2 = sslog_new_property(SSLOG_TEST_URI "Property2D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_data3 = sslog_new_property(SSLOG_TEST_URI "Property3D", SSLOG_PROPERTY_TYPE_DATA);
    sslog_property_t *property_object1 = sslog_new_property(SSLOG_TEST_URI "Property1O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object2 = sslog_new_property(SSLOG_TEST_URI "Property2O", SSLOG_PROPERTY_TYPE_OBJECT);
    sslog_property_t *property_object3 = sslog_new_property(SSLOG_TEST_URI "Property3O", SSLOG_PROPERTY_TYPE_OBJECT);

    sslog_individual_t *individual1 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Ind1");
    sslog_individual_t *individual2 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Ind2");
    sslog_individual_t *individual3 = sslog_new_individual(ont_class1, SSLOG_TEST_URI "Ind3");

    // Properties to individual1.
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data1, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data2, "value2.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data2, "value2.2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data3, "value3.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data3, "value3.2"),SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_data3, "value3.3"), SSLOG_ERROR_NO);

    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object1, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object2, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object2, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object3, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object3, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual1, property_object3, individual3), SSLOG_ERROR_NO);

    // Properties to individual2.
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data1, "value1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data2, "value2.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data2, "value2.2"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data3, "value3.1"), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data3, "value3.2"),SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_data3, "value3.3"), SSLOG_ERROR_NO);

    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object1, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object2, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object2, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object3, individual1), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object3, individual2), SSLOG_ERROR_NO);
    CU_ASSERT_EQUAL_FATAL(sslog_insert_property(individual2, property_object3, individual3), SSLOG_ERROR_NO);


    CU_ASSERT_EQUAL_FATAL(sslog_remove_individual(individual1), SSLOG_ERROR_NO);

    sslog_triple_t *triple_template = sslog_new_triple_detached("Ind1",
                                                                SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    list_t *triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);
    sslog_free_triples(triples);
    sslog_free_triple(triple_template);

    triple_template = sslog_new_triple_detached(sslog_entity_get_uri(individual2),
                                                                    SSLOG_TRIPLE_ANY, SSLOG_TRIPLE_ANY,
                                                                    SSLOG_RDF_TYPE_URI, SSLOG_RDF_TYPE_URI);

    triples = sslog_query_triple(triple_template);

    // 9 properties (3 properties are removed, links to individual1) + 1 rdf:type
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 10);
    sslog_free_triples(triples);

    CU_ASSERT_EQUAL_FATAL(sslog_remove_individual(individual2), SSLOG_ERROR_NO);

    triples = sslog_query_triple(triple_template);
    CU_ASSERT_EQUAL_FATAL(list_count(triples), 0);
    sslog_free_triples(triples);
    sslog_free_triple(triple_template);

    clean_local_store();
}



/* The main() function for setting up and running the tests.
 * Returns a CUE_SUCCESS on successful running, another
 * CUnit error code on failure.
 */
int main()
{
    CU_pSuite pSuite_node_sparql = NULL;

    /* initialize the CUnit test registry */
    if (CUE_SUCCESS != CU_initialize_registry())
        return CU_get_error();


    /* add a suite to the registry */
    pSuite_node_sparql = CU_add_suite("Suite: Node - SPARQL queries", init_suite_nodes, clean_suite_nodes);
    if (NULL == pSuite_node_sparql) {
        CU_cleanup_registry();
        return CU_get_error();
    }

     /* add the tests to the suite */
    if ((NULL == CU_add_test(pSuite_node_sparql, "Node - SPARQL ASK", test_node_sparql_ask))
            || (NULL == CU_add_test(pSuite_node_sparql, "Node - SPARQL CONSTRUCT", test_node_sparql_construct))
            || (NULL == CU_add_test(pSuite_node_sparql, "Node - SPARQL SELECT", test_node_sparql_select))
            || (NULL == CU_add_test(pSuite_node_sparql, "Node - SPARQL-endpoint SELECT", test_node_sparql_enpoint_select))
            ) {
        CU_cleanup_registry();
        return CU_get_error();
    }

    /* Run all tests using the CUnit Basic interface */
    CU_basic_set_mode(CU_BRM_VERBOSE);
    CU_basic_run_tests();

    int failed_tests = CU_get_number_of_tests_failed();

    CU_cleanup_registry();

    return failed_tests;
}
