﻿//  ------------------------------------------------------------------------
//  <copyright file="GeneralProperty.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The class contains a set of <see cref="Property"/> (slave-properties) and 
    /// <see cref="Handler"/>.
    /// Each operation under this class is translated to slave-properties using <see cref="Handler"/>.
    /// The <see cref="Handler"/> is used to configure rules to perform operation.
    /// This class doesn't perform real actions ander properties, it only prepares a <see cref="HandlerConfig"/>. 
    /// The config stores rules to perform real actions under properties.
    /// </summary>
    /// <seealso cref="GeneralProperty"/>
    /// <seealso cref="Handler"/>
    /// <seealso cref="IHandlerConfig"/>
    public class GeneralProperty
    {
        /// <summary>
        /// The properties for operations (slave-properties).
        /// </summary>
        private List<Property> slaveProperties = new List<Property>();

        /// <summary>
        /// Configurators to build rules for perfoming operations.
        /// </summary>
        private List<Handler> handlers = new List<Handler>();

        /// <summary>
        /// Initializes a new instance of the GeneralProperty with default configurator.
        /// <remarks>
        /// Operations under GeneralProperty will be translated to all slave-properies using <see cref="DefaultHandler"/>. 
        /// </remarks>
        /// </summary>
        /// <param name="properties">Slave-properties.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when a properties list or one of the property is null.</exception>
        /// <seealso cref="DefaultHandler"/>
        public GeneralProperty(IEnumerable<Property> properties)
        {
            if (properties == null)
            {
                throw new ArgumentNullException("properties");
            }

            foreach (Property p in properties)
            {
                if (p == null)
                {
                    throw new ArgumentNullException("properties", "Properties list contains a null property.");
                }
            }

            slaveProperties = new List<Property>(properties);
            handlers.Add(new DefaultHandler());
        }

        /// <summary>
        /// Initializes a new instance of the GeneralProperty with given configurator.
        /// </summary>
        /// <param name="properties">Slave-properties.</param>
        /// <param name="configurator">Configurator to build rules.</param>
        public GeneralProperty(IEnumerable<Property> properties, Handler configurator)
        {
            if (properties == null)
            {
                throw new ArgumentNullException("properties");
            }

            slaveProperties = new List<Property>(properties);
            handlers.Add(configurator);
        }

        /// <summary>
        /// Initializes a new instance of the GeneralProperty with configurators.
        /// </summary>
        /// <param name="properties"></param>
        /// <param name="configurators">Configurators to build rules.</param>
        public GeneralProperty(IEnumerable<Property> properties, IEnumerable<Handler> configurators)
        {
            if (properties == null)
            {
                throw new ArgumentNullException("properties");
            }

            slaveProperties = new List<Property>(properties);
            handlers = new List<Handler>(configurators);
        }

        /// <summary>
        /// Prepare configuration to perform action under GeneralProperty. 
        /// </summary>
        /// <param name="action">Action under GeneralProperty.</param>
        /// <param name="individual">Individual to perform an action.</param>
        /// <param name="data">Data for action.</param>
        /// <returns></returns>
        internal HandlerConfig PrepareConfig(ActionType action, Individual individual, object data)
        {
            HandlerConfig config = null;

            if (action == ActionType.Insert)
            {
                config = new HandlerConfig(individual, slaveProperties, data, null);
            }
            else if (action == ActionType.Remove)
            {
                config = new HandlerConfig(individual, slaveProperties, null, data);
            }
           
            foreach (Handler h in handlers)
            {
                if (h.CanConfig(config))
                {
                    h.ConfigureRules(config, action);
                }
            }

            return config;
        }

        /// <summary>
        /// Prepare configuration to perform action under GeneralProperty. 
        /// </summary>
        /// <param name="action">Action under GeneralProperty.</param>
        /// <param name="individual">Individual to perform an action.</param>
        /// <param name="data">Data for action.</param>
        /// <returns></returns>
        internal HandlerConfig PrepareConfig(ActionType action, Individual individual, object oldData, object newData)
        {
            HandlerConfig config = new HandlerConfig(individual, slaveProperties, newData, oldData);

            foreach (Handler h in handlers)
            {
                if (h.CanConfig(config))
                {
                    h.ConfigureRules(config, action);
                }
            }

            return config;
        }


    }
}
