﻿//  ------------------------------------------------------------------------
//  <copyright file="Handler.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------
namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The class to configure actions for a <see cref="GeneralProperty"/>. 
    /// The <see cref="GeneralProperty"/> can contains different configurators.
    /// Each configurator makes rules using <see cref="IHandlerConfig"/>, 
    /// based on these rules the <see cref="GeneralProperty"/> 
    /// performs actions from the <see cref="IHandlerConfig"/>.
    /// </summary>
    /// <seealso cref="GeneralProperty"/>
    /// <seealso cref="IHandlerConfig"/>
    public abstract class Handler
    {
        /// <summary>
        /// Configures rules based on action.
        /// </summary>
        /// <param name="config">Config for rules.</param>
        /// <param name="action">Action that was made under <see cref="GeneralProperty"/></param>
        /// <seealso cref="IHandlerConfig"/>
        internal void ConfigureRules(IHandlerConfig config, ActionType action)
        {
            switch (action)
            {
                case ActionType.Insert:
                    this.ConfigureInsert(config);
                    break;
                case ActionType.Remove:
                    this.ConfigureRemove(config);
                    break;
                case ActionType.Update:
                    this.ConfigureUpdate(config);
                    break;
            }
        }

        /// <summary>
        /// Checks what this configurator can be used to build rules.
        /// </summary>
        /// <param name="config">Config for rules.</param>
        /// <returns>true if the configurator can be used.</returns>
        /// <seealso cref="IHandlerConfig"/>
        protected internal virtual bool CanConfig(IHandlerConfig config)
        {
            return true;
        }

        /// <summary>
        /// Configures rules for insert actions.
        /// </summary>
        /// <param name="config">Config for rules.</param>
        protected virtual void ConfigureInsert(IHandlerConfig config) 
        {
        }

        /// <summary>
        /// Configures rules for remove actions.
        /// </summary>
        /// <param name="config">Config for rules.</param>
        protected virtual void ConfigureRemove(IHandlerConfig config) 
        {
        }

        /// <summary>
        /// Configures rules for update actions.
        /// </summary>
        /// <param name="config">Config for rules.</param>
        protected virtual void ConfigureUpdate(IHandlerConfig config) 
        {
        }
    }
}
