﻿//  ------------------------------------------------------------------------
//  <copyright file="IndicationChangesData.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using PetrSU.SmartSlog.NativeStructures;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// Class describes changes that were made after subscription indication. 
    /// </summary>
    internal class IndicationChangesData : ISubscriptionChangesData
    {
        /// <summary>
        /// Indication sequence number.
        /// First changes are made after subscription operation and this changes has 0 as sequence number.
        /// </summary>
        private int sequenceNumber;

        /// <summary>
        /// List with inserted to the smart space individuals.  
        /// </summary>
        private List<Individual> insertedIndividuals = new List<Individual>();

        /// <summary>
        /// List with removed from smart space individuals uri.
        /// </summary>
        private List<string> removedIndividuals = new List<string>();

        /// <summary>
        /// List with individuals that have properties that were changed.
        /// </summary>
        private List<Individual> updatedIndividuals = new List<Individual>();

        /// <summary>
        /// List with inserted to the smart space triples.  
        /// </summary>
        private List<ITriple> insertedTriples = new List<ITriple>();

        /// <summary>
        /// List with removed from smart space triples.
        /// </summary>
        private List<ITriple> removedTriples = new List<ITriple>();

        /// <summary>
        /// List with triples that have properties that were changed.
        /// </summary>
        private List<ITriple> updatedTriples = new List<ITriple>();

        /// <summary>
        /// Sparql result with new data that was changed.
        /// </summary>
        private SparqlResult sparqlNewResult;

        /// <summary>
        /// Sparql result with old data that was changed.
        /// </summary>
        private SparqlResult sparqlOldResult;

        /// <summary>
        /// Sparql result with new data that was changed.
        /// </summary>
        public SparqlResult SparqlNewResult { get { return sparqlNewResult; } }

        /// <summary>
        /// Sparql result with old data that was changed.
        /// </summary>
        public SparqlResult SparqlOldResult { get { return sparqlOldResult; } }

        /// <summary>
        /// The correspondence between the altered properties and individuals.
        /// </summary>
        private Dictionary<Individual, List<PropertyChangesData>> changedProperties = new Dictionary<Individual, List<PropertyChangesData>>();

        /// <summary>
        /// This list is used to return empty list in some methods instead return null.
        /// </summary>
        private static ReadOnlyCollection<PropertyChangesData> emptyPropetiesList = new ReadOnlyCollection<PropertyChangesData>(new List<PropertyChangesData>());
        private IntPtr sslog_changes;

        /// <summary>
        /// Gets or sets sequence number of this indication. Firts changes are made after subscription operation 
        /// and this changes always has 0 as sequence number.
        /// </summary>
        public int SequenceNumber
        {
            get { return this.sslog_changes != null ? NativeMethods.sslog_sbcr_ch_get_sequence_number(this.sslog_changes) : sequenceNumber; }
            internal set { sequenceNumber = value; }
        }

        /// <summary>
        /// Initialize triples structures
        /// </summary>
        /// <param name="changes">Pointer on change_list structure</param>
        public IndicationChangesData()
        {

        }

        /// <summary>
        /// Initialize triples structures
        /// </summary>
        /// <param name="changes">Pointer on change_list structure</param>
        public IndicationChangesData(IntPtr changes)
        {
            sslog_changes = changes;
            /// DEBUG
            try
            {
                IntPtr list_t = NativeMethods.sslog_sbcr_ch_get_triples(changes, (int)sslog_action_type.INSERT);
                insertedTriples = new List<ITriple>(NativeMethods.ToManagedList(list_t));

                list_t = NativeMethods.sslog_sbcr_ch_get_triples(changes, (int)sslog_action_type.REMOVE);
                removedTriples = new List<ITriple>(NativeMethods.ToManagedList(list_t));

                list_t = NativeMethods.sslog_sbcr_ch_get_triples(changes, (int)sslog_action_type.UPDATE);
                updatedTriples = new List<ITriple>(NativeMethods.ToManagedList(list_t));
            }
            catch
            {

            }


            IntPtr new_result;

            try
            {
                new_result = NativeMethods.sslog_sbcr_ch_get_new_sparql_select(changes);
            }
            catch { new_result = IntPtr.Zero; }

            if (!new_result.Equals(IntPtr.Zero))
                sparqlNewResult = new SparqlResult(new_result);

            IntPtr old_result;

            try
            {
                old_result = NativeMethods.sslog_sbcr_ch_get_old_sparql_select(changes);
            }
            catch { old_result = IntPtr.Zero; }

            if (!old_result.Equals(IntPtr.Zero))
                sparqlOldResult = new SparqlResult(old_result);



            try
            {
                Node node = (SessionManager.GetDefaultSession() as Session).nodes.FirstOrDefault();
                Collection<Individual> indList;

                IntPtr list_t = NativeMethods.sslog_sbcr_ch_get_individual_by_action(changes, (int)sslog_action_type.INSERT);
                List<string> individualsUri = MarshalHelp.UriListToManagedList(list_t);

                indList = node.GetIndividuals(individualsUri);
                if (indList.Count != 0)
                    insertedIndividuals = indList.ToList();

                list_t = NativeMethods.sslog_sbcr_ch_get_individual_by_action(changes, (int)sslog_action_type.REMOVE);
                removedIndividuals = MarshalHelp.UriListToManagedList(list_t);
                
                list_t = NativeMethods.sslog_sbcr_ch_get_individual_by_action(changes, (int)sslog_action_type.UPDATE);
                individualsUri = MarshalHelp.UriListToManagedList(list_t);

                indList = node.GetIndividuals(individualsUri);
                if (indList.Count != 0)
                    updatedIndividuals = indList.ToList();
            }
            catch
            {

            }
        }

        /// <summary>
        /// Adds individuals to the changes. 
        /// </summary>
        /// <param name="individual">Individual that has been chaged.</param>
        /// <param name="action">Changes action: inserted, removed, updated.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.ArgumentNullException">Thrown when individual is null.</exception>
        internal void AddChange(Individual individual, ActionType action)
        {
            if (individual == null)
            {
                throw new ArgumentNullException("individual");
            }

            List<Individual> individuals = GetRawIndividuals(action);

            if (individuals.Contains(individual) == false)
            {
                individuals.Add(individual);
            }
        }

        /// <summary>
        /// Adds property to the changes. Also individual will be added as updated to the changes.
        /// </summary>
        /// <param name="individual">Individual that has this property.</param>
        /// <param name="property">Property that was changed.</param>
        /// <param name="newValue">New value of the property, if the property was removed then pass null as value.</param>
        /// <param name="oldValue">Old value of the property, if the property was inserted (set), then pass null as value.</param>
        /// <param name="action">Action that was perfom under property. </param>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.Property" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.ArgumentNullException">Thrown when the individual or the property is null.</exception>
        internal void AddChange(Individual individual, Property property, object newValue, object oldValue, ActionType action)
        {
            if (individual == null)
            {
                throw new ArgumentNullException("individual");
            }

            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            PropertyChangesData pcd = new PropertyChangesData(individual, property, newValue, oldValue, action);

            // Add the individuals as updated individual if there is no such individual in the changes.
            if (updatedIndividuals.Contains(individual) == false)
            {
                updatedIndividuals.Add(individual);
            }

            List<PropertyChangesData> properties = null;

            // Create new of find old list with chnaged properties for the individual.
            if (changedProperties.ContainsKey(individual) == false)
            {
                properties = new List<PropertyChangesData>();
                changedProperties.Add(individual, properties);
            }
            else
            {
                properties = changedProperties[individual];
            }

            properties.Add(pcd);
        }

        /// <summary>
        /// Gets list with individuals by action. Individuals can be inserted, removed or updated. 
        /// </summary>
        /// <param name="action">Action that were perfom under individuals.</param>
        /// <returns>List with individuals.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.InvalidOperationException">Thrown when there is no a list for the given action.</exception>
        public ReadOnlyCollection<Individual> GetIndividuals(ActionType action)
        {
            List<Individual> individuals = null;

            switch (action)
            {
                case ActionType.Insert:
                    individuals = insertedIndividuals;
                    break;

                //case ActionType.Remove:
                //    individuals = removedIndividuals;
                //    break;

                case ActionType.Update:
                    individuals = updatedIndividuals;
                    break;

                default:
                    throw new InvalidOperationException("Can't perfome '" + action + "' action.");
            }

            return new ReadOnlyCollection<Individual>(individuals);
        }

        /// <summary>
        /// Gets list with removed individuals uri.
        /// </summary>
        /// <returns>List with individuals uri.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.InvalidOperationException">Thrown when there is no a list for the given action.</exception>
        public ReadOnlyCollection<string> GetRemovedIndividualsUri()
        {
            return new ReadOnlyCollection<string>(removedIndividuals);
        }

        /// <summary>
        /// Gets list with all individuals that were added to the chnanges (without removed). The list contains 
        /// inserted or updated individuals. 
        /// </summary>
        /// <returns>List with all individuals.</returns>
        public ReadOnlyCollection<Individual> GetIndividuals()
        {
            List<Individual> individuals = new List<Individual>(insertedIndividuals);
            individuals.AddRange(updatedIndividuals);
            //individuals.AddRange(removedIndividuals);

            return new ReadOnlyCollection<Individual>(individuals);
        }

        /// <summary>
        /// Gets list with triples by action. Triples can be inserted, removed or updated. 
        /// </summary>
        /// <param name="action">Action that were perfom under triples.</param>
        /// <returns>List with triples.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Triple" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.InvalidOperationException">Thrown when there is no a list for the given action.</exception>
        public ReadOnlyCollection<ITriple> GetTriples(ActionType action)
        {
            List<ITriple> triples = null;

            switch (action)
            {
                case ActionType.Insert:
                    triples = insertedTriples;
                    break;

                case ActionType.Remove:
                    triples = removedTriples;
                    break;

                case ActionType.Update:
                    triples = updatedTriples;
                    break;

                default:
                    throw new InvalidOperationException("Can't perfome '" + action + "' action.");
            }

            return new ReadOnlyCollection<ITriple>(triples);
        }

        /// <summary>
        /// Gets list with all triples that were added to the changes. The list contains 
        /// inserted, removed or updated triples. 
        /// </summary>
        /// <returns>List with all triples.</returns>
        public ReadOnlyCollection<ITriple> GetTriples()
        {
            List<ITriple> triples = new List<ITriple>(insertedTriples);
            triples.AddRange(updatedTriples);
            triples.AddRange(removedTriples);

            return new ReadOnlyCollection<ITriple>(triples);
        }

        /// <summary>
        /// Gets properties chnaged data for the individual by action. 
        /// Properties can be inserted, removed or updated. 
        /// </summary>
        /// <param name="individual">Individual that was updated and has changed properties.</param>
        /// <param name="action">Action that was perfom under properties.</param>
        /// <returns>List with data about changed properties, or empty list if there are no data about changes.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <seealso cref="PetrSU.SmartSlog.PropertyChangesData" />
        public ReadOnlyCollection<PropertyChangesData> GetProperties(Individual individual, ActionType action)
        {
            List<PropertyChangesData> properties = null;

            if (changedProperties.TryGetValue(individual, out properties) == false)
            {
                return emptyPropetiesList;
            }

            List<PropertyChangesData> propertiesByAction = (from p in properties
                                                            where p.Action == action
                                                            select p).ToList<PropertyChangesData>();

            return new ReadOnlyCollection<PropertyChangesData>(propertiesByAction);
        }

        /// <summary>
        /// Gets all properties changed data for the individual. This list contains properties 
        /// that were inserted, removed or updated.
        /// </summary>
        /// <param name="individual">Individual that was updated and has changed properties.</param>
        /// <returns>List with data about changed properties, or empty list if there are no data about changes.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.PropertyChangesData" />
        public ReadOnlyCollection<PropertyChangesData> GetProperties(Individual individual)
        {
            List<PropertyChangesData> properties = null;

            if (changedProperties.TryGetValue(individual, out properties) == false)
            {
                return emptyPropetiesList;
            }

            return new ReadOnlyCollection<PropertyChangesData>(properties);
        }


        /// <summary>
        /// Gets list with individuals by action. Individuals can be inserted, removed or updated. 
        /// </summary>
        /// <param name="action">Action that were perfom under individuals.</param>
        /// <returns>List with individuals.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.InvalidOperationException">Thrown when there is no a list for the given action.</exception>
        private List<Individual> GetRawIndividuals(ActionType action)
        {
            List<Individual> individuals = null;

            switch (action)
            {
                case ActionType.Insert:
                    individuals = insertedIndividuals;
                    break;

                //case ActionType.Remove:
                //    individuals = removedIndividuals;
                //    break;

                case ActionType.Update:
                    individuals = updatedIndividuals;
                    break;

                default:
                    throw new InvalidOperationException("Can't perfome '" + action + "' action.");
            }

            return individuals;
        }
    }
}
