﻿//  ------------------------------------------------------------------------
//  <copyright file="Individual.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using PetrSU.SmartSlog.NativeStructures;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The <c>Individual</c> class represents ontology individual.
    /// <remarks>
    ///   <para>
    ///   Instances of this class can't be directly created from ecternal code. 
    ///   You need use <see cref="PetrSU.SmartSlog.Node"/> method to get instace of this class.
    ///   </para>
    ///   </remarks>
    ///   <seealso cref="PetrSU.SmartSlog.Node"/>
    /// </summary>
    public class Individual : OntologyEntity
    {
        #region Private fields
        /// <summary>
        /// UUID of individual.
        /// </summary>
        private string uuid;

        /// <summary>
        /// Parent(Base) class of individual.
        /// </summary>
        private OntologyClass parentClass;

        /// <summary>
        /// Properties of individual.
        /// </summary>
        private List<PropertyValue> properties = new List<PropertyValue>();

        //// <summary>
        //// Represents a lock that is used to manage access to a resource, 
        //// allowing multiple threads for reading or exclusive access for writing. 
        //// </summary>
        //// private ReaderWriterLockSlim propertiesAccessLock = new ReaderWriterLockSlim(LockRecursionPolicy.SupportsRecursion);

        /// <summary>
        /// For locking simultaneous access to individual's properties.
        /// </summary>
        private object propertiesAccessLock = new object();

        //// TODO: need to make a proxy class for semantic and other checking actions.
        /// <summary>
        /// Session where individual is registered.
        /// </summary>
        private Session session;

        private IntPtr nativePtr;

        internal IntPtr NativePtr { get { return nativePtr; } set { nativePtr = value; } }

        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the Individual class.
        /// </summary>
        /// <param name="ontologyClass">Base(parent) class of individual, parent class.</param>
        /// <exception cref="System.ArgumentNullException">
        /// if <paramref name="ontologyClass"/> is <c>null</c>.
        /// </exception>
        /// <see cref="PetrSU.SmartSlog.OntologyClass"/>
        internal Individual(OntologyClass ontologyClass, Session session)
        {
            if (ontologyClass == null)
            {
                throw new ArgumentNullException("ontologyClass");
            }

            this.parentClass = ontologyClass;

            if (String.IsNullOrEmpty(ontologyClass.Namespace))
            {
                this.uuid = System.Guid.NewGuid().ToString();
            }
            else
            {
                this.uuid = ontologyClass.Namespace + "#" + System.Guid.NewGuid().ToString();
            }

            SetUri(this.uuid);

            this.session = session;

            IntPtr individualPtr = NativeMethods.sslog_new_individual(this.parentClass.NativePtr, this.uuid);

            this.NativePtr = individualPtr;
        }

        /// <summary>
        /// Initializes a new instance of the Individual class.
        /// </summary>
        /// <param name="ontologyClass">Base(parent) class of individual, parent class.</param>
        /// <param name="uri">URI that is identically identifier individual.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when ontologyClass is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when uuid is null or empty.</exception>
        /// <see cref="PetrSU.SmartSlog.OntologyClass"/>
        internal Individual(OntologyClass ontologyClass, string uri, Session session)
        {
            if (ontologyClass == null)
            {
                throw new System.ArgumentNullException("ontologyClass");
            }

            if (String.IsNullOrEmpty(uri) == true)
            {
                throw new System.ArgumentException("Identifier can not be null or empty.", "uuid");
            }

            this.parentClass = ontologyClass;

            this.uuid = uri;

            SetUri(uri);

            this.session = session;
        }


        /// <summary>
        /// Initializes a new instance of the Individual class.
        /// </summary>
        /// <param name="ontologyClass">Base(parent) class of individual, parent class.</param>
        /// <param name="uri">URI that is identically identifier individual.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when ontologyClass is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when uuid is null or empty.</exception>
        /// <see cref="PetrSU.SmartSlog.OntologyClass"/>
        public Individual(OntologyClass ontologyClass, string uri)
        {
            if (ontologyClass == null)
            {
                throw new System.ArgumentNullException("ontologyClass");
            }

            if (String.IsNullOrEmpty(uri) == true)
            {
                throw new System.ArgumentException("Identifier can not be null or empty.", "uuid");
            }

            this.parentClass = ontologyClass;

            this.uuid = uri;

            IntPtr individualPtr = NativeMethods.sslog_new_individual(this.parentClass.NativePtr, this.uuid);

            this.NativePtr = individualPtr;
        }

        /// <summary>
        /// Initializes a new instance of the Individual class.
        /// </summary>
        /// <param name="ontologyClass">Base(parent) class of individual, parent class.</param>
        /// <param name="uri">URI that is identically identifier individual.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when ontologyClass is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when uuid is null or empty.</exception>
        /// <see cref="PetrSU.SmartSlog.OntologyClass"/>
        internal Individual(string uri)
        {
            if (String.IsNullOrEmpty(uri) == true)
            {
                throw new System.ArgumentException("Identifier can not be null or empty.", "uuid");
            }

            this.uuid = uri;
        }

        internal Individual(IntPtr nativeIndividual, Session session = null)
        {
            if (nativeIndividual.ToInt32().Equals(0))
                return;

            this.nativePtr = nativeIndividual;

            sslog_individual e = (sslog_individual)Marshal.PtrToStructure(nativePtr, typeof(sslog_individual));

            this.uuid = e.uri;

            if (session != null)
            { 
                this.session = session;
                this.parentClass = this.session.storedClasses.Find(x => x.Uri == e.class_uri);
            }
            else
            {
                this.session = SessionManager.GetDefaultSession() as Session;
                this.parentClass = this.session.storedClasses.Find(x => x.Uri == e.class_uri);
            }
        }

        internal static Individual Create(IntPtr nativeIndividual, Session session = null)
        {
            return new Individual(nativeIndividual, session);
        }

        internal static Individual Create(string uri, Session session = null)
        {
            //IntPtr nativeIndividual = NativeMethods.sslog_get_individual(uri);
            return new Individual(uri);
        }

        #endregion

        #region Enums
        /// <summary>
        /// Cardinality types.
        /// </summary>
        private enum Cardinality
        {
            /// <summary>
            /// Maximum cardinality.
            /// </summary>
            Max,

            /// <summary>
            /// Minimum cardinality.
            /// </summary>
            Min
        }
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the parent class of individual.
        /// </summary>
        public OntologyClass ParentClass
        {
            get { return this.parentClass; }
        }

        /// <summary>
        /// Gets individual UUID.
        /// </summary>
        private string Uuid
        {
            get { return this.uuid; }
        }
        #endregion

        #region Public methods
        // TODO: Check cardinality

        /// <summary>
        /// Sets a new data property. This function checks cardinality, you can't set more properties than maximum cardinality.
        /// If the individuals has a given property with given value, then method does nothing.
        /// </summary>
        /// <param name="property">Property of the value.</param>
        /// <param name="value">String that represents the value.</param>
        /// <exception cref="System.ArgumentException">Thrown when the property cannot be set for the individual 
        /// (not registered in the session or parent class has no the property).</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>"
        public void InsertProperty(Property property, string value)
        {
            //SemanticChecker.CheckPropertySet(session.Repository, this, property, value);

            try
            {
                lock (propertiesAccessLock)
                {
                    InsertProperty(property, (object)value);
                }
            }
            catch (SemanticException ex)
            {
                throw new ArgumentOutOfRangeException("value", ex.ToString());
            }
        }

        /// <summary>
        /// Sets object property. Object property is as links to another individual. 
        /// This function checks cardinality, you can't set more properties than maximum cardinality.
        /// If the individuals has a given property with given value, then method does nothing.
        /// </summary>
        /// <param name="property">Property of the value.</param>
        /// <param name="value">Object property value, <see cref="PetrSU.SmartSlog.Individual"/>.</param>
        /// <exception cref="System.ArgumentException">Thrown when the property cannot be set for the individual 
        /// (not registered in the session or parent class has no the property, or value is not registered in the session).</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public void InsertProperty(Property property, Individual value)
        {
            //SemanticChecker.CheckPropertySet(session.Repository, this, property, value);

            try
            {
                lock (propertiesAccessLock)
                {
                    InsertProperty(property, (object)value);
                }
            }
            catch (SemanticException ex)
            {
                throw new ArgumentOutOfRangeException("value", ex.ToString());
            }
        }

        /// <summary>
        /// Get one data-property or object-property value. 
        /// </summary>
        /// <param name="property">Property to get.</param>
        /// <returns>Firs founded value of the property or null if the property not found.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public object GetProperty(Property property)
        {
            if (this.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Individual has no pointer");
            }

            if (property == null)
            {
                return null;
            }

            IntPtr propertyValue = NativeMethods.sslog_get_property(this.NativePtr, property.NativePtr);

            if (propertyValue.ToInt32().Equals(0))
                return null;

            if (property.IsObject())
            {
                Individual ind_value = new Individual(propertyValue);
                return ind_value;
            }
            else
            {
                string str_value = MarshalHelp.PtrToString(propertyValue);
                return str_value;
            }

            //lock (propertiesAccessLock)
            //{
            //    foreach (PropertyValue p in this.properties)
            //    {
            //        if (p.Property.Equals(property))
            //        {
            //            return p.GetValue();
            //        }
            //    }
            //}

            //return null;
        }

        /// <summary>
        /// Get all values for the given property.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <returns>Values of the property.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public Collection<object> GetProperties(Property property)
        {
            if (property == null)
            {
                return new Collection<object>();
            }

            IntPtr propertyValue = NativeMethods.sslog_get_properties(this.nativePtr, property.NativePtr);

            if (propertyValue.ToInt32().Equals(0))
                return null;

            if (property.IsObject())
            {
                Collection<Individual> ind_value = NativeMethods.IndividualListToManagedList(propertyValue);
                /// TEST IT!!!
                return new Collection<object>(ind_value.Select(x => (object)x).ToList());
            }
            else
            {
                List<string> str_value = MarshalHelp.PtrToStringArray(propertyValue);
                /// TEST IT!!!
                return new Collection<object>(str_value.Select(x => (object)x).ToList());
            }
            //lock (propertiesAccessLock)
            //{
            //    return new Collection<object>((from p in this.properties
            //                                   where p.Property.Equals(property)
            //                                   select p.GetValue()).ToList<object>());
            //}
        }

        /// <summary>
        /// Get all object values by given property.
        /// </summary>
        /// <param name="property">Proprty to get values.</param>
        /// <returns>Values of the property.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public Collection<Individual> GetObjectProperties(Property property)
        {
            if (property == null)
            {
                return new Collection<Individual>();
            }

            if (property.IsObject() == false)
            {
                return new Collection<Individual>();
            }

            lock (propertiesAccessLock)
            {
                return new Collection<Individual>((from p in this.properties
                                                   where p.Property.Equals(property)
                                                   select p.GetObjectValue()).ToList<Individual>());
            }
        }

        /// <summary>
        /// Get all data values by given property.
        /// </summary>
        /// <param name="property">Proprty to get values.</param>
        /// <returns>Values of the property.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public Collection<string> GetDataProperties(Property property)
        {
            if (property == null)
            {
                return new Collection<string>();
            }

            if (property.IsObject() == true)
            {
                return new Collection<string>();
            }

            lock (propertiesAccessLock)
            {
                return new Collection<string>((
                    from p in this.properties
                    where p.Property.Equals(property)
                    select p.GetDataValue()).ToList<string>());
            }
        }

        /// <summary>
        /// Gets all values of all properties (object and data).
        /// </summary>
        /// <returns>Values of the property.</returns>
        public Collection<object> GetProperties()
        {
            lock (propertiesAccessLock)
            {
                return new Collection<object>((from p in this.properties
                                               select p.GetValue()).ToList<object>());

            }
        }

        /// <summary>
        /// Gets first founded data-property value.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <returns>Value of property or null if the value not found.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public string GetDataProperty(Property property)
        {
            if (property == null)
            {
                return null;
            }

            if (property.IsObject() == true)
            {
                return null;
            }

            return (string)this.GetProperty(property);
        }

        /// <summary>
        /// Gets first founded object-property value.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <returns>Value of property or null if the value not found.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public Individual GetObjectProperty(Property property)
        {
            if (property == null)
            {
                return null;
            }

            if (property.IsObject() == false)
            {
                return null;
            }

            return (Individual)this.GetProperty(property);
        }

        /// <summary>
        /// Update data-property value. It does nothing if old or new value is null.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <param name="oldValue">Old value of property.</param>
        /// <param name="newValue">New value for property.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when value (old or new) is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when the property cannot be set for the individual 
        /// (not registered in the session or parent class has no the property).</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void UpdateProperty(Property property, string oldValue, string newValue)
        {
            SemanticChecker.ChekPropertyUpdate(session.Repository, this, property, oldValue, newValue);

            lock (propertiesAccessLock)
            {
                UpdateProperty(property, (object)oldValue, (object)newValue);
            }
        }

        /// <summary>
        /// Update object-property value. It does nothing if old or new value is null.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <param name="oldValue">Old value of property.</param>
        /// <param name="newValue">New value of property.</param>
        /// <exception cref="System.ArgumentException">Thrown when the property cannot be set for the individual 
        /// (not registered in the session or parent class has no the property, or value (new or old) is not registered in the session).</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public void UpdateProperty(Property property, Individual oldValue, Individual newValue)
        {
            SemanticChecker.ChekPropertyUpdate(session.Repository, this, property, oldValue, newValue);

            lock (propertiesAccessLock)
            {
                UpdateProperty(property, (object)oldValue, (object)newValue);
            }
        }

        /// <summary>
        /// Remove property value.
        /// </summary>
        /// <remarks>Find property with given value and remove it. 
        /// This function removes only one, first found property.</remarks>
        /// <param name="property">Property of value.</param>
        /// <param name="value">Property value.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when property is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void RemoveProperty(Property property, string value)
        {
            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            if (value == null)
            {
                return;
            }

            try
            {
                lock (propertiesAccessLock)
                {
                    RemoveProperty(property, value as object);
                }
            }
            catch (SemanticException ex)
            {
                throw new ArgumentOutOfRangeException("value", ex.ToString());
            }
        }

        /// <summary>
        /// Removes property value.
        /// </summary>
        /// <remarks>Find property with given value and remove it. 
        /// This function removes only one, first found property.</remarks>
        /// <param name="property">Property of value.</param>
        /// <param name="value">Property value.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when property is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void RemoveProperty(Property property, Individual value)
        {
            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            if (value == null)
            {
                return;
            }

            try
            {
                lock (propertiesAccessLock)
                {
                    RemoveProperty(property, value as object);
                }
            }
            catch (SemanticException ex)
            {
                throw new ArgumentOutOfRangeException("value", ex.ToString());
            }
        }

        /// <summary>
        /// Removes one first founded property.
        /// </summary>
        /// <param name="property">Property of value.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when property is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        internal void RemoveProperty(Property property)
        {
            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            

            //lock (propertiesAccessLock)
            //{
            //    if (IsCardinalityViolation(property, this.properties, Cardinality.Min) == true)
            //    {
            //        throw new ArgumentOutOfRangeException("property", ExceptionManager.CardinalityMinimumMessage);
            //    }

            //    PropertyValue value = this.properties.FirstOrDefault(pv => pv.Property == property);

            //    this.properties.Remove(value);
            //}
        }


        /// <summary>
        /// Remove all properties values by given property.
        /// </summary>
        /// <param name="property">Property of values.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when property is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        public void RemoveProperties(Property property)
        {
            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            if (property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            int result = NativeMethods.sslog_remove_properties(this.NativePtr, property.NativePtr);
           
            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            return;
        }

        #endregion

        #region Internal methods

        /// <summary>
        /// Return all property values (object and data).
        /// </summary>
        /// <returns>List of values.</returns>
        internal List<PropertyValue> GetPropertiesValues()
        {
            return properties;
        }

        /// <summary>
        /// Tries to find the property with the given value. It checks object and data properties.
        /// </summary>
        /// <param name="property">Property to check.</param>
        /// <param name="value">Value of the property.</param>
        /// <returns>True if there is property with given value or false otherwise.</returns>
        internal bool IsPropertySet(Property property, object value)
        {
            if (GetPropertieValueWithData(property, value) != null)
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Finds property value structure with given data. It finds object and data properties.
        /// </summary>
        /// <param name="property">Property to find.</param>
        /// <param name="data">Data to find.</param>
        /// <returns><see cref="PetrSU.SmartSlog.PropertyValue"/> class with given property and value or null otherwise.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        internal PropertyValue GetPropertieValueWithData(Property property, object data)
        {
            PropertyValue value = this.properties.FirstOrDefault(pv => pv.Property == property && pv.GetValue().Equals(data));

            return value;
        }

        /// <summary>
        /// Sets the property. It works with object and data properties. Without blocking.
        /// </summary>
        /// <param name="property">Property to set.</param>
        /// <param name="value">The property value: <see cref="PetrSU.SmartSlog.Individual"/> 
        /// for the object property or string for the data property.</param>
        /// <returns>True if property is set or null otherwise.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        internal bool InsertProperty(Property property, object value)
        {
            int result;

            if (property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            if (property.IsObject())
                result = NativeMethods.sslog_insert_property(this.NativePtr, property.NativePtr, (value as Individual).NativePtr);
            else
                result = NativeMethods.sslog_insert_property(this.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(value as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            //if (IsPropertySet(property, (object)value) == true)
            //{
            //    return true;
            //}

            //if (IsCardinalityViolation(property, this.properties, Cardinality.Max) == true)
            //{
            //    throw new SemanticException("Maximum cardinality violation for property: " + property.ToString());
            //}

            //if (property.IsObject())
            //{
            //    this.properties.Add(new PropertyValue(property, value as Individual));
            //}
            //else
            //{
            //    this.properties.Add(new PropertyValue(property, value as string));
            //}

            return true;
        }

        //// TODO: need checks value

        /// <summary>
        /// Removes data or object property with given value. Without blocking.
        /// </summary>
        /// <param name="property">Property to remove.</param>
        /// <param name="value">Value of the property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        internal void RemoveProperty(Property property, object value)
        {
            int result;

            if (property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            if (property.IsObject())
                result = NativeMethods.sslog_remove_property(this.NativePtr, property.NativePtr, (value as Individual).NativePtr);
            else
                result = NativeMethods.sslog_remove_property(this.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(value as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            return;
            //if (IsCardinalityViolation(property, this.properties, Cardinality.Min) == true)
            //{
            //    Debug.WriteLine(
            //        "Min cardinality check is fail for property '" + property.Name + "'",
            //        "RemoveProperty");

            //    return;
            //}

            //var values = from p in this.properties
            //             where p.Property == property
            //                && p.GetValue().Equals(value)
            //             select p;

            //if (values.Count() == 0)
            //{
            //    return;
            //}

            //this.properties.Remove(values.First());
        }

        //// TODO: make private, subscriber need to call some of public method

        /// <summary>
        /// Update data or object property property with given values.
        /// It finds property with old value and update it with new value. If there is no
        /// properties with old value it sets new property. 
        /// </summary>
        /// <param name="property">Property to update.</param>
        /// <param name="oldValue">Old value of the property.</param>
        /// <param name="newValue">New value of the property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        internal void UpdateProperty(Property property, object oldValue, object newValue)
        {
            int result;

            if (property.NativePtr.ToInt32().Equals(0))
            {
                

                throw new ArgumentException("Not registered property");
            }

            if (property.IsObject())
                result = NativeMethods.sslog_update_property(this.NativePtr, property.NativePtr, (oldValue as Individual).NativePtr, (newValue as Individual).NativePtr);
            else
                result = NativeMethods.sslog_update_property(this.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(oldValue as string), MarshalHelp.StringToPtr(newValue as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }
            //PropertyValue value = GetPropertieValueWithData(property, oldValue);

            //if (value == null)
            //{
            //    InsertProperty(property, newValue);
            //    return;
            //}

            //if (IsPropertySet(property, newValue) == true)
            //{
            //    RemoveProperty(property, oldValue);
            //}
            //else
            //{
            //    value.UpdateValue(newValue);
            //}
        }

        ///// TODO: Subscride, need to remove

        /// <summary>
        /// Function for the <see cref="PetrSU.SmartSlog.Subscriber"/> class, 
        /// this function can set subscribed properties. (BAD!!! Later will be removed)
        /// </summary>
        /// <param name="property">Property to set.</param>
        /// <param name="value">Value of the property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Subscriber"/>
        internal void InsertPropertyS(Property property, object value)
        {
            if (property == null || value == null)
            {
                return;
            }

            lock (propertiesAccessLock)
            {
                if (IsPropertySet(property, value) == true)
                {
                    return;
                }

                if (IsCardinalityViolation(property, this.properties, Cardinality.Max) == true)
                {
                    Debug.WriteLine(
                        "Max cardinality check is fail for property '" + property.Name + "'",
                        "SetProperty");
                    return;
                }

                if (property.IsObject() == true)
                {
                    this.properties.Add(new PropertyValue(property, (Individual)value));
                }
                else
                {
                    this.properties.Add(new PropertyValue(property, (string)value));
                }
            }
        }

        //// TODO: need checks value

        /// <summary>
        /// Function for the <see cref="PetrSU.SmartSlog.Subscriber"/> class, 
        /// this function can remove subscribed properties. (BAD!!! Later will be removed)
        /// </summary>
        /// <param name="property">Property to remove.</param>
        /// <param name="value">Value of the property.</param>
        internal void RemovePropertyS(Property property, object value)
        {
            if (property == null || value == null)
            {
                return;
            }

            lock (propertiesAccessLock)
            {
                if (IsCardinalityViolation(property, this.properties, Cardinality.Min) == true)
                {
                    Debug.WriteLine(
                        "Min cardinality check is fail for property '" + property.Name + "'",
                        "RemoveProperty");

                    return;
                }

                var values = from p in this.properties
                             where p.Property == property
                                && p.GetValue().Equals(value)
                             select p;

                if (values.Count() == 0)
                {
                    return;
                }

                this.properties.Remove(values.First());
            }
        }

        internal void ResetSession()
        {
            this.session = null;
        }

        /// <summary>
        /// Gets rdf:type triple of individual. 
        /// Individual is described with rdf:type triple (URI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - classURI)).
        /// </summary>
        /// <returns>Triple on success or NULL otherwise (individual is NULL).</returns>
        public Triple toTriple()
        {
            return toTriple(this);
        }
        
         /// <summary>
        /// Gets rdf:type triple of individual. 
        /// 
        /// Individual is described with rdf:type triple (URI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - classURI)).
        /// </summary>
        /// <param name="individual">Individual to convert</param>
        /// <returns>Triple on success or NULL otherwise (individual is NULL).</returns>
        static internal Triple toTriple(Individual individual)
        {
            if (individual == null)
                throw new ArgumentNullException("individual");

            IntPtr triplePtr = NativeMethods.sslog_individual_to_triple(individual.nativePtr);

            if (triplePtr.ToInt32().Equals(0))
                return null;
            else
                return new Triple(triplePtr);
        }

        #endregion

        #region Private methods
        /// <summary>
        /// Checks cadinality violation.
        /// </summary>
        /// <remarks>Count properties values and return result of check. Checks maximum and minimum cardinality.</remarks>
        /// <param name="property">Property of values.</param>
        /// <param name="properties">List with properties values.</param>
        /// <param name="cardinality"><see cref="PetrSU.SmartSlog.Individual.Cardinality"/> type.</param>
        /// <returns><see cref="System.Boolean"/>: true if cardinality violated; otherwise, false.</returns>
        private static bool IsCardinalityViolation(Property property, List<PropertyValue> properties, Cardinality cardinality)
        {
            if (property.MaxCardinality < 0)
            {
                return false;
            }

            int count = (from p in properties where p.Property == property select p).Count();

            if (cardinality == Cardinality.Max && count >= property.MaxCardinality)
            {
                return true;
            }
            else if (cardinality == Cardinality.Min && count <= property.MinCardinality)
            {
                return true;
            }

            return false;
        }
        #endregion
    }
}

