﻿//  ------------------------------------------------------------------------
//  <copyright file="LoggerManager.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// Internal logger helper.
    /// </summary>
    internal static class LoggerManager
    {
        /// <summary>
        /// Type operations to log: join, leave operations.
        /// </summary>
        internal const int SessionManagement = 1;

        /// <summary>
        /// Type operations to log: properties.
        /// </summary>
        internal const int PropertiesManagement = 1;

        /// <summary>
        /// Type operations to log: operations with triples.
        /// </summary>
        internal const int TriplesManagement = 2;

        /// <summary>
        /// Type operations to log: subscription.
        /// </summary>
        internal const int SubscripitonManagement = 3;

        /// <summary>
        /// Type operations to log: operations with high-level entites.
        /// </summary>
        internal const int EntitiesManagement = 4;

        /// <summary>
        /// Type operations to log: operations with mapping triples to entities and vice versa.
        /// </summary>
        internal const int OwlMappingManagement = 5;

        /// <summary>
        /// Type operations to log: nodes general operations.
        /// </summary>
        internal const int NodeManagement = 6;

        
        /// <summary>
        /// Name of the programm to output.
        /// </summary>
        private const string ProgramName = "SmartSlog - ";

        /// <summary>
        /// Switch to control session operation. 
        /// </summary>
        private static readonly BooleanSwitch sessionSwitch = new BooleanSwitch(
            "SmartSlog.SessionManagementSwitch",
            "Switch on/off information about session management.",
            "0");

        /// <summary>
        /// Switch to controll operation with triples. 
        /// </summary>
        private static readonly BooleanSwitch triplesSwitch = new BooleanSwitch(
            "SmartSlog.TriplesManagementSwitch",
            "Switch on/off information about triples management.",
            "0");

        /// <summary>
        /// Switch to controll subscription. 
        /// </summary>
        private static readonly BooleanSwitch subscriptionSwitch = new BooleanSwitch(
            "SmartSlog.SubscriptionManagementSwitch",
            "Switch on/off information about subscription management.",
            "0");

        /// <summary>
        /// Switch to controll entities management. 
        /// </summary>
        private static readonly BooleanSwitch entitiesSwitch = new BooleanSwitch(
            "SmartSlog.EntitiesnManagementSwitch",
            "Switch on/off information about entities management.",
            "0");

        /// <summary>
        /// Switch to controll entities management. 
        /// </summary>
        private static readonly BooleanSwitch owlMappingSwitch = new BooleanSwitch(
            "SmartSlog.OwlMappingManagementSwitch",
            "Switch on/off information about owl mapping operations.",
            "0");

        /// <summary>
        /// Switch to controll entities management. 
        /// </summary>
        private static readonly BooleanSwitch nodeSwitch = new BooleanSwitch(
            "SmartSlog.NodeManagementSwitch",
            "Switch on/off information about node's operations.",
            "0");



        /// <summary>
        /// Prints triplets (subject, predicate, object) to the console. (Function for the debug).
        /// </summary>
        /// <param name="triples">List with triples</param>
        internal static void PrintTriples(ICollection<ITriple> triples)
        {
            foreach (ITriple triple in triples)
            {
                Console.WriteLine("\t - Subject: '" + triple.Subject + "'");
                Console.WriteLine("\t - Predicate: '" + triple.Predicate + "'");
                Console.WriteLine("\t - Object: '" + triple.ObjectElement + "'\n");
            }
        }

        /// <summary>
        /// Writes stating method message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        internal static void WriteStartMessage(int managementType, string methodName)
        {
            WriteStartMessage(managementType, methodName, String.Empty);
        }

        /// <summary>
        /// Writes stating method message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        /// <param name="message">Message to output.</param>
        internal static void WriteStartMessage(int managementType, string methodName, string message)
        {
            BooleanSwitch sw = GetSwitch(managementType);
            Trace.WriteLineIf(sw.Enabled, String.Empty);
            Trace.WriteLineIf(
                sw.Enabled,
                ProgramName + methodName + " " + message + " STARTS - " + DateTime.Now.ToString("hh:mm:ss.fff tt", CultureInfo.InvariantCulture));
        }

        /// <summary>
        /// Writes ending method message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        internal static void WriteEndMessage(int managementType, string methodName)
        {
            WriteEndMessage(managementType, methodName, String.Empty);
        }

        /// <summary>
        /// Writes ending method message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        /// <param name="message">Message to output.</param>
        internal static void WriteEndMessage(int managementType, string methodName, string message)
        {
            BooleanSwitch sw = GetSwitch(managementType);
            Trace.WriteLineIf(sw.Enabled, ProgramName + methodName + " " + message + " ENDS");
        }

        /// <summary>
        /// Writes information message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        /// <param name="message">Message to output.</param>
        internal static void WriteMessage(int managementType, string methodName, string message)
        {
            BooleanSwitch sw = GetSwitch(managementType);
            Trace.Indent();
            Trace.WriteLineIf(sw.Enabled, ProgramName + methodName + ": " + message);
            Trace.Unindent();
        }

        /// <summary>
        /// Writes error message.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        /// <param name="message">Message to output.</param>
        internal static void WriteErrorMessage(int managementType, string methodName, string message)
        {
            BooleanSwitch sw = GetSwitch(managementType);
            Trace.Indent();
            Trace.WriteLineIf(sw.Enabled, ProgramName + methodName + " ERROR: " + message);
            Trace.Unindent();
        }

       
        /// <summary>
        /// Writes error message based on execption, also writes infromation about inner exception.
        /// </summary>
        /// <param name="managementType">Type of the operation.</param>
        /// <param name="methodName">Method name to output.</param>
        /// <param name="ex">Exception to output.</param>
        internal static void WriteExceptionMessage(int managementType, string methodName, Exception ex)
        {
            BooleanSwitch sw = GetSwitch(managementType);
            Trace.Indent();
            Trace.WriteLineIf(sw.Enabled, ProgramName + methodName + " EXCEPTION: " + ex.ToString());
            
            if (ex.InnerException != null)
            {
                Trace.WriteLineIf(sw.Enabled && ex.InnerException != null, " INTERNAL EXCEPTION: " + ex.InnerException.ToString());
            }
            Trace.Unindent();
        }

        /// <summary>
        /// Gets a debug swith.
        /// </summary>
        /// <param name="managementType">Type of the management.</param>
        /// <returns>Swith of the given management type.</returns>
        private static BooleanSwitch GetSwitch(int managementType)
        {
            switch (managementType)
            {
                case SessionManagement: return sessionSwitch;
                case TriplesManagement: return triplesSwitch;
                case SubscripitonManagement: return subscriptionSwitch;
                case EntitiesManagement: return entitiesSwitch;
                case OwlMappingManagement: return owlMappingSwitch;
                case NodeManagement: return nodeSwitch;
                default:
                    {
                        Debug.Fail("You should not be here.");
                        return null;
                    }
            }
        }

        

    }
}
