﻿//  ------------------------------------------------------------------------
//  <copyright file="Node.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using PetrSU.SmartSlog.NativeStructures;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The <c>Node</c> class is intermediate between a SmartSpace and a developer.
    /// <remarks>
    /// All entities (<see cref="PetrSU.SmartSlog.OntologyClass"/>, <see cref="PetrSU.SmartSlog.Property"/>) must be registered by this class. 
    /// Individuals are created by this class, you can't create it manually.
    /// </remarks>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/>
    /// <seealso cref="PetrSU.SmartSlog.Property"/>
    /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
    /// </summary>
    public class Node
    {
        /// <summary>
        /// Node session. Node is registered in this session.
        /// </summary>
        private Session session;

        /// <summary>
        /// Identifier of the node, this identifier is user for communicating with the smart space.
        /// </summary>
        private string nodeId;

        private string spaceId;

        private string address;

        private int port;

        /// <summary>
        /// Identifier of the node, this identifier is user for communicating with the smart space.
        /// </summary>
        private string nodeSessionId;

        

        /// <summary>
        /// Link to the <see cref="PetrSU.SmartSlog.Repository"/>"/> class. 
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Repository"/>
        private IRepository repository = null;

        /// <summary>
        /// <see cref="PetrSU.SmartSlog.Subscriber"/> of the node, node use it to manage subscriptions.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Subscriber"/>
        private Subscriber subscriber;

        /// <summary>
        /// <see cref="PetrSU.SmartSlog.KpiWrapper"/> as a wrapper of the KPI libraray. 
        /// This wrapper is used to communicate with smart spaces through KPI.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.KpiWrapper"/>
        //private IKpi kpi;

        /// <summary>
        /// The flag: is node joined to the smart space or not.
        /// If it equals true node is joined, if it equals false the node is not joined.
        /// </summary>
        private bool isJoined;

        //static bool isSslogInit = false;

        private IntPtr sslog_node;

        internal IntPtr NativeNode { get { return sslog_node; } }

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the Node class with smart space parameters.
        /// </summary>
        /// <param name="nodeId">Identifier of node.</param>
        /// <param name="spaceId">Identifier of smart space.</param>
        /// <param name="address">Address of smart space (IP for TCP/IP).</param>
        /// <param name="port">Port of smart space.</param>
        public Node(string nodeId, string spaceId, string address, int port)
            : this(nodeId, SessionManager.GetDefaultSession() as Session)
        {
            //this.kpi.SetSpaceInfo(nodeId, spaceId, address, port);
            this.nodeId = nodeId;
            this.spaceId = spaceId;
            this.address = address;
            this.port = port;

            if (NativeMethods.sslog_init() != Constant.SSLOG_ERROR_NO)
            {
                throw new InvalidOperationException("Error " + NativeMethods.sslog_error_get_last_code() + ". "
                     + NativeMethods.sslog_error_get_last_text());
            }


            sslog_node = NativeMethods.sslog_new_node(nodeId, spaceId, address, port);

            if (sslog_node == IntPtr.Zero)
            {
                throw new InvalidOperationException("Error " + NativeMethods.sslog_error_get_last_code() + ". "
                     + NativeMethods.sslog_error_get_last_text());
            }
        }


        public void UpdateTriple(Triple oldTriple, Triple newTriple)
        {
            if (oldTriple == null && newTriple == null) {
                return;
            }

            if (oldTriple == null)
            {
                NativeMethods.sslog_node_update_triple(sslog_node, IntPtr.Zero, newTriple.NativePtr);
            }
            else if (newTriple == null)
            {
                NativeMethods.sslog_node_update_triple(sslog_node, oldTriple.NativePtr, IntPtr.Zero);
            }
            else
            {
                NativeMethods.sslog_node_update_triple(sslog_node, oldTriple.NativePtr, newTriple.NativePtr) ;
            }
        }

        public void RemoveTriple(Triple triple)
        {
            if (triple == null)
            {
                throw new ArgumentNullException("triple");
            }
            
            int result = NativeMethods.sslog_node_remove_triple(sslog_node, triple.NativePtr);
        }

        public Collection<ITriple> QueryTriple(Triple tripleTemplate)
        {
            if (tripleTemplate == null)
            {
                throw new ArgumentNullException("tripleTemplate");
            }

            if (tripleTemplate.NativePtr == null)
            {
                throw new ArgumentException("Native triple is null");
            }

            IntPtr list = NativeMethods.sslog_node_query_triple(sslog_node, tripleTemplate.NativePtr);

            return NativeMethods.ToManagedList(list);
        }

        /// <summary>
        /// Executes the SSAP format sparql ask query operation.
        /// </summary>
        /// <param name="query">SPARQL ASK query in text format</param>
        /// <returns>Status of the operation when completed (true if data exists, false when no data or otherwise).</returns>
        public bool SparqlAsk(string query)
        {
            if (query == null)
            {
                throw new ArgumentNullException("query");
            }

            int result;

            try
            {
                result = NativeMethods.sslog_node_sparql_ask(sslog_node, query);
            }
            catch
            {
                Debug.WriteLine(GetLastError());
                return false;
            }

            return result == 1 ? true : false;
        }

        /// <summary>
        /// Performs SPARQL SELECT query.
        /// 
        /// Returns all, or a subset of, the variables bound in a query pattern match. 
        /// Values  of variables are returned as a table there columns are variables and 
        /// rows contains values for variables.
        /// </summary>
        /// <param name="query">SPARQL query (ver. 2)</param>
        /// <param name="triples_template">Template to construct and store triples from 
        /// variables in a form '?a ?b ?c . ?d ?e ?f . ?q ?w ?e'. If the parameter is NULL then
        /// no triples will be constructed and stored in the local store.</param>
        /// <returns>result for SELECT query on success and NULL otherwise (on error)</returns>
        public SparqlResult SparqlSelect(string query, string triples_template = null)
        {
            if (query == null)
            {
                throw new ArgumentNullException("query");
            }

            //if (triples_template == null)
            //{
            //    throw new ArgumentNullException("triples_template");
            //}

            IntPtr list;

            try
            {
                list = NativeMethods.sslog_node_sparql_select(sslog_node, query, triples_template);
            }
            catch {
                Debug.WriteLine(GetLastError());
                return null;
            }

            if (list.Equals(IntPtr.Zero))
                return null;

            SparqlResult result = new SparqlResult(list);

            return result;
        }

        /// <summary>
        /// Performs SPARQL CONSTRUCT query.
        /// The CONSTRUCT query form returns a single RDF graph specified by a graph template.
        /// The result is an RDF graph formed by taking each query solution in the solution
        /// sequence, substituting for the variables in the graph template, and combining
        /// the triples into a single RDF graph by set union.
        /// </summary>
        /// <param name="query">String with CONSTRUCT query form</param>
        /// <returns>list with constructed triples on success, empty list if there are no data, or null in error.</returns>
        public List<ITriple> SparqlConstruct(string query)
        {
            if (query == null)
            {
                throw new ArgumentNullException("query");
            }

            IntPtr result;

            try
            {
                result = NativeMethods.sslog_node_sparql_construct(sslog_node, query);
            }
            catch
            {
                Debug.WriteLine(GetLastError());
                return null;
            }

            List<ITriple> triples = new List<ITriple>(NativeMethods.ToManagedList(result));

            return triples;
        }


        internal static string Utf16ToUtf8(string utf16String)
        {
            // Get UTF16 bytes and convert UTF16 bytes to UTF8 bytes
            byte[] utf16Bytes = Encoding.Unicode.GetBytes(utf16String);
            byte[] utf8Bytes = Encoding.Convert(Encoding.Unicode, Encoding.UTF8, utf16Bytes);

            // Return UTF8 bytes as ANSI string
            return Encoding.Default.GetString(utf8Bytes);
        }

        /// <summary>
        /// Initializes a new instance of the Node class with smart space parameters and session.
        /// </summary>
        /// <param name="session">Session for node.</param>
        /// <param name="nodeId">Identifier of node.</param>
        /// <param name="spaceId">Identifier of smart space.</param>
        /// <param name="address">Address of smart space (IP for TCP/IP).</param>
        /// <param name="port">Port of smart space.</param>
        public Node(ISession session, string nodeId, string spaceId, string address, int port)
            : this(nodeId, session as Session)
        {
           // this.kpi.SetSpaceInfo(nodeId, spaceId, address, port);
        }

        /// <summary>
        /// Initializes a new instance of the Node class for local work.
        /// </summary>
        /// <param name="nodeId">Identifier of node.</param>
        /// <param name="session">Session for node.</param>
        private Node(string nodeId, Session session)
        {
            if (String.IsNullOrEmpty(nodeId))
            {
                throw new ArgumentNullException("nodeId", "Node identifier cannot be null or empty.");
            }

            if (session == null)
            {
                throw new ArgumentNullException("session", "Session has an incorrect type or null.");
            }

            this.nodeId = nodeId;

            this.session = session;

           // repository = this.session.Repository;

            this.nodeSessionId = this.session.RegisterNode(this);

            //kpi = session.CreateKpi();
        }

        #endregion

        /// <summary>
        /// Gets node's identifier.
        /// </summary>
        public string Id
        {
            get { return nodeId; }
        }

        /// <summary>
        /// Gets a value indicating whether node has joined to the smart space.
        /// </summary>
        public bool IsJoined
        {
            get { return this.isJoined; }
        }

        /// <summary>
        /// Gets node's session.
        /// </summary>
        public ISession Session
        {
            get { return session; }
        }

        internal string NodeSessionId
        {
            get { return nodeSessionId; }
        }

        #region LocalFunctions
        /// <summary>
        /// Gets new individual of given class.
        /// </summary>
        /// <param name="ontologyClass">Class of individual.</param>
        /// <returns>New individual or null if individual can not be created.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        public Individual CreateIndividual(OntologyClass ontologyClass)
        {
            return new Individual(ontologyClass, session);
        }

        ////TODO: add checking uuid in the SS

        /// <summary>
        /// Gets new individual of given class with given uuid.
        /// </summary>
        /// <param name="ontologyClass">Class of individual.</param>
        /// <param name="uri">URI to identify individual.</param>
        /// <returns>New individual or null if individual can not be created.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        public Individual CreateIndividual(OntologyClass ontologyClass, string uri)
        {
            return new Individual(ontologyClass, uri);
        }

       
        /// <summary>
        /// Create a new subscription. The subscritption can be synchronous and asynchronous. For the synchronous subscription
        /// you need to call function that blocks current thread and waits data from the smart space. 
        /// Asynchronous subscriptions are processed in a background thread.
        /// </summary>
        /// <param name="isAsynchronous">Is the subscription asynchronous.</param>
        /// <returns>New created subscription.</returns>
        public Subscription CreateSubscription(bool isAsynchronous = false)
        {
            return new Subscription(this, isAsynchronous);
        }

        #endregion

        #region SSFunctions

        /// <summary>
        /// Joins to the smart space.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Thrown when node is joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Join()
        {
            //const string MethodName = "Join"; // Only for debug, without runtime getting to support WP
            //const int ManagementType = LoggerManager.SessionManagement; // Only for debugging, type of the function

            if (isJoined == true)
            {
                throw new InvalidOperationException("Cannot join to the smart space, leave it first.");
            }

            if (NativeMethods.sslog_node_join(sslog_node) != Constant.SSLOG_ERROR_NO)
            {
                throw new InvalidOperationException("Error " + NativeMethods.sslog_error_get_last_code() + ". "
                     + NativeMethods.sslog_error_get_last_text());
            }

            subscriber = new Subscriber(session);

            isJoined = true;

            LoggerManager.WriteMessage(LoggerManager.SessionManagement, "Join", String.Format("Node '{0}' has joined.", nodeId));
        }

        /// <summary>
        /// Leaves the smart space. It also try to unsubscribe all active sibscription (sinchronous and asynschronous). 
        /// Sometimes it can take a long time, it depends on the number of subscriptions and communication speed with the smart space.
        /// It does nothing if the node is not joined to the smart space.
        /// </summary>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Leave()
        {
            //const string MethodName = "Join"; // Only for debug, without runtime getting to support WP
            //const int ManagementType = LoggerManager.SessionManagement; // Only for debugging, type of the function

            if (isJoined == false)
            {
                return;
            }

            subscriber.UnsubscribeAll();

            isJoined = false;
            if (NativeMethods.sslog_node_join(sslog_node) != Constant.SSLOG_ERROR_NO)
            {
                throw new InvalidOperationException("Error " + NativeMethods.sslog_error_get_last_code() + ". "
                     + NativeMethods.sslog_error_get_last_text());
            }


            try
            {
             ///   KpiHelper.DoKpi(kpi.Leave, ManagementType, MethodName);
            }
            finally
            {
                isJoined = false;
                subscriber = null;
            }
        }

        /// <summary>
        /// Sets parameters of the node. You can use this function only if the node is not joined to the smart space.
        /// </summary>
        /// <param name="spaceId">Identifier of the smart space.</param>
        /// <param name="address">Address of the smart space or SIB (IP for TCPIP).</param>
        /// <param name="port">Port of the smart space or SIB.</param>
        /// <exception cref="System.InvalidOperationException">Thrown when node is joined to the smart space.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown when the smart space identifier or address is null or empty.</exception>
        private void SetParameters(string spaceId, string address, int port)
        {
            if (isJoined == true)
            {
                throw new InvalidOperationException("Cannot change settings while node is joined to the smart space");
            }

            if (String.IsNullOrWhiteSpace(spaceId))
            {
                throw new ArgumentNullException("spaceId", "Parameter can not be null or empty.");
            }

            if (String.IsNullOrWhiteSpace(address))
            {
                throw new ArgumentNullException("address", "Parameter can not be null or empty.");
            }

            //kpi.SetSpaceInfo(this.nodeId, spaceId, address, port);
        }

        /// <summary>
        /// Insert a whole individual to smart space.
        /// This functions gets all propertis of the given individual and send it to smart space. 
        /// </summary>
        /// <param name="individual">Individual to insert.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <exception cref="System.ArgumentException">Thrown when individual is not registered in the session.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Insert(Individual individual)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (individual == null)
            {
                throw new ArgumentNullException("individual");
            }
            //if (!repository.IsEntityExists(individual))
            //{
            //    throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            //}

            //List<ITriple> triples = OwlMapping.IndividualToTriples(kpi, individual);

            //KpiHelper.DoKpi(this.kpi.Insert, triples, LoggerManager.TriplesManagement, "Insert(Individual)");

            if (individual.NativePtr.ToInt32().Equals(0))
                throw new ArgumentException("individual");

            int result = NativeMethods.sslog_node_insert_individual(this.NativeNode, individual.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }
        }

        

        //// TODO: Checks property for given individual
        //// TODO: Set errors or throw

        /// <summary>
        /// Gets property from smart space.
        /// </summary>
        /// <remarks>
        /// Tries to gets triple from SmartSpace that represents given property for the given individual.
        /// If it's object-property then a new individual will be created or will be found locally if it was created early.
        /// This function also set property locally for the given individual.
        /// </remarks>
        /// <param name="individual">Individual that probably has property in SmartSpace.</param>
        /// <param name="property">A property that's need to get from the server.</param>
        /// <returns>An individual if the property is object, string if property is literal or null if there is no property in the smart space.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public object GetProperty(Individual individual, Property property)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            //SemanticChecker.CheckPropertyGet(repository, individual, property);

            //const string MethodName = "GetProperty"; // Only for debug, without runtime getting to support WP
            //const int ManagementType = LoggerManager.SessionManagement; // Only for debugging, type of the function

            //List<ITriple> requestTriples = new List<ITriple>();
            //List<ITriple> resultTriples = new List<ITriple>();

            IntPtr result = NativeMethods.sslog_node_get_property(this.NativeNode, individual.NativePtr, property.NativePtr);
                
            if(result.ToInt32().Equals(0))
                return null;

            //requestTriples.Add(kpi.CreateTriple(individual.Uri, property.Uri, kpi.AnyUri, RdfType.Uri, RdfType.Uri));

            //KpiHelper.DoKpi(this.kpi.Query, requestTriples, resultTriples, ManagementType, MethodName);

            //if (resultTriples.Count == 0)
            //{
            //    return null;
            //}

            object value = PrepareValueForProperty(property, result);

            if (value == null)
            {
                return null;
            }

            if (property.IsObject())
            {
                individual.InsertProperty(property, value as Individual);
            }
            else
            {
                individual.InsertProperty(property, value as string);
            }

            return value;
        }

        /// <summary>
        /// Sets property in smart space, also set it for the local individual. 
        /// If the local individual has the property with given vlaue, it dose not set another property.
        /// </summary>
        /// <param name="individual">Individual for setting property.</param>
        /// <param name="property">Property for setting.</param>
        /// <param name="value">Value of the property (string for the data-property, or individual for the object-property).</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        private void SetProperty(Individual individual, Property property, Individual value)
        {
            SetProperty(individual, property, value as object);
        }

        /// <summary>
        /// Sets property in smart space, also set it for the local individual. 
        /// If the local individual has the property with given vlaue, it dose not set another property.
        /// </summary>
        /// <param name="individual">Individual for setting property.</param>
        /// <param name="property">Property for setting.</param>
        /// <param name="value">Value of the property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        private void SetProperty(Individual individual, Property property, string value)
        {
            SetProperty(individual, property, value as object);
        }

        /// <summary>
        /// Remove property with given value from smart space and from local indivudal.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="individual">Individual for removing property.</param>
        /// <param name="property">Property for removing.</param>
        /// <param name="value">Value of property, it must be string for data-property of individual for object-property.</param>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void RemoveProperty(Individual individual, Property property, Individual value)
        {
            RemoveProperty(individual, property, value as object);
        }

        /// <summary>
        /// Remove property with given value from smart space and from local indivudal.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="individual">Individual for removing property.</param>
        /// <param name="property">Property for removing.</param>
        /// <param name="value">Value of property, it must be string for data-property of individual for object-property.</param>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void RemoveProperty(Individual individual, Property property, string value)
        {
            RemoveProperty(individual, property, value as object);
        }

        public static void Discoverу(string broadcast_address = null, int port = -1)
        {
            IntPtr ss_info = NativeMethods.ss_discovery(broadcast_address, port);

            Collection<ss_info_t> ss_info_list = new Collection<ss_info_t>();
            
            int count = MarshalHelp.CountList(ss_info);

            for (int i = 0; i < count; ++i)
            {
                IntPtr s = Marshal.ReadIntPtr(ss_info, i * IntPtr.Size);
                ss_info_t e = (ss_info_t)Marshal.PtrToStructure(s, typeof(ss_info_t));
                ss_info_list.Add(e);
            }
        }

        //// TODO: Individual update function works as Set then no old Value finded

        /// <summary>
        /// Update means update existence property or create a new object property. Firs it updates data in the smart space, then it updates data in the local individuals.
        /// </summary>
        /// <param name="individual">Individual for updating property.</param>
        /// <param name="property">Property for updating.</param>
        /// <param name="oldValue">Old value of the property.</param>
        /// <param name="newValue">New value of the property.</param>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void UpdateProperty(Individual individual, Property property, Individual oldValue, Individual newValue)
        {
            UpdateProperty(individual, property, oldValue as object, newValue as object);
        }

        /// <summary>
        /// Update means update existence property or create a new object property. Firs it updates data in the smart space, then it updates data in the local individuals.
        /// </summary>
        /// <param name="individual">Individual for updating property.</param>
        /// <param name="property">Property for updating.</param>
        /// <param name="oldValue">Old value of the property.</param>
        /// <param name="newValue">New value of the property.</param>
        /// <exception cref="System.ArgumentException">Thrown when individual or property is not registered in the session.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when cardinality is violated.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void UpdateProperty(Individual individual, Property property, string oldValue, string newValue)
        {
            UpdateProperty(individual, property, oldValue as object, newValue as object);
        }

        /// <summary>
        /// Get individuals of given ontology class
        /// </summary>
        /// <param name="ontologyClass">Searching ontology class.</param>
        /// <returns>List of individuals of given class or empty List, if not exists.</returns>
        /// <exception cref="System.ArgumentException">Thrown when ontology class is not registered in the session.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public Collection<Individual> GetIndividuals(OntologyClass ontologyClass)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            IntPtr nativeIndividuals = NativeMethods.sslog_node_get_individuals_by_class(this.NativeNode, ontologyClass.NativePtr);

            Collection<Individual> individuals = NativeMethods.IndividualListToManagedList(nativeIndividuals, this.session);

            return individuals;
        }

        /// <summary>
        /// Get individuals of given URIs list
        /// </summary>
        /// <param name="ontologyClass">URIs list.</param>
        /// <returns>List of individuals of given URIs list or empty List, if not exists.</returns>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public Collection<Individual> GetIndividuals(List<string> uriList)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            Collection<Individual> individuals = new Collection<Individual>();

            foreach(string uri in uriList)
            {
                Individual ind = this.GetIndividual(uri);
                if(ind != null)
                    individuals.Add(ind);
            }

            return individuals;
        }

        /// <summary>
        /// Get individuals of given URIs list
        /// </summary>
        /// <param name="ontologyClass">URIs list.</param>
        /// <returns>List of individuals of given URIs list or empty List, if not exists.</returns>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public Individual GetIndividual(string uri)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            IntPtr nativeIndividual = NativeMethods.sslog_node_get_individual_by_uri(this.NativeNode, uri);

            if (nativeIndividual.ToInt32().Equals(0))
                return null;

            Individual individual = new Individual(nativeIndividual, this.session);

            return individual;
        }

        //// TODO: handle errors

        /// <summary>
        /// Search individual with properties in the smart space.
        /// </summary>
        /// <param name="individual">individual for searching.</param>
        /// <returns>True if individual in the smart space or false if it not exists.</returns>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        private bool Exist(Individual individual)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (!repository.IsEntityExists(individual))
            {
                //    throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            }

            //// TODO: need to send only uuid triple without properties
            //List<ITriple> requestTriples = OwlMapping.IndividualToTriples(kpi, individual);
            List<ITriple> resultTriples = new List<ITriple>();

            //KpiHelper.DoKpi(kpi.Query, requestTriples, resultTriples, LoggerManager.EntitiesManagement, "Exist");

            if (resultTriples.Count > 0)
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Removes individual from the smart space. All properties and all information about indivudlal removed by this function only from smart space.
        /// </summary>
        /// <param name="individual">Individual to remove.</param>
        /// <param name="clearLocalStore">Remove individual from local store too.</param>
        /// <exception cref="System.ArgumentException">Thrown when the individual is not registered in the session.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public void Remove(Individual individual, bool withLocal = false)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if(individual == null)
            {
                throw new ArgumentNullException("individual");
            }

            int result;

            if(withLocal.Equals(false))
                result = NativeMethods.sslog_node_remove_individual(this.NativeNode, individual.NativePtr);
            else
                result = NativeMethods.sslog_node_remove_individual_with_local(this.NativeNode, individual.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            //if (!repository.IsEntityExists(individual))
            //{
            //    throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            //}

            //List<ITriple> triples = OwlMapping.IndividualToAnyAny(kpi, individual);

            //KpiHelper.DoKpi(kpi.Remove, triples, LoggerManager.EntitiesManagement, "Remove");
        }

        /// <summary>
        /// Updates individual in the smart space. Update is not one operation, this function removes individual from the smart space
        /// and then inserts it.
        /// </summary>
        /// <param name="individual">Individual to update.</param>
        /// <exception cref="System.ArgumentException">Thrown when individual is not registered in the session.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        private void Update(Individual individual)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (!repository.IsEntityExists(individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            }

            Remove(individual);
            Insert(individual);
        }

      

        /// <summary>
        /// Populates individual from the smart space.
        /// </summary>
        /// <param name="individual">individual to populate.</param>
        /// <exception cref="System.ArgumentException">Thrown when the individual is not registered in the session.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        public void Populate(Individual individual)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            int result = NativeMethods.sslog_node_populate(this.NativeNode, individual.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            return;


            //if (!repository.IsEntityExists(individual))
            //{
            //    throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            //}

            //List<ITriple> requestedTriples = OwlMapping.IndividualToTriplesAny(kpi, individual);
            //List<ITriple> resultTriples = new List<ITriple>();



            //KpiHelper.DoKpi(kpi.Query, requestedTriples, resultTriples, LoggerManager.EntitiesManagement, "Populate");

            //if (resultTriples.Count <= 0)
            //{
            //    return;
            //}

            //Individual ind = OwlMapping.TripplesToIndividualOne(kpi, repository, resultTriples);

        }

        
        #endregion

        #region SubscribeMethods

        /// <summary>
        /// Subscribes session. 
        /// </summary>
        /// <param name="subscription">Subscripiton to subscribe.</param>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        /// <exception cref="System.ArgumentNullException">Thrown when the subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Subscribe(Subscription subscription)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            subscriber.Subscribe(subscription);
        }

        /// <summary>
        /// Waits subscribed data for the synchronous subscription session. 
        /// This function does nothing for asynscronous sessions and sessions that is not subscribed.
        /// </summary>
        /// <param name="subscription">Subscription to wait notifications.</param>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        /// <exception cref="System.ArgumentNullException">Thrown when the subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void WaitSubscribe(Subscription subscription)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            subscriber.WaitSubscribe(subscription);
        }

        /// <summary>
        /// Indicates what checking of the subscription must be stopped.
        /// <remarks>
        /// This function works only with synchronous subscriptions.
        /// If you want to stop waiting, then you call this function.
        /// It only indicates, but not really stops checking, you need to wait while
        /// the <see cref="WaitSubscribe"/> ends, and then make other actions under subscription.
        /// </remarks>
        /// </summary>
        /// <param name="subscription">Subscription to stop waiting.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the subscription is null.</exception>
        public void StopWaiting(Subscription subscription)
        {
            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            //subscriber.StopWaiting(subscription);
        }

        /// <summary>
        /// Unsubscribe subscripiron. After this session will be in inactive state.j
        /// </summary>
        /// <param name="subscription">Subscribed subscripiron.</param>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        /// <exception cref="System.ArgumentNullException">Thrown when the subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joined to the smart space.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Unsubscribe(Subscription subscription)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            subscriber.Unsubscribe(subscription);
        }

        /// <summary>
        /// Unsubscribe all subscripiton that were subscribed. 
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        /// <exception cref="System.InvalidOperationException">Thrown when node is not joioned to the smart space.</exception>
        public void UnsubscribeAll()
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            //subscriber.UnsubscribeAll();
        }

        #endregion

        /// <summary>
        /// Destroys the node. It unsubscribe all subscripiton and leave the smart space.
        /// After this method the node cannot be used.
        /// </summary>
        internal void Destroy()
        {
            try
            {
                if (IsJoined)
                {
                    //subscriber.UnsubscribeAll();
                    Leave();
                    //kpi.Shutdown();
                }
            }
            catch (Exception ex)
            {
                LoggerManager.WriteExceptionMessage(LoggerManager.NodeManagement, "Destroy", ex);
            }
            finally
            {
                this.nodeId = null;
                //this.kpi = null;
                this.session = null;
                //this.subscriber = null;
                this.repository = null;
            }
        }

        #region Private functions

        /// <summary>
        /// Sets property in smart space, also set it for the local individual. 
        /// If the local individual has the property with given vlaue, it dose not set another property.
        /// </summary>
        /// <param name="individual">Individual for setting property.</param>
        /// <param name="property">Property for setting.</param>
        /// <param name="value">Value of the property (string for the data-property, or individual for the object-property).</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <exception cref="System.InvalidOperationException">Thrown when a node is not joined to the smart space.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown when some of parameters are null.</exception>
        private void SetProperty(Individual individual, Property property, object value)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            //SemanticChecker.CheckPropertySet(repository, individual, property, value);

            individual.InsertProperty(property, value);

            //List<ITriple> triples = new List<ITriple>();

            ////OwlMapping.PropertyToTriples(kpi, individual, property, value, triples);
            ////triples.Add(OwlMapping.GetRdfTypeTriple(kpi, individual));

            //try
            //{
            //    //KpiHelper.DoKpi(kpi.Insert, triples, LoggerManager.PropertiesManagement, "SetProperty");
            //}
            //catch
            //{
            //    //// Remove property that was set.
            //    try
            //    {
            //        individual.RemoveProperty(property, value);
            //    }
            //    catch (Exception ex)
            //    {
            //        LoggerManager.WriteExceptionMessage(LoggerManager.PropertiesManagement, "SetProperty", ex);
            //    }

            //    throw;
            //}
        }

        /// <summary>
        /// Remove property with given value from smart space and from local indivudal.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="individual">Individual for removing property.</param>
        /// <param name="property">Property for removing.</param>
        /// <param name="value">Value of property, it must be string for data-property of individual for object-property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        private void RemoveProperty(Individual individual, Property property, object value)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (individual.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered individual");
            }

            if (property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            int result;

            if (property.IsObject())
                result = NativeMethods.sslog_node_remove_property(this.NativeNode, individual.NativePtr, property.NativePtr, (value as Individual).NativePtr);
            else
                result = NativeMethods.sslog_node_remove_property(this.NativeNode, individual.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(value as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            //SemanticChecker.CheckPropertyRemove(repository, individual, property, value);

            //List<ITriple> triples = new List<ITriple>();
            ////OwlMapping.PropertyToTriples(kpi, individual, property, value, triples);

            ////KpiHelper.DoKpi(kpi.Remove, triples);
            //individual.RemoveProperty(property, value);
        }

        /// <summary>
        /// Updates property. Firs it updates data in the smart space, then it updates data in the local individuals.
        /// Use this method as internal or prive, this method is not for API, for API there methods with individuals 
        /// and strings as value for property.
        /// </summary>
        /// <param name="individual">Individual for updating property.</param>
        /// <param name="property">Property for updating.</param>
        /// <param name="oldValue">Old value of the property.</param>
        /// <param name="newValue">New value of the property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        private void UpdateProperty(Individual individual, Property property, object oldValue, object newValue)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (individual.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered individual");
            }

            if(property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            int result;

            if (property.IsObject())
                result = NativeMethods.sslog_node_update_property(this.NativeNode, individual.NativePtr, property.NativePtr, (oldValue as Individual).NativePtr, (newValue as Individual).NativePtr);
            else
                result = NativeMethods.sslog_node_update_property(this.NativeNode, individual.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(oldValue as string), MarshalHelp.StringToPtr(newValue as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }


            //SemanticChecker.ChekPropertyUpdate(repository, individual, property, oldValue, newValue);

            //List<ITriple> insertingTriples = new List<ITriple>();
            //List<ITriple> removingTriples = new List<ITriple>();

            //OwlMapping.PropertyToTriples(kpi, individual, property, newValue, insertingTriples);
            //OwlMapping.PropertyToTriples(kpi, individual, property, oldValue, removingTriples);

            //insertingTriples.Add(OwlMapping.GetRdfTypeTriple(kpi, individual));

            //KpiHelper.DoKpiUpdate(kpi.Update, removingTriples, insertingTriples, LoggerManager.PropertiesManagement, "UpdateProperty");
            //individual.UpdateProperty(property, oldValue, newValue);
        }

        /// <summary>
        /// Checks if given property set for given individual with given value in SS
        /// </summary>
        /// <param name="individual">Individual to check.</param>
        /// <param name="property">Property to check.</param>
        /// <param name="value">Object property value.</param>
        /// <returns>True if property is set, false otherwise</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        private bool IsPropertySet(Individual individual, Property property, Individual value)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if (!property.IsObject())
            {
                return false;
            }

            object check = GetProperty(individual, property);

            if (check == null)
            {
                return false;
            }

            Individual check_obj = (Individual)check;

            if (check_obj.Uri.Equals(value.Uri))
            {
                return true;
            }

            return false;
        }

        //// TODO: to other class
        private object PrepareValueForProperty(Property property, ITriple triple)
        {
            if (!property.IsObject())
            {
                return triple.ObjectElement;
            }

            string uuid = triple.ObjectElement;

            Individual ind = repository.GetIndividual(uuid);

            if (ind == null)
            {
                //OntologyClass ontClass = OwlMapping.GetClassByUuid(kpi, repository, uuid);
                //ind = repository.GetOrCreateIndividual(ontClass, uuid);
            }

            return ind;
        }

        private object PrepareValueForProperty(Property property, IntPtr value)
        {
            if (!property.IsObject())
            {
                return MarshalHelp.PtrToString(value);
            }

            Individual ind = new Individual(value);

            if (ind == null)
            {
                //OntologyClass ontClass = OwlMapping.GetClassByUuid(kpi, repository, uuid);
                //ind = repository.GetOrCreateIndividual(ontClass, uuid);
            }

            return ind;
        }

        #endregion


        #region Insert-property

        public void InsertProperty(Individual individual, Property property, object value)
        {
            if (isJoined == false)
            {
                throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
            }

            if(value.Equals(null))
            {
                throw new ArgumentException("value");
            }

            if (individual.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered individual");
            }

            if (property.NativePtr.ToInt32().Equals(0))
            {
                throw new ArgumentException("Not registered property");
            }

            int result;

            if(property.IsObject())
                result = NativeMethods.sslog_node_insert_property(this.NativeNode, individual.NativePtr, property.NativePtr, (value as Individual).NativePtr);
            else
                result = NativeMethods.sslog_node_insert_property(this.NativeNode, individual.NativePtr, property.NativePtr, MarshalHelp.StringToPtr(value as string));

            MarshalHelp.FreeIntPtr();

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            //HandlerConfig config = property.PrepareConfig(ActionType.Insert, individual, value);

            //Collection<ITriple> insertedTriples = new Collection<ITriple>();
            //Collection<ITriple> removedTriples = new Collection<ITriple>();

            //MakeAction(config, insertedTriples, removedTriples);

            //KpiHelper.Do(kpi, removedTriples, insertedTriples);
        }

        public void InsertProperty(Individual individual, Property property, Individual value)
        {
            InsertProperty(individual, property, (object)value);
        }

        public void InsertProperty(Individual individual, Property property, string value)
        {
            InsertProperty(individual, property, (object)value);
        }


        public string GetLastError()
        {
            return "Error " + NativeMethods.sslog_error_get_last_code() + ". "
                     + NativeMethods.sslog_error_get_last_text();
        }

        private void MakeAction(HandlerConfig config, ICollection<ITriple> insertedTriples, ICollection<ITriple> removedTriples)
        {
            foreach (PropertyChangesData change in config.Changes)
            {
                Individual individual = change.Owner;

                switch (change.Action)
                {
                    // Local changesa for local individual only.                        
                    case ActionType.LocalInsert:
                        change.Owner.InsertProperty(change.Property, change.CurrentValue);
                        break;
                    case ActionType.LocalRemove:
                        change.Owner.RemoveProperty(change.Property, change.CurrentValue);
                        break;
                    case ActionType.LocalUpdate:
                        change.Owner.UpdateProperty(change.Property, change.PreviousValue, change.CurrentValue);
                        break;

                    // Changes for loacal individual and remote triple store.
                    case ActionType.Insert:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.CurrentValue, insertedTriples);
                        change.Owner.InsertProperty(change.Property, change.CurrentValue);
                        break;
                    case ActionType.Remove:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.PreviousValue, removedTriples);
                        change.Owner.RemoveProperty(change.Property, change.PreviousValue);
                        break;
                    case ActionType.Update:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.CurrentValue, insertedTriples);
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.PreviousValue, removedTriples);
                        change.Owner.UpdateProperty(change.Property, change.PreviousValue, change.CurrentValue);
                        break;

                    // Changes for remote triple store.
                    case ActionType.RemoteInsert:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.CurrentValue, insertedTriples);
                        break;
                    case ActionType.RemoteRemove:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.PreviousValue, removedTriples);
                        break;
                    case ActionType.RemoteUpdate:
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.CurrentValue, insertedTriples);
                        //OwlMapping.PropertyToTriples(kpi, change.Owner, change.Property, change.PreviousValue, removedTriples);
                        break;

                }
            }

            foreach (Individual i in config.Individuals)
            {
                //insertedTriples.Add(OwlMapping.GetRdfTypeTriple(kpi, i));
            }
        }

        //public void Remove(Individual individual, GeneralProperty property, object value)
        //{
        //    if (isJoined == false)
        //    {
        //        throw new InvalidOperationException(ExceptionManager.NotJoinMessage);
        //    }

        //    HandlerConfig config = property.PrepareConfig(ActionType.Remove, individual, value);

        //    Collection<ITriple> insertedTriples = new Collection<ITriple>();
        //    Collection<ITriple> removedTriples = new Collection<ITriple>();

        //    MakeAction(config, insertedTriples, removedTriples);

        //    KpiHelper.Do(kpi, removedTriples, insertedTriples);
        //}

        #endregion General-property

    }
}

