﻿//  ------------------------------------------------------------------------
//  <copyright file="OntologyClass.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// This class is used for representing class entity from ontology. It is used as base for creating <see cref="PetrSU.SmartSlog.Individual"/>.
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/>
    public class OntologyClass : OntologyEntity
    {
        /// <summary>
        /// Individuals of this class.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        private List<Individual> individuals = new List<Individual>();

        /// <summary>
        /// Superclasses of this class.
        /// </summary>
        private List<OntologyClass> superClasses = new List<OntologyClass>();

        /// <summary>
        /// Properties of this class.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        private List<Property> properties = new List<Property>();

        /// <summary>
        /// Pointer on SmartSlog native structure of Ontology Class
        /// </summary>
        private IntPtr nativePtr;

        /// <summary>
        /// Pointer on SmartSlog native structure of Ontology Class
        /// </summary>
        internal IntPtr NativePtr { get { return nativePtr; } set { nativePtr = value; } }



        /// <summary>
        /// Initializes a new instance of the OntologyClass class.
        /// </summary>
        /// <param name="uri">Full URI of the class.</param>
        /// <param name="properties">List with properties.</param>
        /// <param name="superClasses">Superclasses of this class.</param>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public OntologyClass(string uri, List<Property> properties = null, List<OntologyClass> superClasses = null)
        {
            if (String.IsNullOrEmpty(uri) == true)
            {
                throw new System.ArgumentException("URI can not be null or empty", "uri");
            }

            SetUri(uri);

            if (properties != null)
            {
                this.properties = properties;
            }

            if (superClasses != null)
            {
                this.superClasses = superClasses;
            }
        }

        /// <summary>
        /// Gets current count of individual of the class.
        /// </summary>
        internal int IndividualCount
        {
            get { return this.individuals.Count; }
        }

        /// <summary>
        /// Gets list with properties of this class.
        /// </summary>
        /// <returns>Properties that are related to the class.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public ReadOnlyCollection<Property> GetProperties()
        {
#if (PORTABLE || WINDOWS_PHONE)
            return new ReadOnlyCollection<Property>(this.properties);
#else
            return this.properties.AsReadOnly(); 
#endif

        }

        /// <summary>
        /// Gets superclasses.
        /// </summary>
        /// <returns>Classes that are parent for the current class.</returns>
        public ReadOnlyCollection<OntologyClass> GetSuperclasses()
        {
#if (PORTABLE || WINDOWS_PHONE)
            return new ReadOnlyCollection<OntologyClass>(this.superClasses);
#else
            return this.superClasses.AsReadOnly(); 
#endif
        }

        /// <summary>
        /// Adds new individual to the instances list of this class.
        /// </summary>
        /// <param name="individual">Individual to insetr.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        internal void AddIndividual(Individual individual)
        {
            if (this.individuals.Contains(individual) == false && individual.ParentClass == this)
            {
                this.individuals.Add(individual);
            }
        }

        /// <summary>
        /// Removes individual from instances.
        /// </summary>
        /// <param name="individual">Individual to remove.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        internal void RemoveIndividual(Individual individual)
        {
            this.individuals.Remove(individual);
        }

        /// <summary>
        /// Gets rdf:type triple of class.
        /// 
        /// Rdf:type triple of class:
        /// ClassURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/2000/01/rdf-schema#Class.
        /// </summary>
        /// <returns>Triple on success or NULL otherwise (class is NULL).</returns>
        public Triple toTriple()
        {
            return toTriple(this);
        }
        
        /// <summary>
        /// Gets rdf:type triple of class.
        /// 
        /// Rdf:type triple of class:
        /// ClassURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/2000/01/rdf-schema#Class.
        /// </summary>
        ///<param name="ontologyClass">Class to convert</param>
        /// <returns>Triple on success or NULL otherwise (class is NULL).</returns>
        static internal Triple toTriple(OntologyClass ontologyClass)
        {
            if (ontologyClass == null)
                throw new ArgumentNullException("ontologyClass");

            IntPtr triplePtr = NativeMethods.sslog_class_to_triple(ontologyClass.nativePtr);

            if (triplePtr.ToInt32().Equals(0))
                return null;
            else
                return new Triple(triplePtr);
        }
    }
}

