﻿//  ------------------------------------------------------------------------
//  <copyright file="Property.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// This class represents a property from ontology. 
    /// <remarks>
    /// You can set different properties for <see cref="PetrSU.SmartSlog.Individual"/> class. The setting means set value fo property for individual.
    /// Properties can be object or data. Object-propeties links one object with another object. 
    /// For data-properties some simple data (strings, numbers etc) is used as values.
    /// In the SmartSlog object-properties manipulates with individuals and data-properties manipulates with strings only, 
    /// it's mean that a developer translate data (for example numbers) to the string format by hand.
    /// Also properties has cardinality, it can be max and min. Cardinality is used as a restriction of properties count. You can't set more properties value 
    /// than maximum cardinality and can't remove more properties value than minimum cardinality.
    /// </remarks>
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/> 
    public class Property : OntologyEntity
    {
        /// <summary>
        /// Parent properties list.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Property"/> 
        private List<Property> parentProperties = new List<Property>();

        /// <summary>
        /// Maximum cardinality. The higher restriction of this property for setting. Then it equals -1 it means no restriction.
        /// </summary>
        private int maxCardinality;

        /// <summary>
        /// Minimum cardinality. The lower restriction of this property for setting. Then it equals -1 it means no restriction.
        /// </summary>
        private int minCardinality;

        /// <summary>
        /// Flag of the property type.
        /// </summary>
        private bool isObject = false;

        /// <summary>
        /// Pointer on SmartSlog native structure of Ontology Property
        /// </summary>
        private IntPtr nativePtr;

        /// <summary>
        /// Pointer on SmartSlog native structure of Ontology Property
        /// </summary>
        internal IntPtr NativePtr { get { return nativePtr; } set { nativePtr = value; } }


        /// <summary>
        /// Initializes a new instance of the Property class without cardinality.
        /// </summary>
        /// <param name="uri">URI of the property.</param>
        /// <param name="isObject">Flag of the property type.</param>
        /// <param name="parentProperties">List of parant properties.</param>
        public Property(string uri, bool isObject, List<Property> parentProperties)
        {
            if (String.IsNullOrWhiteSpace(uri) == true)
            {
                throw new System.ArgumentException("Uri can not be null or empty", "uri");
            }

            SetUri(uri);

            if (parentProperties != null)
            {
                this.parentProperties = parentProperties;
            }

            this.maxCardinality = -1;
            this.minCardinality = -1;

            this.isObject = isObject;
        }

        /// <summary>
        /// Initializes a new instance of the Property class with cardinality.
        /// </summary>
        /// <param name="uri">URI of the property.</param>
        /// <param name="maxCardinality">Maximum cardinality. </param>
        /// <param name="minCardinality">Minimum cardinality.</param>
        /// <param name="isObject">Flag of the property type.</param>
        /// <param name="parentProperties">List of parant properties.</param>
        public Property(string uri, int maxCardinality, int minCardinality, bool isObject, List<Property> parentProperties)
            : this(uri, isObject, parentProperties)
        {
            if (maxCardinality >= 0 && maxCardinality < minCardinality)
            {
                throw new System.ArgumentException("Maximum cardinality can not be less than minimal cardinality");
            }

            this.maxCardinality = maxCardinality;
            this.minCardinality = minCardinality;
        }

        /// <summary>
        /// Initializes a new instance of the Property class with cardinality.
        /// </summary>
        /// <param name="uri">URI of the property.</param>
        /// <param name="maxCardinality">Maximum cardinality. </param>
        /// <param name="minCardinality">Minimum cardinality.</param>
        /// <param name="isObject">Flag of the property type.</param>
        public Property(string uri, int maxCardinality, int minCardinality, bool isObject)
            : this(uri,  maxCardinality, minCardinality, isObject, null)
        {
            
        }

        /// <summary>
        /// Initializes a new instance of the Property class without cardinality.
        /// </summary>
        /// <param name="uri">URI of the property.</param>
        /// <param name="isObject">Flag of the property type.</param>
        public Property(string uri, bool isObject)
            :this (uri, isObject, null)
        {
            
        }

        /// <summary>
        /// Gets the maximum cardinality, the higher restriction of this property for setting. Then it equals -1 it means no restriction.
        /// </summary>
        public int MaxCardinality
        {
            get { return maxCardinality; }
        }

        /// <summary>
        /// Gets the minimum cardinality, the lower restriction of this property for setting. Then it equals -1 it means no restriction.
        /// </summary>
        public int MinCardinality
        {
            get { return minCardinality; }
        }

        /// <summary>
        /// Gets the list with parent properties. 
        /// </summary>
        /// <returns>List with parent properties.</returns>
        public List<Property> GetParentProperties()
        {
            return parentProperties;
        }

        /// <summary>
        /// Gets the flag property type. 
        /// </summary>
        /// <returns>True if the property is object and false otherwise.</returns>
        public bool IsObject()
        {
            return isObject;
        }

        /// <summary>
        /// Gets rdf:type triple of property.
        /// Rdf:type triple of property:
        /// PropertyURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/1999/02/22-rdf-syntax-ns#Property.
        /// </summary>
        /// <returns>Triple on success or NULL otherwise (property is NULL).</returns>
        public Triple toTriple()
        {
            return toTriple(this);
        }

        /// <summary>
        /// Gets rdf:type triple of property.
        /// Rdf:type triple of property:
        /// PropertyURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/1999/02/22-rdf-syntax-ns#Property.
        /// </summary>
        /// <param name="property">Property to convert</param>
        /// <returns>Triple on success or NULL otherwise (property is NULL).</returns>
        static internal Triple toTriple(Property property)
        {
            if (property == null)
                throw new ArgumentNullException("property");

            IntPtr triplePtr = NativeMethods.sslog_property_to_triple(property.nativePtr);

            if (triplePtr.ToInt32().Equals(0))
                return null;
            else
                return new Triple(triplePtr);
        }
    }
}
