﻿//  ------------------------------------------------------------------------
//  <copyright file="PropertyChangesData.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// Class that stores changes of a property. 
    /// Property can be set/removed/updated for an individual, this class stores action and old and new value of a property. 
    /// </summary>
    public class PropertyChangesData
    {
        /// <summary>
        /// Property of changes.
        /// </summary>
        private Property property;

        /// <summary>
        /// Previous value of the property.
        /// </summary>
        private object oldValue;

        /// <summary>
        /// Current or new value of the property.
        /// </summary>
        private object newValue;

        /// <summary>
        /// Owner of the property. Individual that has changed property.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        private Individual owner;

        /// <summary>
        /// Action that was perfome under property (set, update, remove).
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        private ActionType action;

        /// <summary>
        /// Initializes a new instance of the PropertyChangesData class.
        /// </summary>
        /// <param name="owner">Owner of the property. Individual that has changed property.</param>
        /// <param name="property">Property that was changed.</param>
        /// <param name="newValue">New value of the property.</param>
        /// <param name="oldValue">Old value of the property.</param>
        /// <param name="action">Action that waas perfom under property.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual" />
        /// <seealso cref="PetrSU.SmartSlog.Property" />
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        /// <exception cref="System.ArgumentNullException">Thrown when owner or property is null.</exception>
        internal PropertyChangesData(Individual owner, Property property, object newValue, object oldValue, ActionType action)
        {
            if (owner == null)
            {
                throw new ArgumentNullException("owner");
            }
            else if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            this.property = property;
            this.oldValue = oldValue;
            this.newValue = newValue;
            this.owner = owner;
            this.action = action;
        }

        /// <summary>
        /// Gets individual that has changed property.
        /// </summary>
        public Individual Owner
        {
            get { return owner; }
        }

        /// <summary>
        /// Gets Property that was changed.
        /// </summary>
        public Property Property
        {
            get { return property; }
        }

        /// <summary>
        /// Gets previous value of the property. If property was inserted (set) than it returns null.
        /// </summary>
        public object PreviousValue
        {
            get { return oldValue; }
        }

        /// <summary>
        /// Gets current value of the property. If property was removed than it returns null.
        /// </summary>
        public object CurrentValue
        {
            get { return newValue; }
        }

        /// <summary>
        /// Gets action that was perfom under property. Property can be inserted(set), removed or updated.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.ActionType" />
        public ActionType Action
        {
            get { return action; }
        }
    }
}
