﻿//  ------------------------------------------------------------------------
//  <copyright file="PropertyValue.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// This class is used to store property value, it stores pair of <see cref="PetrSU.SmartSlog.Property"/> and value of this property 
    /// (string or <see cref="PetrSU.SmartSlog.Individual"/>).  
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Property"/>
    internal class PropertyValue
    {
        #region Privat fields
        /// <summary>
        /// Property for which value is set.
        /// </summary>
        private Property property;

        /// <summary>
        /// State of property.
        /// </summary>
        private State state;

        /// <summary>
        /// Value of property.
        /// </summary>
        private object value;
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the PropertyValue class with data value.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// </summary>
        /// <param name="property">Property for which value setting.</param>
        /// <param name="value">Value of property.</param>
        public PropertyValue(Property property, string value)
        {
            if (value != null && property != null
                && property.IsObject() == false)
            {
                this.value = value;
                this.property = property;
                this.state = State.Set;
            }
            else
            {
                throw new ArgumentException("Property/value is null or value is incorrect for the given property.");
            }
        }

        /// <summary>
        /// Initializes a new instance of the PropertyValue class with object(<see cref="PetrSU.SmartSlog.Individual" />) value.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="property">Property for which value setting.</param>
        /// <param name="value">Value of property.</param>
        public PropertyValue(Property property, Individual value)
        {
            if (value != null && property != null
                && property.IsObject() == true)
            {
                this.value = value;
                this.property = property;
                this.state = State.Set;
            }
            else
            {
                throw new ArgumentException("Property/value is null or value is incorrect for the given property.");
            }
        }

        /// <summary>
        /// Initializes a new instance of the PropertyValue class with object(<see cref="PetrSU.SmartSlog.Pattern" />) value.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Pattern"/>
        /// </summary>
        /// <param name="property">Property for which value setting.</param>
        /// <param name="value">Value of property.</param>
        //public PropertyValue(Property property, Pattern value)
        //{
        //    if (value != null && property != null
        //        && property.IsObject() == true)
        //    {
        //        this.value = value;
        //        this.property = property;
        //        this.state = State.Set;
        //    }
        //    else
        //    {
        //        throw new ArgumentException("Property/value is null or value is incorrect for the given property.");
        //    }
        //}
        #endregion

        #region Enums
        /// <summary>
        /// Described states of value.
        /// </summary>
        ///  <remarks>
        ///  This states useful for synchronization management of individual.
        ///  State can be one of:
        ///  <list type="bullet">
        /// <listheader>
        /// <term>Enum field</term>
        /// <description>description</description>
        /// </listheader>
        /// <item>
        ///     <term>Set</term>
        ///     <description>Value was set</description>
        /// </item>
        /// <item>
        ///     <term>Update</term>
        ///     <description>Value was updated</description>
        /// </item>
        /// <item>
        ///     <term>Remove</term>
        ///     <description>Value was removed from individual</description>
        /// </item>
        /// <item>
        ///     <term>None</term>
        ///     <description>Nothing to happend with value</description>
        /// </item>
        /// </list>
        /// </remarks>
        internal enum State
        {
            /// <summary>
            /// Value was set.
            /// </summary>
            Set,

            /// <summary>
            /// Value was update.
            /// </summary>
            Update,

            /// <summary>
            /// Value was removed.
            /// </summary>
            Remove,

            /// <summary>
            /// Nothing to happend with value.
            /// </summary>
            None
        }
        #endregion

        #region Public properties
        /// <summary>
        /// Gets <see cref="PetrSU.SmartSlog.Property "/> if value.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// </summary>
        public Property Property
        {
            get { return this.property; }
        }
        #endregion

        #region Public methods
        /// <summary>
        /// Gets data value of property.
        /// </summary>
        /// <returns>Value if it is value of data-property or <c>null</c> otherwise.</returns>
        public string GetDataValue()
        {
            // TODO: check property...
            return (string)this.GetValue();
        }

        /// <summary>
        /// Gets object value of property.
        /// </summary>
        /// <returns>Value if it is value of object-property or <c>null</c> otherwise.</returns>
        public Individual GetObjectValue()
        {
            // TODO: check property...
            return (Individual)this.GetValue();
        }

        /// <summary>
        /// Gets value of property.
        /// </summary>
        /// <remarks>
        /// This method do not check a proprty type, 
        /// if you need just object or data value use <see cref="GetDataValue"/> or <see cref="GetDataValue"/>.
        /// </remarks>
        /// <returns>Value of proprty.</returns>
        public object GetValue()
        {
            return this.value;
        }

        /// <summary>
        /// Checks value type.
        /// </summary>
        /// <remarks>
        /// Type of value maybe data or object. 
        /// Object type of value is mean <see cref="PetrSU.SmartSlog.Individual"/>.
        /// </remarks>
        /// <returns><c>true</c> if type is object or <c>false</c> otherwise.</returns>
        public bool IsObjectValue()
        {
            return Property.IsObject();
        }
        #endregion

        #region Internal methods
        /// <summary>
        /// Gets <see cref="PetrSU.SmartSlog.PropertyValue.State"/> of value.
        /// </summary>
        /// <returns>Current state of value.</returns>
        internal State GetState()
        {
            return this.state;
        }

        /// <summary>
        /// Update value.
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="value">New value of property. 
        /// It can be <see cref="System.String"/> for data property 
        /// or <see cref="PetrSU.SmartSlog.Individual"/> for object property.</param>
        /// <returns>true if value upgrade successfully; otherwise, false.</returns>
        internal bool UpdateValue(object value)
        {
            if (value == null)
            {
                return false;
            }

            if ((this.IsObjectValue() == true && value is Individual)
                || (this.IsObjectValue() == false && value is string))
            {
                this.value = value;
                this.state = State.Update;

                return true;
            }

            return false;
        }
        #endregion
    }
}
