﻿//  ------------------------------------------------------------------------
//  <copyright file="Repository.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The <c>Repository</c> class represents ontology entities store.
    /// <remarks>
    ///   <para>
    ///     This is the singleton internal class. It contains information about clases, properties, individuals and can create new individuals.  
    ///   </para>
    ///   </remarks>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/>
    /// <seealso cref="PetrSU.SmartSlog.Property"/>
    /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
    /// </summary>
    internal sealed class Repository : IRepository
    {
        private class PropertyReferncesCounter
        {
            private Property property;

            private Individual individual;

            public PropertyReferncesCounter(Property property, Individual individual)
            {
                this.property = property;

                this.individual = individual;
            }

            public int Count { get; set; }

            public Property Property { get { return this.property; } }

            public Individual Individual { get { return this.individual; } }
        }

        #region Private fields
        /// <summary>
        /// Contains instance of repository.
        /// </summary>
        //private static Repository instance;

        /// <summary>
        /// Store for properties.
        /// <seealso cref="Property"/>
        /// </summary>
        private List<Property> properties = new List<Property>();

        /// <summary>
        /// Store for classes.
        /// <seealso cref="OntologyClass"/>
        /// </summary>
        private List<OntologyClass> classes = new List<OntologyClass>();

        /// <summary>
        /// Store for individuals.
        /// <seealso cref="Individual"/>
        /// </summary>
        private List<Individual> individuals = new List<Individual>();

        //private Dictionary<String, List<Property>> subscribedProperties = new Dictionary<string, List<Property>>();

        private LinkedList<PropertyReferncesCounter> propertyReferences = new LinkedList<PropertyReferncesCounter>();

        private Session session;


        /// <summary>
        /// Object is used to lock multiple acces to repository.
        /// </summary>
        private object repositoryLock = new object();

        //private Dictionary<string, int> propertyRefences = new Dictionary<string, int>();
        #endregion

        public Repository(Session session)
        {
            this.session = session;
        }

        #region Public methods
        /// <summary>
        /// Add property to repository.
        /// </summary>
        /// <param name="property">property to adding to repository</param>
        public void AddEntity(Property property)
        {
            if (property != null
                    && this.properties.Contains(property) == false)
            {
                this.properties.Add(property);
            }
        }

        /// <summary>
        /// Add class to repository.
        /// </summary>
        /// <param name="ontologyClass">class to adding to repository</param>
        public void AddEntity(OntologyClass ontologyClass)
        {
            if (ontologyClass != null && IsExists(this.classes, ontologyClass) == false)
            {
                this.classes.Add(ontologyClass);
            }
        }

        /// <summary>
        /// Add individual to repository.
        /// </summary>
        /// <param name="individual">Individual to adding to repository.</param>
        public void AddEntity(Individual individual)
        {
            if (individual != null && this.individuals.Contains(individual) == false)
            {
                this.individuals.Add(individual);
            }
        }

        //// TODO: Make as URI
        /// <summary>
        /// Gets class by given class type.
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        /// </summary>
        /// <param name="uri">URI of class.</param>
        /// <returns>Class <see cref="PetrSU.SmartSlog.OntologyClass"/> 
        /// or null if there is no class with given class type</returns>
        public OntologyClass GetClassByUri(string uri)
        {
            foreach (OntologyClass ontologyClass in this.classes)
            {
                if (ontologyClass.Uri.Equals(uri) == true)
                {
                    return ontologyClass;
                }
            }

            return null;
        }

        /// <summary>
        /// Try to find individual by given URI.
        /// </summary>
        /// <param name="uri">URI of individual.</param>
        /// <returns>Individual<see cref="PetrSU.SmartSlog.Individual"/> or 
        /// <c>null</c> if individual not found.</returns>
        public Individual GetIndividual(string uuid)
        {
            foreach (Individual individual in this.individuals)
            {
                if (uuid.Equals(individual.Uri) == true)
                {
                    return individual;
                }
            }

            return null;
        }

        /// <summary>
        /// Gets property by given URI.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// </summary>
        /// <param name="uri">URI of the property.</param>
        /// <returns><see cref="PetrSU.SmartSlog.Property"/> 
        /// or null if there is no property with given name</returns>
        public Property GetPropertyByUri(string uri)
        {
            foreach (Property property in this.properties)
            {
                if (property.Uri.Equals(uri) == true)
                {
                    return property;
                }
            }

            return null;
        }

        #endregion

        /// <summary>
        /// Remove property from repository.
        /// </summary>
        /// <param name="property">property for removing from repository</param>
        /// <exception cref="System.InvalidOperationException">Thrown when a property is set for some individual.</exception>"
        public void RemoveEntity(Property property)
        {
            lock (repositoryLock)
            {
                // Check if the property is set for some individual.
                foreach (Individual individual in individuals)
                {
                    if (individual.GetProperty(property) != null)
                    {
                        throw new InvalidOperationException("The property cannot be removed, bacause it set for some individual.");
                    }
                }

                this.properties.Remove(property);
            }
        }

        /// <summary>
        /// Remove class from repository.
        /// </summary>
        /// <param name="ontologyClass">class for removing from repository</param>
        /// <exception cref="System.InvalidOperationException">Thrown when a class or sub-classes are have instances (individuals).</exception>"
        public void RemoveEntity(OntologyClass ontologyClass)
        {
            // Check if the class has instances.
            if (ontologyClass.IndividualCount > 0)
            {
                throw new InvalidOperationException("The class cannot be removed, bacause it or has instance (individual).");
            }

            lock (repositoryLock)
            {
                // Check if the sub-classes have instances.
                foreach (Individual individual in individuals)
                {
                    if (individual.ParentClass.GetSuperclasses().Count(sc => sc.Equals(ontologyClass)) > 0)
                    {
                        throw new InvalidOperationException("The class cannot be removed, bacause some of it's sub-classes have an instance (individual).");
                    }
                }

                this.classes.Remove(ontologyClass);
            }
        }

        /// <summary>
        /// Remove individual from repository.
        /// </summary>
        /// <param name="individual">individual for removing from repository</param>
        /// <exception cref="System.InvalidOperationException">Thrown when some individual has a link to current individual..</exception>"
        public void RemoveEntity(Individual individual)
        {
            lock (repositoryLock)
            {
                int linksCounter = (from i in this.individuals
                                    from p in i.GetProperties()
                                    where p == individual
                                    select p).Count();

                if (linksCounter > 0)
                {
                    throw new InvalidOperationException("The individual cannot be removed, bacause some other individual has a link to current individual.");
                }

                this.individuals.Remove(individual);
            }
        }

        /// <summary>
        /// Remove individual from repository.
        /// </summary>
        /// <param name="individual">individual for removing from repository</param>
        /// <exception cref="System.InvalidOperationException">Thrown when some individual has a link to current individual..</exception>"
        public void RemoveEntity(Individual individual, bool withLinks)
        {
            lock (repositoryLock)
            {
                foreach (Individual checkIndividual in this.individuals)
                {
                    var filteredProperties = checkIndividual.GetPropertiesValues().Where(pv => pv.IsObjectValue() == true
                        && pv.GetObjectValue().Equals(individual)).ToList();

                    if (filteredProperties.Count() > 0 && withLinks == false)
                    {
                        throw new InvalidOperationException("The individual cannot be removed, bacause some other individual has a link to current individual.");
                    }

                    foreach (PropertyValue pv in filteredProperties)
                    {
                       checkIndividual.RemoveProperty(pv.Property, individual);
                    }
                }

                this.individuals.Remove(individual);
            }
        }

        public void RemoveAllIndividuals()
        {
            lock (repositoryLock)
            {
                individuals.Clear();
            }
        }

        /// <summary>
        /// Check is entity in the repository.
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// </summary>
        /// <param name="property"><see cref="PetrSU.SmartSlog.Property"/> for check.</param>
        /// <returns>true if entity is found in the repository; otherwise, false.</returns>
        public bool IsEntityExists(Property property)
        {
            lock (repositoryLock)
            {
                return this.properties.Contains(property);
            }
        }

        /// <summary>
        /// Check is entity in the repository.
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        /// </summary>
        /// <param name="ontologyClass"><see cref="PetrSU.SmartSlog.OntologyClass"/> for check.</param>
        /// <returns>true if entity is found in the repository; otherwise, false.</returns>
        public bool IsEntityExists(OntologyClass ontologyClass)
        {
            lock (repositoryLock)
            {
                return this.classes.Contains(ontologyClass);
            }
        }

        /// <summary>
        /// Check is entity in the repository.
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// </summary>
        /// <param name="individual"><see cref="PetrSU.SmartSlog.Individual"/> for check.</param>
        /// <returns>true if entity is found in the repository; otherwise, false.</returns>
        public bool IsEntityExists(Individual individual)
        {
            lock (repositoryLock)
            {
                return this.individuals.Contains(individual);
            }
        }

        /// <summary>
        /// Create new <see cref="PetrSU.SmartSlog.Individual"/> by given <see cref="PetrSU.SmartSlog.OntologyClass"/>.
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        /// </summary>
        /// <param name="ontologyClass">Class of individual.</param>
        /// <returns>New individual or null if it can't find ontology class in repository.</returns>
        public Individual CreateIndividual(OntologyClass ontologyClass)
        {
            if (this.IsEntityExists(ontologyClass) == false)
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "ontologyClass");
            }

            Individual individual = new Individual(ontologyClass, session);
            
            lock (repositoryLock)
            {
                this.AddEntity(individual);
            }

            return individual;
        }

        /// <summary>
        /// Create new <see cref="PetrSU.SmartSlog.Individual"/> by given <see cref="PetrSU.SmartSlog.OntologyClass"/> and given UUID.
        /// <remarks>It tries to find individuals with given UUID, if individual found it returns it, if no then creates new indivdual.</remarks>
        /// </summary>
        /// <param name="ontologyClass">Class of individual.</param>
        /// <param name="uuid">Individual's UUID.</param>
        /// <returns>Old individual if it found in the repository, new created individual or null if it can't find ontology class in repository.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        public Individual CreateIndividual(OntologyClass ontologyClass, string uuid)
        {
            if (this.IsEntityExists(ontologyClass) == false)
            {
                Debug.WriteLine("Can't create individual: a class is not registered.");
                return null;
            }

            if (String.IsNullOrWhiteSpace(uuid))
            {
                Debug.WriteLine("Can't create individual: uuid is empty.");
                return null;
            }

            Individual individual = GetIndividual(uuid);

            if (individual == null)
            {
                individual = new Individual(ontologyClass, uuid, session);

                lock (repositoryLock)
                {
                    this.AddEntity(individual);
                }
            }

            return individual;
        }

        /// <summary>
        /// Get individual by UUID or create it using given parameters.
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        /// </summary>
        /// <remarks>This function try to find <see cref="PetrSU.SmartSlog.Individual"/> by given UUID and return it.
        /// It creates an <see cref="PetrSU.SmartSlog.Individual"/> by given class name with given UUID if it can not find it in repository.
        /// If it can't find <see cref="PetrSU.SmartSlog.OntologyClass"/> by given calss name it can't create new <see cref="PetrSU.SmartSlog.Individual"/>.
        /// </remarks>
        /// <param name="uuid">UUID of individual.</param>
        /// <param name="ontologyClassName">Name of class.</param>
        /// <returns><see cref="PetrSU.SmartSlog.Individual"/> or null if individual can not be created.</returns>
        public Individual GetOrCreateIndividual(string baseClassUri, string uuid)
        {
            lock (repositoryLock)
            {
                Individual individual = this.GetIndividual(uuid);

                if (individual == null)
                {
                    OntologyClass ontClass = this.GetClassByUri(baseClassUri);

                    if (ontClass == null)
                    {
                        return individual;
                    }
                    else
                    {
                        individual = new Individual(ontClass, uuid, session);
                        this.AddEntity(individual);
                    }
                }

                return individual;
            }
        }

        public Individual GetOrCreateIndividual(OntologyClass ontologyClass, string uuid)
        {
            lock (repositoryLock)
            {
                Individual individual = this.GetIndividual(uuid);

                if (individual != null)
                {
                    return individual;
                }

                if (IsEntityExists(ontologyClass) == false)
                {
                    return null;
                }

                individual = new Individual(ontologyClass, uuid, session);
                this.AddEntity(individual);

                return individual;
            }
        }

        #region Private methods
        /// <summary>
        /// Checks is there entity in repository.
        /// </summary>
        /// <typeparam name="T">Ontology entity: <see cref="OntologyClass"/>, <see cref="Property"/>, <see cref="Individual"/>.</typeparam>
        /// <param name="store">Repository entity store.</param>
        /// <param name="entity">Entity for check.</param>
        /// <returns><c>true</c> if entity is in repository or <c>false</c> otherwise.</returns>
        private bool IsExists<T>(List<T> store, T entity) where T : class
        {
            Debug.Assert(store != null, "Store is null.");
            Debug.Assert(entity != null, "Entity is null.");

            foreach (T storeEntity in store)
            {
                if (storeEntity == entity)
                {
                    return true;
                }
            }

            return false;
        }

        private PropertyReferncesCounter GetPropertyCounter(Property property, Individual individual)
        {
            return (from r in this.propertyReferences
                    where r.Individual.Equals(individual)
                    && r.Property.Equals(property)
                    select r).FirstOrDefault();
        }

        #endregion






    } // Class end
} // Namespace end
