﻿//  ------------------------------------------------------------------------
//  <copyright file="SemanticChecker.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Linq;

namespace PetrSU.SmartSlog
{
    internal static class SemanticChecker
    {

        public static bool IsMaxCardinalityViolation(Individual individual, Property property)
        {
            if (property.MaxCardinality < 0)
            {
                return false;
            }

            int count = (from p in individual.GetPropertiesValues() where p.Property == property select p).Count();

            if (count >= property.MaxCardinality)
            {
                return true;
            }

            return false;
        }

        public static bool IsMinCardinalityViolation(Individual individual, Property property)
        {
            if (property.MinCardinality < 0)
            {
                return false;
            }

            int count = (from p in individual.GetPropertiesValues() where p.Property == property select p).Count();

            if (count <= property.MinCardinality)
            {
                return true;
            }

            return false;
        }




        public static bool IsIndividualCanHaveProperty(Individual individual, Property property)
        {
            OntologyClass oc = individual.ParentClass;

            if (oc.Uri.Equals("http://www.w3.org/2002/07/owl#Thing")
                  || oc.Uri.Equals("http://www.w3.org/2002/07/owl#Nothing"))
            {
                return false;
            }

            if (oc.GetProperties().Contains(property))
            {
                return true;
            }
            
            return IsParentClassesHasProperty(oc, property);
        }

        private static bool IsParentClassesHasProperty(OntologyClass subClass, Property property)
        {
            foreach (OntologyClass oc in subClass.GetSuperclasses())
            {
                //// TODO: Codegen made incorrect creating of objects
                if (oc == null)
                {
                    continue;
                }

                if (oc.Uri.Equals("http://www.w3.org/2002/07/owl#Thing")
                    || oc.Uri.Equals("http://www.w3.org/2002/07/owl#Nothing"))
                {
                    return false;
                }

                if (oc.GetProperties().Contains(property))
                {
                    return true;
                }

                if (IsParentClassesHasProperty(oc, property))
                {
                    return true;
                }
            }

            return false;
        }

        public static void ChekPropertyUpdate(IRepository repository, Individual individual, Property property, object oldValue, object newValue)
        {
            if (!repository.IsEntityExists(individual))
            {
                throw ExceptionManager.CreateArgumentNotRegisteredException("individual", individual);
            }

            if (!repository.IsEntityExists(property))
            {
                throw ExceptionManager.CreateArgumentNotRegisteredException("property", property);
            }

            if (!SemanticChecker.IsIndividualCanHaveProperty(individual, property))
            {
                throw new ArgumentException(
                    String.Format(ExceptionManager.IncorrectPropertyForIndividual, property.Uri),
                    "property");
            }

            if (property.IsObject())
            {
                if (!repository.IsEntityExists(oldValue as Individual))
                {
                    throw ExceptionManager.CreateArgumentNotRegisteredException("oldValue", oldValue as Individual);
                }

                if (!repository.IsEntityExists(newValue as Individual))
                {
                    throw ExceptionManager.CreateArgumentNotRegisteredException("oldValue", newValue as Individual);
                }
            }
            else
            {
                string value = oldValue as string;

                if (value == null)
                {
                    throw new ArgumentNullException("oldValue");
                }

                value = newValue as string;

                if (value == null)
                {
                    throw new ArgumentNullException("newValue");
                }
            }


        }

        public static void CheckPropertyGet(IRepository repository, Individual individual, Property property)
        {
            if (!repository.IsEntityExists(individual))
            {
                throw ExceptionManager.CreateArgumentNotRegisteredException("individual");
            }

            if (!repository.IsEntityExists(property))
            {
                throw ExceptionManager.CreateArgumentNotRegisteredException("property");
            }

            if (!SemanticChecker.IsIndividualCanHaveProperty(individual, property))
            {
                throw new ArgumentException(
                    String.Format(ExceptionManager.IncorrectPropertyForIndividual, property.Uri),
                    "property");
            }

        }


        public static void CheckPropertySet(IRepository repository, Individual individual, Property property, object value)
        {
            if (!repository.IsEntityExists(individual))
            {
                string uri = (individual == null) ? "null" : individual.Uri;

                throw new ArgumentException(
                    String.Format(ExceptionManager.EntityUriNotRegisteredMessage, uri),
                    "individual");
            }

            if (!repository.IsEntityExists(property))
            {
                string uri = (property == null) ? "null" : property.Uri;
                throw new ArgumentException(
                    String.Format(ExceptionManager.EntityUriNotRegisteredMessage, uri),
                    "property");
            }

            if (!IsIndividualCanHaveProperty(individual, property))
            {
                throw new ArgumentException(
                    String.Format(ExceptionManager.IncorrectPropertyForIndividual, property.Uri),
                    "property");
            }

            if (property.IsObject() && !repository.IsEntityExists(value as Individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "value");
            }

            if (!property.IsObject() && (value as string == null))
            {
                throw new ArgumentException(ExceptionManager.IncorrectPropertyValueMessage, "value");
            }



        }


        public static void CheckPropertyRemove(IRepository repository, Individual individual, Property property, object value)
        {
            if (!repository.IsEntityExists(individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            }

            if (!repository.IsEntityExists(property))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "property");
            }

            if (property.IsObject() && !repository.IsEntityExists(value as Individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "value");
            }

            if (!property.IsObject() && (value as string == null))
            {
                throw new ArgumentException(ExceptionManager.IncorrectPropertyValueMessage, "value");
            }
        }

        public static void CheckPropertyRemoveLocal(IRepository repository, Individual individual, Property property, object value)
        {
            if (!repository.IsEntityExists(individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "individual");
            }

            if (!repository.IsEntityExists(property))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "property");
            }

            if (property.IsObject() && !repository.IsEntityExists(value as Individual))
            {
                throw new ArgumentException(ExceptionManager.EntityNotRegisteredMessage, "value");
            }

            if (!property.IsObject() && (value as string == null))
            {
                throw new ArgumentException(ExceptionManager.IncorrectPropertyValueMessage, "value");
            }
        }
    }
}
