﻿//  ------------------------------------------------------------------------
//  <copyright file="Session.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The Session class represents a session between KP and smart spaces. The <see cref="Node"/> class is used as SSAP session. 
    /// This class manipulates with many nodes and SSAP sessions. Each node in the session store data to session and data share between 
    /// all nodes in the session.
    /// To work with the objects (calsses, properties, nodes) you need them in the session.
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Node"/>
    /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
    /// <seealso cref="PetrSU.SmartSlog.Property"/>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/>
    internal class Session : ISession
    {
        /// <summary>
        /// The name of the KPI library, that is used as a default library when session has many KPIs.
        /// </summary>
        private const string DefaultKpiName = "ckpi";

        /// <summary>
        /// Identifier of the session.
        /// </summary>
        private string id = String.Empty;

        /// <summary>
        /// Nodes that are registered in the session.
        /// </summary>
        internal List<Node> nodes = new List<Node>();

        /// <summary>
        /// Classes List of registered ontology
        /// </summary>
        internal List<OntologyClass> storedClasses = new List<OntologyClass>();

        /// <summary>
        /// Properties List of registered ontology
        /// </summary>
        internal List<Property> storedProperties = new List<Property>();

        private List<Triple> storedTriples = new List<Triple>();

        /// <summary>
        /// KPI loaders to prepare KPI libraries to work.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.KpiWrapper.IKpi"/>
        /// <seealso cref="PetrSU.SmartSlog.KpiWrapper.IKpiLoader"/>
        //private List<IKpiLoader> kpiLoaders;

        /// <summary>
        /// Initializes a new instance of the Session with identifier.
        /// </summary>
        internal Session(string id)
        {
            //kpiLoaders = KpiFinder.GetKpiLoaders();
            //this.id = id;
            //repository = new Repository(this);
        }

        /// <summary>
        /// Gets the session identifier.
        /// </summary>
        public string Id
        {
            get { return id; }
        }

        /// <summary>
        /// Creates a KPI. It tries to find and instantiates the C_KPI wrapper, 
        /// otherwise it instantiates first KPI, that was found.
        /// </summary>
        /// <returns>The interface to work with the KPI.</returns>
        /// <exception cref="System.InvalidOperationException">Thrown when can't find KPIs.</exception>
        //internal IKpi CreateKpi()
        //{
        //    if (kpiLoaders.Count == 0)
        //    {
        //        throw new InvalidOperationException("Can't create a KPI. KPIs are not found.");
        //    }

        //    var kpi = (from l in kpiLoaders where DefaultKpiName.Equals(l.Name) select l).FirstOrDefault();

        //    if (kpi != null)
        //    {
        //        return kpi.CreateKpi();
        //    }

        //    return kpiLoaders.First().CreateKpi();
        //}


        /// <summary>
        /// Registers the node in the session.
        /// </summary>
        /// <param name="node">Node to register</param>
        /// <returns>Identifier if node's session (not used now).</returns>
        /// <exception cref="System.InvalidOperationException">Thrown when node is already registered.</exception>
        internal string RegisterNode(Node node)
        {
            if (nodes.Where(n => n.Id.Equals(node.Id)).Count() > 0)
            {
                throw new InvalidOperationException("Session already has a node with the same identifier: " + node.Id);
            }

            nodes.Add(node);

            return node.Id;
        }

        /// <summary>
        /// Gets a repository of the session.
        /// </summary>
        /// DELETE TOTALY
        internal IRepository Repository { get { return null; } }
        
        /// <summary>
        /// Registers a new class in the session. 
        /// It does nothing if the ontology class with the same URI is already registered.
        /// </summary>
        /// <param name="ontologyClass">New ontology class to registartion.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the ontology class is null.</exception>
        public void RegisterEntity(OntologyClass ontologyClass)
        {
            if (ontologyClass == null)
            {
                throw new ArgumentNullException("ontologyClass");
            }

            ontologyClass.NativePtr = NativeMethods.sslog_new_class(ontologyClass.Uri);

            storedClasses.Add(ontologyClass);
            //Class_S new_class = new Class_S(ontologyClass.NativePtr);

            //if (!repository.IsEntityExists(ontologyClass))
            //{
            //    repository.AddEntity(ontologyClass);
            //}
        }

        public void RegisterEntity(Property ontologyProperty)
        {
            if (ontologyProperty == null)
            {
                throw new ArgumentNullException("ontologyProperty");
            }

            ontologyProperty.NativePtr = NativeMethods.sslog_new_property(ontologyProperty.Uri, ontologyProperty.IsObject() ? 1 : 2);

            storedProperties.Add(ontologyProperty);
            //if (!repository.IsEntityExists(ontologyProperty))
            //{
            //    repository.AddEntity(ontologyProperty);
            //}
        }

        public  Triple CreateTriple(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType)
        {
            Triple triple = new Triple(subject, predicate, objectElement, subjectType, objectType);  

            foreach (Triple storedTriple in storedTriples)
            {
                // Check if the triple is realy new or not.
                if(storedTriple.NativePtr == triple.NativePtr) {
                    triple.NativePtr = IntPtr.Zero;
                    return storedTriple;
                }
            }

            return triple;
        }

        public void UnregisterTriple(Triple triple)
        {
            storedTriples.Remove(triple);
        }

        public  Triple CreateDetachedTriple(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType)
        {
            return new Triple(subject, predicate, objectElement, subjectType, objectType, true);
        }

        public void UnregisterNode(Node node)
        {
            if (node == null)
            {
                throw new ArgumentNullException("node");
            }

            if (!this.nodes.Contains(node))
            {
                throw new InvalidOperationException("The node dosen't registered in the current session.");
            }

            nodes.Remove(node);
            node.Destroy();
        }

        /// <summary>
        /// Removes ontologyClass from local store
        /// </summary>
        /// <param name="ontologyClass">OntologyClass to remove</param>
        private void UnregisterEntity(OntologyClass ontologyClass)
        {
            if (ontologyClass == null)
            {
                throw new ArgumentNullException("ontologyClass");
            }

            //// TODO: Check instances of for the low classes.
            if (ontologyClass.IndividualCount > 0)
            {
                throw new InvalidOperationException("You can not unregister a class while it has instances (individuals)");
            }

            //this.repository.RemoveEntity(ontologyClass);
        }

        /// <summary>
        /// Removes property from local store
        /// </summary>
        /// <param name="property">Property to remove</param>
        private void UnregisterEntity(Property property)
        {
            if (property == null)
            {
                throw new ArgumentNullException("property");
            }

            //repository.RemoveEntity(property);
        }

        /// <summary>
        /// Removes individual from local store
        /// </summary>
        /// <param name="individual">Individual to remove</param>
        public void UnregisterEntity(Individual individual)
        {
            int result = NativeMethods.sslog_remove_individual(individual.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }
        }

        /// <summary>
        /// Removes individual from local store with links
        /// </summary>
        /// <param name="individual">Individual to remove</param>
        private void UnregisterEntity(Individual individual, bool withLinks)
        {
            if (individual == null)
            {
                return;
            }

            //repository.RemoveEntity(individual, withLinks);
            individual.ResetSession();
        }

        /// <summary>
        /// Removes triple from local store
        /// </summary>
        /// <param name="triple">Triple to remove</param>
        public void RemoveTriple(Triple triple)
        {
            if (triple == null)
            {
                throw new ArgumentNullException("triple");
            }

            storedTriples.Remove(triple);

            NativeMethods.sslog_remove_triple(triple.NativePtr);
            
            //triple.NativePtr = IntPtr.Zero;
        }

        internal void Destroy()
        {
            foreach (Node node in this.nodes)
            {
                node.Destroy();
            }

            this.id = null;
            this.nodes = null;
            this.storedClasses = null;
            this.storedTriples = null;
            //this.repository = null;
            //this.kpiLoaders = null;
        }
    }
}
