﻿//  ------------------------------------------------------------------------
//  <copyright file="SessionManager.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// This class is used to create and get sessions.
    /// There is one default session, it is used for node's default session.
    /// You can create a new session and use it for nodes.
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Node"/>
    /// <seealso cref="PetrSU.SmartSlog.ISession"/>
    public class SessionManager
    {
        /// <summary>
        /// Prefix for the session identifier.
        /// </summary>
        private const string SessionIdPrefix = "SSLOG_SESSID:";

        /// <summary>
        /// Default session identifier.
        /// </summary>
        private const string DefaultSessionId = SessionIdPrefix + "default";

        /// <summary>
        /// The default session is set for each node that are created without session.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.Node"/>
        private static Session defaultSession;

        /// <summary>
        /// Field is used to create a new session identifier.
        /// </summary>
        private static int nextSessionId;

        //private static bool isSslogInit = false;

        /// <summary>
        /// List of all sessions that were created.
        /// </summary>
        private static List<Session> sessions = new List<Session>();

        /// <summary>
        /// Initializes a new instance of the SessionManager class.
        /// </summary>
        protected SessionManager()
        {
        }

        /// <summary>
        /// Gets the default session.
        /// </summary>
        /// <returns>Default session.</returns>
        /// <seealso cref="PetrSU.SmartSlog.ISession"/>
        public static ISession GetDefaultSession()
        {
            if (defaultSession == null)
            {
                defaultSession = new Session(DefaultSessionId);
            }

            return defaultSession;
        }

        /// <summary>
        /// Gets all session that were created without default session.
        /// </summary>
        /// <returns>Created sessions.</returns>
        /// <seealso cref="PetrSU.SmartSlog.ISession"/>
        public static ReadOnlyCollection<ISession> GetSessions()
        {
#if (PORTABLE || WINDOWS_PHONE)
            IList<ISession> isessions = new List<ISession>(sessions.Count);

            foreach (ISession s in sessions)
            {
                isessions.Add(s);
            }

            return new ReadOnlyCollection<ISession>(isessions);
#else
            return new ReadOnlyCollection<ISession>(sessions.AsEnumerable<ISession>().ToList());    
#endif
        }

        /// <summary>
        /// Creates a new session.
        /// </summary>
        /// <returns>New created session.</returns>
        /// <seealso cref="PetrSU.SmartSlog.ISession"/>
        public static ISession CreateSession()
        {
            ++nextSessionId;
            Session session = new Session(SessionIdPrefix + nextSessionId.ToString());

            sessions.Add(session);

            return session;
        }

        /// <summary>
        /// Destroys the session. After this method the session cannot be used and all nodes will be unregistered and cannot be used too.
        /// </summary>
        /// <param name="session">Session to destroy.</param>
        /// <exception cref="System.ArgumentException">Thrown when the session has incorrect type or null.</exception>
        public static void DestroySession(ISession session)
        {
            if (session.Id.Equals(DefaultSessionId))
            {
                defaultSession = null;
            }

            Session destroySession = session as Session;

            if (destroySession == null)
            {
                throw new ArgumentException("session", "Session has an incorrect type.");
            }

            destroySession.Destroy();

            sessions.Remove(destroySession);
        }
    }
}
