﻿//  ------------------------------------------------------------------------
//  <copyright file="Subscriber.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// This class work with subscription session. 
    /// <remarks>Subscription can be asynchronous and synchronous. 
    /// You need to create <see cref="PetrSU.SmartSlog.Subscription"/>, add <see cref="PetrSU.SmartSlog.Individual"/> and list of
    /// <see cref="PetrSU.SmartSlog.Property"/> to the session, set type of subscripion and subscribe subscription. 
    /// When the subscription is no longer needed unregistered subscription.</remarks> 
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.Individual"/>
    /// <seealso cref="PetrSU.SmartSlog.Property"/>
    /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
    internal class Subscriber
    {
        #region Private fields

        /// <summary>
        /// Name of the class for debug output.
        /// </summary>
        private const string DebugName = "Subscriber ";

        /// <summary>
        /// Name of thread processing asynchronous subscription.
        /// </summary>
        private const string ThreadName = "SmartSlog.AsyncSubscriptions";

        /// <summary>
        /// Default timeout(milliseconds) to receive notification for the subscription. 
        /// </summary>
        private readonly int indicationTimeout = 1000;

        /// <summary>
        /// Thread for asynchronous subscription.
        /// </summary>
        private Thread subscribeThread;

        private bool isAsyncProcessStopped;

        /// <summary>
        /// Represents a lock that is used to manage access to a resource, allowing multiple threads for reading or exclusive access for writing. 
        /// </summary>
        //// private ReaderWriterLockSlim subscriptionSessionsLock = new ReaderWriterLockSlim();

        /// <summary>
        /// Object is used to lock multiple acces to subscripitons.
        /// </summary>
        private object subscriptionSessionsLock = new object();

        /// <summary>
        /// KPI_Low wrapper, it implements access to the smart space (SSAP).
        /// </summary>
        //private IKpi kpi;

        private Session session;

        /// <summary>
        /// List with containers that are subscribed asynchronously.
        /// </summary>
        private List<Subscription> asyncSessions = new List<Subscription>();

        /// <summary>
        /// List with containers that are subscribed synchronously.
        /// </summary>
        private List<Subscription> syncSessions = new List<Subscription>();

        #endregion

        /// <summary>
        /// Initializes a new instance of the Subscriber class.
        /// Indication timeout determines time to wait new data for the subscription.  
        /// When you wait a synchronous subscription with timeout you need to use value that is greater or equal than the indication 
        /// timeout.
        /// Asynchronous subscriptions are processed sequentially and each subscription is waiting new data for this time.
        /// </summary>
        /// <param name="session">Session of the subscriber.</param>
        /// <param name="kpi">KPI to work with the smart space.</param>
        /// <param name="subscriptionIndicationTimout">Indication timeout determines time to check new data for the subscription. 
        /// Default value is 500 milliseconds, minimal is 10 milliseconds.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when indication timeout parameter is to small.</exception>
        internal Subscriber(Session session)
        {
            this.session = session;
          //  this.kpi = kpi;
        }

        /// <summary>
        /// Initializes a new instance of the Subscriber class.
        /// Indication timeout determines time to wait new data for the subscription.  
        /// When you wait a synchronous subscription with timeout you need to use value that is greater or equal than the indication 
        /// timeout.
        /// Asynchronous subscriptions are processed sequentially and each subscription is waiting new data for this time.
        /// </summary>
        /// <param name="session">Session of the subscriber.</param>
        /// <param name="kpi">KPI to work with the smart space.</param>
        /// <param name="subscriptionIndicationTimout">Indication timeout determines time to check new data for the subscription. 
        /// Default value is 1000 milliseconds, minimal is 10 milliseconds.</param>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when indication timeout parameter is to small.</exception>
        internal Subscriber(Session session, IKpi kpi, int subscriptionIndicationTimout)
        {
            if (subscriptionIndicationTimout < 10)
            {
                throw new ArgumentOutOfRangeException("subscriptionIndicationTimout", "Timeout is too small.");
            }

            this.session = session;
            this.indicationTimeout = subscriptionIndicationTimout;
            
        }

        #region Public methods

        /// <summary>
        /// Subscribe session. It does nothing if the subscription is already active or has no data to subscribe.
        /// </summary>
        /// <param name="subscription">Session to subscribe.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when KPI is in inconsistent state.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        public void Subscribe(Subscription subscription)
        {
            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            if (subscription.IsActive == true)
            {
                LoggerManager.WriteMessage(LoggerManager.SubscripitonManagement, "Subscribe", "Subscription with id='" + subscription.Id + "' is already subscribed");
                return;
            }

            List<ITriple> subscribeTriples = new List<ITriple>();
            List<ITriple> resultTriples = new List<ITriple>();

            //OwlMapping.SubscriptionToTriples(kpi, subscription, subscribeTriples);

            if (subscribeTriples.Count == 0)
            {
                //LoggerManager.WriteMessage(
                //    LoggerManager.SubscripitonManagement,
                //    "Subscribe",
                //    "Subscription with id='" + subscription.Id + "' has no data to subscribe.");
                //return;
            }

            if (subscription.IsAsynchronous)
            {
                IntPtr intptrDelegate = Marshal.GetFunctionPointerForDelegate(subscription._subscriptionHandler);

                NativeMethods.sslog_sbcr_set_changed_handler(subscription.NativeSubscription, subscription._subscriptionHandler);
            }

            int result = NativeMethods.sslog_sbcr_subscribe(subscription.NativeSubscription);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            this.RegisterSubscription(subscription);
        }

        /// <summary>
        /// Waits new data for the given subscription.
        /// This method is used for synchronous subscriptions. 
        /// It blocks current process and waits data from the smart space. It does not blocks asynchronous subcription thread.
        /// For unsubscribed containers and asynchronous containers it does nothing.
        /// If the value of timeout is less then 0, then it is considered as an infinite timeout. 
        /// </summary>
        /// <returns>true if an indication is received, otherwise (session is inactive, asynchronous or timeout is over) false.</returns>
        /// <param name="subscription">Session that will be used for a waiting indication.</param>
        /// <param name="millisecondTimeout">Now use it as a number of attempts: number * indication timeout.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when KPI is in inconsistent state.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        //public bool WaitSubscribe(Subscription subscription, int millisecondTimeout)
        //{
        //    const string MethodName = "WaitSubscribe"; // Only for debug, without runtime getting to support WP
        //    const int ManagementType = LoggerManager.SubscripitonManagement; // Only for debugging, type of the function

        //    if (subscription == null)
        //    {
        //        throw new ArgumentNullException("subscription");
        //    }

        //    if (subscription.IsActive == false)
        //    {
        //        LoggerManager.WriteMessage(ManagementType, MethodName, "Subscription is not subscribed.");
        //        return false;
        //    }

        //    if (subscription.IsAsynchronous == true)
        //    {
        //        LoggerManager.WriteMessage(ManagementType, MethodName, String.Format("Subscription({0}) is asyncronous.", subscription.Id));
        //        return false;
        //    }

        //    //// Call infinity version of 'Wait' method.
        //    if (millisecondTimeout < 0)
        //    {
        //        return WaitSubscribe(subscription);
        //    }

        //    //// Set status  and subscription's isStopped status to start checking indication. 
        //    SubscriptionStatus status = SubscriptionStatus.Timeout;
        //    subscription.isStopped = false;

        //    //// Indications will be checking while:
        //    //// - subscription has correct state; 
        //    //// - indication status is timeout;
        //    //// - timeout is not over.
        //    while (status == SubscriptionStatus.Timeout
        //        && !subscription.isStopped && subscription.IsActive)
        //    {
        //        status = ProcessSubscription(subscription);

        //        //// Decrease timeout. 
        //        millisecondTimeout -= indicationTimeout;

        //        if (millisecondTimeout < indicationTimeout)
        //        {
        //            break;
        //        }
        //    }

        //    LoggerManager.WriteMessage(ManagementType, MethodName, String.Format("Subscription({0}) has status: {1}", subscription.Id, status));
        //    return (status == SubscriptionStatus.Indication) ? true : false;
        //}

        /// <summary>
        /// Waits new data for the given subscription.
        /// This method is used for synchronous subscriptions. 
        /// It blocks current process and waits data from the smart space. It does not blocks asynchronous subcription thread.
        /// For unsubscribed containers and asynchronous containers it does nothing.
        /// </summary>
        /// <returns>true if an indication is received, otherwise (session is inactive, asynchronous or timeout is over) false.</returns>
        /// <param name="subscription">Synchronous subscription that will be used for a waiting indication.</param>
        /// <param name="millisecondTimeout">Now use it as a number of attempts: number * indication timeout.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when subscription is null.</exception>
        /// <exception cref="System.InvalidOperationException">Thrown when KPI is in inconsistent state.</exception>
        /// <exception cref="System.IO.IOException">Thrown when there is a problem with a smart space access.</exception>
        /// <exception cref="PetrSU.SmartSlog.KpiWrapper.SmartSpaceException">Thrown when the smart space returns an error.</exception>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        public bool WaitSubscribe(Subscription subscription)
        {
            const string MethodName = "WaitSubscribe"; // Only for debug, without runtime getting to support WP
            const int ManagementType = LoggerManager.SubscripitonManagement; // Only for debugging, type of the function

            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            if (subscription.IsActive == false)
            {
                LoggerManager.WriteMessage(ManagementType, MethodName, "Subscription is not subscribed.");
                return false;
            }

            if (subscription.IsAsynchronous == true)
            {
                LoggerManager.WriteMessage(ManagementType, MethodName, String.Format("Subscription({0}) is asyncronous.", subscription.Id));
                return false;
            }

            //SubscriptionStatus status = SubscriptionStatus.Timeout;

            subscription.isStopped = false;
            
            int result = NativeMethods.sslog_sbcr_wait(subscription.NativeSubscription);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                throw ExceptionManager.GetExeption(result);
            }

            IntPtr sslog_changes = NativeMethods.sslog_sbcr_get_changes_last(subscription.NativeSubscription);

            IndicationChangesData changes = new IndicationChangesData(sslog_changes);

            subscription.AddChanges(changes);

            return true;

            //// Indications will be checking while:
            //// - subscription has correct state; 
            //// - indication status is timeout.
            //while (status == SubscriptionStatus.Timeout
            //    && !subscription.isStopped && subscription.IsActive)
            //{
            //    status = ProcessSubscription(subscription);
            //}

            //if (status == SubscriptionStatus.Error)
            //{
            //    SubscriptionEventArgs ea = new SubscriptionEventArgs();
            //    subscription.FireErrorEvent(ea);
            //    if (ea.Unsubscribe)
            //    {
            //        Unsubscribe(subscription);
            //    }
            //}

            //LoggerManager.WriteMessage(ManagementType, MethodName, String.Format("Subscription({0}) has status: {1}", subscription.Id, status));
            //return (status == SubscriptionStatus.Indication) ? true : false;
        }

        /// <summary>
        /// Indicates what checking of the subscription must be stopped.
        /// This method works only with synchronous subscriptions, for asynchronous or inactive subscriptions it does nothing.
        /// If you want to stop waiting, then you call this method.
        /// It only indicates, but not really stops checking, you need to wait while
        /// the <see cref="WaitSubscribe"/> ends, and then make other actions under subscription.
        /// </summary>
        /// <param name="subscription">Subscription to stop waiting.</param>
        public void StopWaiting(Subscription subscription)
        {
            const string MethodName = "StopWaiting"; // Only for debug, without runtime getting to support WP
            const int ManagementType = LoggerManager.SubscripitonManagement; // Only for debugging, type of the function

            LoggerManager.WriteStartMessage(ManagementType, MethodName);

            if (subscription == null)
            {
                LoggerManager.WriteEndMessage(ManagementType, MethodName, "Subscription is null.");
                return;
            }

            if (subscription.IsActive == false)
            {
                LoggerManager.WriteEndMessage(ManagementType, MethodName, String.Format("Subscription({0}) is inactive", subscription.Id));
                return;
            }

            if (subscription.IsAsynchronous == true)
            {
                LoggerManager.WriteEndMessage(ManagementType, MethodName, String.Format("Subscription({0}) is asynchronous", subscription.Id));
                return;
            }

            subscription.isStopped = true;

            LoggerManager.WriteEndMessage(ManagementType, MethodName, String.Format("Subscription({0}) has status as stopped.", subscription.Id));
        }

        /// <summary>
        /// Unsubscribe the subscripiton. After this operation a session state will be inactive.
        /// You need to subscribe it again if you want to get notification for the subscription.
        /// If the subscription has an inactive state, then the method does nothing.
        /// </summary>
        /// <param name="subscription">Subscripition to unsubscribe.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when subscription is null.</exception>
        public void Unsubscribe(Subscription subscription)
        {
            if (subscription == null)
            {
                throw new ArgumentNullException("subscription");
            }

            if (subscription.IsActive == false)
            {
                return;
            }

            //ISubscriptionInfo info = subscription.SubscribeInfo;

            Debug.WriteLine(DebugName + "(Unsubscribe): " + subscription.Id);

            //KpiHelper.DoKpiUnsubscribe(kpi.Unsubscribe, info);
            NativeMethods.sslog_sbcr_unsubscribe(subscription.NativeSubscription);

            // If the Unsubscribe packet did not received, and subscription is in active state, 
            // then unregister the subscription manually.
            if (subscription.IsActive)
            {
                UnregisterSession(subscription);
            }

            return;

            //// TODO: need to make field
            //int maxAttempts = 5;

            //// Wait for SIB unsubscribe packet giving some attempts to receive it.
            //if (subscription.IsAsynchronous)
            //{
            //    while (subscription.IsActive == true && maxAttempts > 0)
            //    {
            //        Thread.Sleep(indicationTimeout);
            //        --maxAttempts;
            //    }
            //}
            //else
            //{
            //    WaitSubscribe(subscription, indicationTimeout * maxAttempts);
            //}
        }

        /// <summary>
        /// Unsubscribes all subscriptions (synchronous and asynchronous).
        /// </summary>
        public void UnsubscribeAll()
        {
            while (true)
            {
                Subscription subscripiton = null;

                lock (this.subscriptionSessionsLock)
                {
                    subscripiton = asyncSessions.FirstOrDefault();
                }

                if (subscripiton == null)
                {
                    break;
                }

                Unsubscribe(subscripiton);

            }

            while (true)
            {
                Subscription subscription = null;

                lock (syncSessions)
                {
                    subscription = syncSessions.FirstOrDefault();
                }

                if (subscription == null)
                {
                    break;
                }

                Unsubscribe(subscription);
            }
        }

        //// TODO: to session/repository

        /// <summary>
        /// Checks is an individual is subscribed. 
        /// Try to finds an individual in active sessions and return search result.
        /// </summary>
        /// <param name="individual">Individual for checking.</param>
        /// <returns>true if individual is found; otherwise, false.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        public bool IsSubscribed(Individual individual)
        {
            if (this.FindSubscribedDataBy(individual) == null)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Checks is property for given individual is subscribed.
        /// <remarks>Try to find individual and property in subscribed containers and return search result.</remarks>
        /// </summary>
        /// <param name="individual">Individual for checking.</param>
        /// <param name="property">Property for checking.</param>
        /// <returns>true if property for individual is found; otherwise, false.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        public bool IsSubscribed(Individual individual, Property property)
        {
            if (this.FindSubscribedDataBy(individual, property) == null)
            {
                return false;
            }

            return true;
        }

        #endregion // Public methods

        /// <summary>
        /// Checks is given property subscribed for given individual.
        /// </summary>
        /// <param name="individual">Individual to check.</param>
        /// <param name="property">Property to check.</param>
        /// <returns>true if the property was subscribed for the given individual, otherwise false, it also returns false then some argument is null.</returns>
        internal bool IsPropertySubscribe(Individual individual, Property property)
        {
            if (individual == null || property == null)
            {
                return false;
            }

            if (FindSubscribedDataBy(individual, property) != null)
            {
                return true;
            }

            return false;
        }

        #region Private methods

        /// <summary>
        /// Prepare data values for property. 
        /// </summary>
        /// <param name="newTriple">New triples that represents property value.</param>
        /// <param name="oldTriple">Old triples that represents property value.</param>
        /// <param name="newValue">Prepared new value of property.</param>
        /// <param name="oldValue">Prepared old value of property.</param>
        private static void PrepareDataValues(ITriple newTriple, ITriple oldTriple, out object newValue, out object oldValue)
        {
            newValue = null;
            oldValue = null;

            if (newTriple != null)
            {
                newValue = newTriple.ObjectElement;
            }

            if (oldTriple != null)
            {
                oldValue = oldTriple.ObjectElement;
            }
        }

        /// <summary>
        /// Registers the subscription. Registration is used to track subscribed(active) sessions.
        /// This method set session as active and store it in an interanl store.
        /// </summary>
        /// <param name="subscription">Subscripition to register.</param>
        /// <param name="subscribeInfo">Subscription information that will be set for a session.</param>
        private void RegisterSubscription(Subscription subscription)
        {
            Debug.Assert(subscription != null, "Subscription cannot be null.");

            subscription.IsActive = true;

            if (subscription.IsAsynchronous)
            {

                lock (this.subscriptionSessionsLock)
                {
                    this.asyncSessions.Add(subscription);
                }

                //this.StartAsyncSessionProcessing();
            }
            else
            {
                lock (this.syncSessions)
                {
                    this.syncSessions.Add(subscription);
                }
            }
        }

        /// <summary>
        /// Unregister a subscription.
        /// </summary>
        /// <param name="subscription">Subscription to unregister.</param>
        private void UnregisterSession(Subscription subscription)
        {
            Debug.Assert(subscription != null, "Subscription cannot be null.");

            if (subscription.IsAsynchronous == true)
            {
                lock (this.subscriptionSessionsLock)
                {
                    this.asyncSessions.Remove(subscription);
                    subscription.IsActive = false;

                }
            }
            else
            {
                lock (this.syncSessions)
                {
                    this.syncSessions.Remove(subscription);
                    subscription.IsActive = false;
                }
            }

            subscription.SubscribeInfo = null;
        }

        /// <summary>
        /// Starts thread for checking asynchronous subscription.
        /// </summary>
        //private void StartAsyncSessionProcessing()
        //{
        //    if (this.subscribeThread != null && this.subscribeThread.IsAlive == true)
        //    {
        //        return;
        //    }

        //    this.subscribeThread = new Thread(this.ProcessAsyncSubscriptions);

        //    this.subscribeThread.Name = ThreadName;

        //    isAsyncProcessStopped = false;

        //    this.subscribeThread.Start();
        //}

        //private void StopAsyncSessionProcessing()
        //{
        //    if (this.subscribeThread != null && this.subscribeThread.IsAlive == true)
        //    {
        //        isAsyncProcessStopped = false;
        //        this.subscribeThread.Join();
        //        this.subscribeThread = null;
        //    }
        //}

        /// <summary>
        /// Process asynchronous subscription in another thread.
        /// </summary>
        //private void ProcessAsyncSubscriptions()
        //{
        //    const string MethodName = "ProcessAsyncSubscriptions"; // Only for debug, without runtime getting to support WP
        //    const int ManagementType = LoggerManager.SubscripitonManagement; // Only for debugging, type of the function

        //    LoggerManager.WriteStartMessage(ManagementType, MethodName);

        //    int index = 0;
        //    while (isAsyncProcessStopped == false)
        //    {
        //        SubscriptionStatus status = SubscriptionStatus.Timeout;
        //        Subscription subscription = null;

        //        lock (this.subscriptionSessionsLock)
        //        {
        //            if (this.asyncSessions.Count <= 0)
        //            {
        //                break;
        //            }
        //            else if (index >= this.asyncSessions.Count)
        //            {
        //                index = 0;
        //            }

        //            subscription = this.asyncSessions[index];

        //            status = ProcessSubscription(subscription);

        //        }

        //        if (status == SubscriptionStatus.Error)
        //        {
        //            SubscriptionEventArgs ea = new SubscriptionEventArgs();
        //            subscription.FireErrorEvent(ea);

        //            if (ea.Unsubscribe)
        //            {
        //                Unsubscribe(subscription);
        //            }
        //        }
        //        else
        //        {
        //            ++index;
        //        }

        //    }

        //    LoggerManager.WriteEndMessage(ManagementType, MethodName);
        //}

        /// <summary>
        /// Process the subscription, it start indication checking and handle result of this operation.
        /// </summary>
        /// <param name="subscription">Subscripition that will be process.</param>
        /// <returns>Status of the processing.</returns>
        //private SubscriptionStatus ProcessSubscription(Subscription subscription)
        //{
        //    const string MethodName = "ProcessAsyncSubscriptions"; // Only for debug, without runtime getting to support WP
        //    const int ManagementType = LoggerManager.SubscripitonManagement; // Only for debugging, type of the function

        //    LoggerManager.WriteStartMessage(ManagementType, MethodName, "SubscriptionId=" + subscription.Id);

        //    SubscriptionStatus status = SubscriptionStatus.Error;// KpiHelper.DoKpiCheckSubscription(kpi.CheckSubscription, subscription.SubscribeInfo, indicationTimeout);

        //    //switch (status)
        //    //{
        //    //    // Indication:  1
        //    //    case SubscriptionStatus.Indication:

        //    //        IndicationChangesData changes = this.UpdateSubscription(subscription);
        //    //        break;

        //    //    // Unsubscribe: 2
        //    //    case SubscriptionStatus.Unsubscribed:
        //    //        LoggerManager.WriteMessage(ManagementType, MethodName, "Unsibscribe indication for subscriptionId=" + subscription.Id);
        //    //        UnregisterSession(subscription);
        //    //        break;

        //    //    // Timeout:     0
        //    //    case SubscriptionStatus.Timeout: break;

        //    //    // Error:      -1
        //    //    case SubscriptionStatus.Error:
        //    //        // Error event is fired by calles method                    
        //    //        break;
        //    //}

        //    //if (status == SubscriptionStatus.Timeout)
        //    //{
        //    //    try
        //    //    {
        //    //        //kpi.Query(
        //    //            //new List<ITriple>() { kpi.CreateTriple("TEST", "NETWORK", "STATE", RdfType.Uri, RdfType.Literal)}, new List<ITriple>());
        //    //    }
        //    //    catch(Exception ex)
        //    //    {
        //    //        status = SubscriptionStatus.Error;
        //    //    }
        //    //}

        //    LoggerManager.WriteStartMessage(ManagementType, MethodName, String.Format("SubscriptionId={0}; status={1}", subscription.Id, status));

        //    return status;
        //}

        /// <summary>
        /// Update subscription using triples that was received with a notification. 
        /// An updating consists of two steps: update subscribed properties and update classes. 
        /// First step can creates/updates/removes properties for subscribed individuals.
        /// Second step creates individuals: some individual was inserted to the smart space and it
        /// has subscribed class as parent class. It doesn't remove local individual when this individual 
        /// was removed from the smart space.
        /// </summary>
        /// <param name="subscription">Subscription for updating.</param>
        /// <param name="newTriples">Triples from SmartSpace with new values.</param>
        /// <param name="oldTriples">Triples from SmartSpace with old values.</param>
        /// <returns>Changes that were made after updating.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        private IndicationChangesData UpdateSubscription(Subscription subscription)
        {
            SubscriptionEventArgs eventArgs = new SubscriptionEventArgs();

            IndicationChangesData changes = new IndicationChangesData();

            List<ITriple> oldTriples = null;// subscription.SubscribeInfo.GetOldTriples().ToList();
            List<ITriple> newTriples = null;//subscription.SubscribeInfo.GetNewTriples().ToList();

            if (oldTriples.Count == 0 && newTriples.Count == 0)
            {
                subscription.AddChanges(changes);
                return changes;
            }

            foreach (SubscriptionData sd in subscription.GetData())
            {
                List<Property> updatedProperties = new List<Property>();
                //this.UpdateIndividual(sd.Individual, newTriples, oldTriples, changes);

                //eventArgs.Add(sd.Individual, updatedProperties);
            }

            UpdateClasses(subscription, newTriples, oldTriples, changes);

            subscription.AddChanges(changes);

            return changes;
        }

        //// It Needs to perfom working with triples, make a LinkedList and remove elements from it after they are processed.
        //// Also - finding classes.


        /// <summary>
        /// Update an information about individuals that were subscribed by class. 
        /// </summary>
        /// <param name="subscription">Subscription to update.</param>
        /// <param name="newTriples">New triples that were inserted to the smart space and received with the subscription indication.</param>
        /// <param name="oldTriples">Old triples that were removed from the smart space and received with the subscription indication.</param>
        /// <param name="changes">Object to store changes, that individuals were inserted or removed.</param>
        /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
        /// <seealso cref="PetrSU.SmartSlog.IndicationChangesData"/>
        private void UpdateClasses(Subscription subscription, List<ITriple> newTriples, List<ITriple> oldTriples, IndicationChangesData changes)
        {
            if (subscription.GetClasses().Count == 0)
            {
                return;
            }

            List<string> uries = (from oc in subscription.GetClasses()
                                  select oc.Uri).ToList();

            List<ITriple> newClassTriples = null; //(from triple in newTriples
            //                                 where triple.Predicate.Equals(kpi.RdfTypeUri)
            //                                     && uries.Contains(triple.ObjectElement)
            //                                 select triple).ToList<ITriple>();

            List<ITriple> oldClassTriples = null;// (from triple in oldTriples
            //                                       where triple.Predicate.Equals(kpi.RdfTypeUri)
            //                                           && uries.Contains(triple.ObjectElement)
            //                                       select triple).ToList<ITriple>();

            foreach (Triple triple in newClassTriples)
            {
                OntologyClass oc = session.Repository.GetClassByUri(triple.ObjectElement);

                if (oc == null)
                {
                    continue;
                }

                Individual individual = session.Repository.GetOrCreateIndividual(oc, triple.Subject);

                changes.AddChange(individual, ActionType.Insert);
            }

            foreach (Triple triple in oldClassTriples)
            {
                OntologyClass oc = session.Repository.GetClassByUri(triple.ObjectElement);

                if (oc == null)
                {
                    continue;
                }

                Individual individual = session.Repository.GetOrCreateIndividual(oc, triple.Subject);
                changes.AddChange(individual, ActionType.Remove);
            }
        }

        /// <summary>
        /// Update individual using given triples.
        /// Method finds triples for given individual and
        /// perfoms actions(update/remove/insert) for properties.
        /// </summary>
        /// <param name="individual">Individual for update.</param>
        /// <param name="newTriples">Triples from SmartSpace with new values.</param>
        /// <param name="oldTriples">Triples from SmartSpace with old values.</param>
        /// <param name="changes">Object to store changes about properties, that were set/removed/updated.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.IndicationChangesData"/>
        private void UpdateIndividual(Individual individual, List<ITriple> newTriples, List<ITriple> oldTriples, IndicationChangesData changes)
        {
            List<ITriple> newIndividualTriples = (from triple in newTriples
                                                  where triple.Subject.Equals(individual.Uri)
                                                  select triple).ToList<ITriple>();

            newIndividualTriples.Sort(new TriplesPredicateComparer());

            List<ITriple> oldIndividualTriples = (from triple in oldTriples
                                                  where triple.Subject.Equals(individual.Uri)
                                                  select triple).ToList<ITriple>();

            oldIndividualTriples.Sort(new TriplesPredicateComparer());

            ActionType actionType = this.GetActionType(newIndividualTriples, oldIndividualTriples);

            while (actionType != ActionType.Unknown)
            {
                List<ITriple> activeTriples = (actionType == ActionType.Remove) ? oldIndividualTriples : newIndividualTriples;
                Property property = session.Repository.GetPropertyByUri(activeTriples.First().Predicate);

                if (property == null)
                {
                    actionType = GoToNextTriple(newIndividualTriples, oldIndividualTriples);

                    continue;
                }

                object oldValue = null;
                object newValue = null;

                this.PrepareValues(property, newIndividualTriples.FirstOrDefault(), oldIndividualTriples.FirstOrDefault(), out newValue, out oldValue);

                switch (actionType)
                {
                    case ActionType.Update:
                        individual.UpdateProperty(property, oldValue, newValue);
                        changes.AddChange(individual, property, newValue, oldValue, ActionType.Update);
                        break;

                    case ActionType.Remove:
                        individual.RemovePropertyS(property, oldValue);
                        changes.AddChange(individual, property, null, oldValue, ActionType.Remove);
                        break;

                    case ActionType.Insert:
                        individual.InsertPropertyS(property, newValue);
                        changes.AddChange(individual, property, newValue, null, ActionType.Insert);
                        break;
                }

                actionType = this.GoToNextTriple(newIndividualTriples, oldIndividualTriples);
            }
        }



        /// <summary>
        /// Removes firts triple in the given lists and gets action for new triples.
        /// </summary>
        /// <param name="newTriples">List with new triples.</param>
        /// <param name="oldTriples">List with old triples.</param>
        /// <returns>Action that was performed in the smart space (insert, remove, update).</returns>
        private ActionType GoToNextTriple(List<ITriple> newTriples, List<ITriple> oldTriples)
        {
            if ((newTriples.Count > 0 && oldTriples.Count > 0)
                && (newTriples[0].Predicate.Equals(oldTriples[0].Predicate) == true))
            {
                newTriples.RemoveAt(0);
                oldTriples.RemoveAt(0);
            }
            else if (newTriples.Count > 0)
            {
                newTriples.RemoveAt(0);
            }
            else if (oldTriples.Count > 0)
            {
                oldTriples.RemoveAt(0);
            }

            return this.GetActionType(newTriples, oldTriples);
        }

        /// <summary>
        /// Prepare values for action under property. Converts value represented by triples to 
        /// individuals for object-property and to string for data-property.
        /// </summary>
        /// <param name="property">Property for which you need to prepare value.</param>
        /// <param name="newTriple">New triples that represents property value.</param>
        /// <param name="oldTriple">Old triples that represents property value.</param>
        /// <param name="newValue">Prepared new value of property.</param>
        /// <param name="oldValue">Prepared old value of property.</param>
        private void PrepareValues(Property property, ITriple newTriple, ITriple oldTriple, out object newValue, out object oldValue)
        {
            if (property.IsObject() == true)
            {
                this.PrepareObjectValues(newTriple, oldTriple, out newValue, out oldValue);
            }
            else
            {
                PrepareDataValues(newTriple, oldTriple, out newValue, out oldValue);
            }
        }

        /// <summary>
        /// Prepare object values for property. It gets individual from repository or create it.
        /// </summary>
        /// <param name="newTriple">New triples that represents property value.</param>
        /// <param name="oldTriple">Old triples that represents property value.</param>
        /// <param name="newValue">Prepared new value of property.</param>
        /// <param name="oldValue">Prepared old value of property.</param>
        private void PrepareObjectValues(ITriple newTriple, ITriple oldTriple, out object newValue, out object oldValue)
        {
            oldValue = null;
            newValue = null;

            if (newTriple != null)
            {
                //OntologyClass classtype = OwlMapping.GetClassByUuid(kpi, session.Repository, newTriple.ObjectElement);// this.GetClasstypeByUUID(newTriple.ObjectElement);
                //newValue = session.Repository.GetOrCreateIndividual(classtype, newTriple.ObjectElement);
            }

            if (oldTriple != null)
            {
                oldValue = session.Repository.GetIndividual(oldTriple.ObjectElement);
            }
        }

        /// <summary>
        /// Find subscribed data with given data.
        /// </summary>
        /// <param name="individual">Individual for searching.</param>
        /// <param name="property">Property for searching.</param>
        /// <returns>Subscribe data or null if it not found.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionData"/>
        private SubscriptionData FindSubscribedDataBy(Individual individual, Property property)
        {
            SubscriptionData data = this.FindSubscribedDataBy(individual);

            if (data == null)
            {
                return null;
            }

            lock (this.subscriptionSessionsLock)
            {
                //foreach (Property p in data.Properties)
                //{
                //    if (p.Equals(property) == true)
                //    {
                //        return data;
                //    }
                //}

            }

            return null;
        }

        /// <summary>
        /// Find subscribed data with given data.
        /// </summary>
        /// <param name="individual">Individual for searching.</param>
        /// <returns>Subscribe data or null if it not found.</returns>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionData"/>
        private SubscriptionData FindSubscribedDataBy(Individual individual)
        {
            lock (this.subscriptionSessionsLock)
            {
                //SubscriptionData data = (from session in this.asyncSessions
                //                         from d in session.GetData()
                //                         where d.Individual.Equals(individual)
                //                         select d).ToList().FirstOrDefault();

                //if (data != null)
                //{
                //    return data;
                //}

                //return (from session in this.syncSessions
                //        from d in session.GetData()
                //        where d.Individual.Equals(individual)
                //        select d).ToList().FirstOrDefault();
            }
            return null;

        }

        /// <summary>
        /// Gets action that needs to perfom for property.
        /// <remarks>There are three methods for properties: update, remove and insert.
        /// Then you get new triples without old it means insert. 
        /// Then old triples without new it menas remove and then you get old and new triples it means update.
        /// Otherwise it means unknown action.
        /// </remarks>
        /// </summary>
        /// <param name="newTriples">New triples.</param>
        /// <param name="oldTriples">Old triples.</param>
        /// <returns>Action type for given triples. </returns>
        private ActionType GetActionType(List<ITriple> newTriples, List<ITriple> oldTriples)
        {
            if (newTriples.Count == 0 && oldTriples.Count == 0)
            {
                return ActionType.Unknown;
            }

            if (newTriples.Count != 0 && oldTriples.Count != 0
              && (newTriples[0].Predicate.Equals(oldTriples[0].Predicate) == true))
            {
                return ActionType.Update;
            }
            else if (newTriples.Count != 0)
            {
                return ActionType.Insert;
            }
            else if (oldTriples.Count != 0)
            {
                return ActionType.Remove;
            }

            return ActionType.Unknown;
        }

        #endregion // Private methods

        private class TriplesPredicateComparer : Comparer<ITriple>
        {
            public override int Compare(ITriple x, ITriple y)
            {
                return x.Predicate.CompareTo(y.Predicate);
            }
        }
    }
}
