﻿//  ------------------------------------------------------------------------
//  <copyright file="Subscription.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using PetrSU.SmartSlog.NativeStructures;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// Delegate type. This delegat is used to handle subscription's updating. To the delegat pass updated subscription and
    /// <see cref="PetrSU.SmartSlog.SubscriptionEventArgs"/> class with arguments that include updated individuals and their updated changedProperties.  
    /// </summary>
    /// <param name="sender">Sender of the event, subscription that was updated.</param>
    /// <param name="e">Argumets for the delegat (is not used now, always null).</param>
    /// <seealso cref="PetrSU.SmartSlog.Subscription"/>
    /// <seealso cref="PetrSU.SmartSlog.SubscriptionEventArgs"/>
    public delegate void SubscriptionChangedHandler(object sender, EventArgs e);



    /// <summary>
    /// The <c>Subscription</c> class is used to manage a session of the subscription.
    /// <remarks>
    /// To subscribe data you need to set <see cref="PetrSU.SmartSlog.Individual"/> class and list with <see cref="PetrSU.SmartSlog.Property"/> classes.
    /// The subscription can store many pairs of individuals and changedProperties. 
    /// The subscription can be synchronous and asynchronous. 
    /// For updating(receiving subscription information) synchronous subscription 
    /// you need to call function from the <see cref="PetrSU.SmartSlog.Node"/> class. Asynchronous subscription is updating in the background mode.
    /// You can set handler for updating subscription event to handle subscription. The event is reised after subscription's updating by new data from the smart space.
    /// </remarks>
    /// </summary>
    /// <see cref="PetrSU.SmartSlog.Individual"/> 
    public class Subscription
    {
        /// <summary>
        /// Contains last changes that was made under subscribed data.
        /// </summary>
        private IndicationChangesData lastChanges;

        /// <summary>
        /// List with sibscription's datas.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionData"/>
        //private List<SubscriptionData> subscriptionDataList = new List<SubscriptionData>();

        /// <summary>
        /// List with classes that will be used in the subscription.
        /// </summary>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionData"/>
        private List<OntologyClass> subscriptionClasses = new List<OntologyClass>();

        /// <summary>
        /// Attributes are "key - value" pairs, developer uses them to suit your needs.
        /// </summary>
        private Dictionary<string, object> attributes = new Dictionary<string, object>();

        /// <summary>
        /// The type of the subscription and subscription. 
        /// </summary>
        private bool isAsynchronous = false;

        /// <summary>
        /// Active means that a session is subscribed and ready to receive notifications from the smart space.
        /// </summary>
        private bool isActive;

        private IntPtr nativeSubscription;

        internal IntPtr NativeSubscription { get { return nativeSubscription; } }

        /// <summary>
        /// Initializes a new instance of the Subscription class.
        /// It can be synchronous and asynchronous. The synchronous subscription updated by subscription only by using function from the
        /// <see cref="PetrSU.SmartSlog.Node"/> class, and asyncronous subscription updated in the background mode.
        /// </summary>
        /// <param name="node">Node to connect to the smart space.</param>
        /// <param name="isAsynchronous">A flag that determines whether an asynchronous session.</param>
        /// <seealso cref="PetrSU.SmartSlog.Node"/>
        public Subscription(Node node, bool isAsynchronous)
        {
            this.isAsynchronous = isAsynchronous;

            nativeSubscription = NativeMethods.sslog_new_subscription(node.NativeNode, isAsynchronous);

            if (nativeSubscription == null)
            {
                throw ExceptionManager.GetExeption(NativeMethods.sslog_error_get_last_code());
            }

            if(isAsynchronous)
            {
                this._subscriptionHandler = subscriptionHandler;
            }
        }

        
        internal NativeMethods.SubscriptionHandler _subscriptionHandler;

        internal void subscriptionHandler(IntPtr subscription)
        {
            IntPtr sslog_changes = NativeMethods.sslog_sbcr_get_changes_last(subscription);

            IndicationChangesData changes = new IndicationChangesData(sslog_changes);

            this.AddChanges(changes);
        }

        /// <summary>
        /// Update event, it is reised after updating subscription. 
        /// </summary>
        //public event SubscriptionChangedHandler ChangedEvent;
        public event EventHandler<EventArgs> ChangedEvent;

        /// <summary>
        /// Update event, it is reised after updating subscription. 
        /// </summary>
        //public event SubscriptionChangedHandler ChangedEvent;
        public event EventHandler<EventArgs> ErrorEvent;

        /// <summary>
        /// Gets or sets a value indicating whether a subscription is asynchronous. 
        /// You cannot change the type after you subscribe a session.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Thrown when trying to set the type on an active session.</exception>
        public bool IsAsynchronous
        {
            get
            {
                return isAsynchronous;
            }

            set
            {
                if (IsActive == true)
                {
                    throw new InvalidOperationException("You cannot change type while session is active.");
                }

                isAsynchronous = value;
            }
        }

        /// <summary>
        /// Gets identifier of a subscription.
        /// </summary>
        public string Id
        {
            get
            {
                return NativeMethods.sslog_sbcr_get_id(nativeSubscription);
                //return (SubscribeInfo == null) ? String.Empty : null;// SubscribeInfo.Id;
            }
        }

        /// <summary>
        /// Gets a value indicating whether subscription subscription status.
        /// </summary>
        /// <returns>True than a session is active or false otherwise.</returns>
        public bool IsActive
        {
            get { return isActive; }//NativeMethods.sslog_sbcr_is_active(nativeSubscription) == 0 ? false : true; }
            set { isActive = value; }
        }

        /// <summary>
        /// Gets last changes that was made after last indictaion or after session subscription (first synchronization).
        /// </summary>
        public ISubscriptionChangesData LastChanges
        {
            get { return lastChanges; }
        }

        /// <summary>
        /// Gets or sets subscription's information about subscription.  
        /// </summary>
        internal ISubscriptionInfo SubscribeInfo
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets flag to stop synchronous subcription waiting.
        /// </summary>
        internal bool isStopped
        {
            get;
            set;
        }

        /// <summary>
        /// Add an individual and properties for a subscription.
        /// </summary>
        /// <param name="individual">Individual for subscription.</param>
        /// <param name="properties">List with properies for subscription.</param>
        /// <seealso cref="PetrSU.SmartSlog.Individual"/>
        /// <seealso cref="PetrSU.SmartSlog.Property"/>
        public void Add(Individual individual, List<Property> properties)
        {
            if (individual == null)
            {
                throw new System.ArgumentNullException("individual");
            }

            if (properties == null)
            {
                throw new System.ArgumentNullException("properties");
            }

            int result = NativeMethods.sslog_sbcr_add_individual(this.NativeSubscription, individual.NativePtr, MarshalHelp.ToNativeList(properties));

            if (result != Constant.SSLOG_ERROR_NO)
            {
                ExceptionManager.GetExeption(result);
            }
            //SubscriptionData sd = new SubscriptionData(individual, properties);
            //this.subscriptionDataList.Add(sd);
        }

        /// <summary>
        /// Add class to the sibscription.
        /// </summary>
        /// <remarks>
        /// When you subscribe to the class, then you will get all individuals 
        /// that have current class as a parent class. The subscription 
        /// to the class is equivalent to the case when you use 
        /// '* -- rdf:type -- classURI' triple for a query.  
        /// </remarks>
        /// <param name="ontologyClass">Class that will be used for subscription.</param>
        public void Add(OntologyClass ontologyClass)
        {
            if (ontologyClass == null)
            {
                throw new System.ArgumentNullException("ontologyClass");
            }

            if (subscriptionClasses.Contains(ontologyClass) == false)
            {
                subscriptionClasses.Add(ontologyClass);
            }

            int result = NativeMethods.sslog_sbcr_add_class(this.nativeSubscription, ontologyClass.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                ExceptionManager.GetExeption(result);
            }
        }


        public void Add(Triple tripleTemplate)
        {
            if (tripleTemplate == null)
            {
                throw new ArgumentNullException("tripleTemplate");
            }

            int result = NativeMethods.sslog_sbcr_add_triple_template(nativeSubscription, tripleTemplate.NativePtr);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                ExceptionManager.GetExeption(result);
            }
        }

        public void Add(string select_query, string select_triples_templates = null)
        {
            if (select_query == null)
            {
                throw new ArgumentNullException("select_query");
            }

            //if (select_triples_templates == null)
            //{
            //    throw new ArgumentNullException("select_triples_templates");
            //}

            int result = NativeMethods.sslog_sbcr_add_sparql_select(nativeSubscription, select_query, select_triples_templates);

            if (result != Constant.SSLOG_ERROR_NO)
            {
                ExceptionManager.GetExeption(result);
            }
        }


        /// <summary>
        /// Sets an attribute with given name. This function adds new or updates existing attribute.
        /// </summary>
        /// <param name="name">Name of an trribute.</param>
        /// <param name="value">Value of an attribute.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when name of an attribute is null.</exception>
        public void SetAttribute(string name, object value)
        {
            attributes[name] = value;
        }

        /// <summary>
        /// Gets an attribute with given name. 
        /// </summary>
        /// <param name="name">Name of an atrribute.</param>
        /// <returns>An attribute with given name or null if there is no such attribute.</returns>
        /// <exception cref="System.ArgumentNullException">Thrown when name of an attribute is null.</exception>
        public object GetAttribute(string name)
        {
            object value = null;
            attributes.TryGetValue(name, out value);

            return value;
        }

        /// <summary>
        /// Removes an attribute with given name. 
        /// </summary>
        /// <param name="name">Name of an atrribute.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when name of an attribute is null.</exception>
        public void RemoveAttribute(string name)
        {
            attributes.Remove(name);
        }

        //// <summary>
        //// Sets a session state. An active session is subscribed and can receive notification. An inactive session is unsubscribed session.  
        //// </summary>
        //// <param name="isActive">Flag of a session state.</param>
        ////internal void IsActive(bool isActive)
        ////{
        ////    this.isActive = isActive;
        ////}

        /// <summary>
        /// Gets list with <see cref="PetrSU.SmartSlog.SubscriptionData"/> classes, 
        /// it is pairs of individuals and their subscribed changedProperties.
        /// </summary>
        /// <returns>List with data that were added to the subscription.</returns>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionData"/>
        internal List<SubscriptionData> GetData()
        {
            
            return null;// subscriptionDataList;
        }

        /// <summary>
        /// Gets list with <see cref="PetrSU.SmartSlog.OntologyClass"/> classes.
        /// </summary>
        /// <returns>List with classes that were added to the subscription.</returns>
        /// <seealso cref="PetrSU.SmartSlog.OntologyClass"/>
        internal List<OntologyClass> GetClasses()
        {
            return subscriptionClasses;
        }

        /// <summary>
        /// Adds changes that were made using data that was received by indication from the smart space.
        /// This method rises ChangedEvent.
        /// </summary>
        /// <param name="changes">Chnages contains individuals and properties that ware changed.</param>
        internal void AddChanges(IndicationChangesData changes)
        {
            if (changes == lastChanges)
            {
                return;
            }

            if (lastChanges == null)
            {
                lastChanges = changes;
                changes.SequenceNumber = 0;
            }
            else
            {
                changes.SequenceNumber = lastChanges.SequenceNumber + 1;
                lastChanges = changes;
            }


            OnChanged(null);
        }

        /// <summary>
        /// Activates event, passing <see cref="PetrSU.SmartSlog.SubscriptionEventArgs"/> class as arguments.
        /// </summary>
        /// <param name="e">Arguments for event, it contains updated individuals and updated changedProperties.</param>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionEventArgs"/>
        internal void FireChangeEvent(SubscriptionEventArgs e)
        {
            OnChanged(e);
        }

        /// <summary>
        /// Activates event, passing <see cref="PetrSU.SmartSlog.SubscriptionEventArgs"/> class as arguments.
        /// </summary>
        /// <param name="e">Arguments for event, it contains updated individuals and updated changedProperties.</param>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionEventArgs"/>
        internal void FireErrorEvent(SubscriptionEventArgs e)
        {
            OnError(e);
        }

        /// <summary>
        /// Invoke the Changed event; called whenever subscription data changes (adding new changes).
        /// </summary>
        /// <param name="e">Arguments for event, it contains updated individuals and updated changedProperties.</param>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionEventArgs"/>
        protected virtual void OnError(SubscriptionEventArgs e)
        {
            if (ErrorEvent == null)
            {
                return;
            }

            //// Now, raise the event by invoking the delegate. Pass in
            //// the object that initated the event (this) as well as SubscriptionEventArgs.
            //// The call must match the signature of FireEventHandler.
            ErrorEvent(this, e);
        }


        /// <summary>
        /// Invoke the Changed event; called whenever subscription data changes (adding new changes).
        /// </summary>
        /// <param name="e">Arguments for event, it contains updated individuals and updated changedProperties.</param>
        /// <seealso cref="PetrSU.SmartSlog.SubscriptionEventArgs"/>
        protected virtual void OnChanged(SubscriptionEventArgs e)
        {
            if (ChangedEvent == null)
            {
                return;
            }

            //// Now, raise the event by invoking the delegate. Pass in
            //// the object that initated the event (this) as well as SubscriptionEventArgs.
            //// The call must match the signature of FireEventHandler.
            ChangedEvent(this, e);
        }
    }
}
