﻿//  ------------------------------------------------------------------------
//  <copyright file="Triple.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>02.02.2016 9:17</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Runtime.InteropServices;

namespace PetrSU.SmartSlog
{
    /// <summary>
    /// The <c>Triple</c> class contains RDF format triple.
    /// <remarks>The RDF triple consists of three parts: subject, predicate and object. 
    /// For the subject and for the object you need to set a type (<see cref="PetrSU.PetrSU.SmartSlog.KpiWrapper.RdfType"/>).
    /// The subject can have type URI or blank node. Possible types for object are URI, blank node or literal.
    /// </remarks>
    /// </summary>
    /// <seealso cref="PetrSU.SmartSlog.KpiWrapper.RdfType"/>
    /// <seealso cref="PetrSU.SmartSlog.KpiWrapper.ITriple"/>
    public class Triple : ITriple
    {
        /// <summary>
        /// Subject of triple. It can have URI or blank node type.
        /// </summary>
        private string subjectElement;

        /// <summary>
        /// Predicate of triple. It is some URI, what links subject and object.
        /// </summary>
        private string predicate;

        /// <summary>
        /// Object of triple. It can be literal, blank node or URI.
        /// </summary>
        private string objectElement;

        /// <summary>
        /// Type of subject.
        /// </summary>
        /// <seealso cref="PetrSU.KpiLowLibWrapper.RDFType"/>
        private RdfType subjectType;

        /// <summary>
        /// Type of object.
        /// </summary>
        /// <seealso cref="PetrSU.KpiLowLibWrapper.RDFType"/>
        private RdfType objectType;

        /// <summary>
        /// Native structure of SmartSlog triple.
        /// </summary>
        private sslog_triple nativeTriple;

        private IntPtr nativePtr;

        internal IntPtr NativePtr { get { return nativePtr; }  set { nativePtr = value; } }

        /// <summary>
        /// Initializes a new instance of the Triple class.
        /// </summary>
        /// <param name="subject">Subject of triple.</param>
        /// <param name="predicate">Predicate of triple.</param>
        /// <param name="objectElement">Object of triple.</param>
        /// <param name="subjectType">RDF type of subject.</param>
        /// <param name="objectType">RDF type of object.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the some triple's element is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when the subject type is literal.</exception>
        /// <seealso cref="PetrSU.SmartSlog.KpiWrapper.RdfType"/>
        public Triple(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType)
        {
            nativePtr = NativeMethods.sslog_new_triple(subject, predicate, objectElement, (int)subjectType, (int)objectType);

            if (nativePtr == null)
            {
                throw new InvalidOperationException(NativeMethods.sslog_error_get_last_code() + ": " + NativeMethods.sslog_error_get_last_text());
            }

            nativeTriple = (sslog_triple)Marshal.PtrToStructure(nativePtr, typeof(sslog_triple));



            return;

            //if (subject == null)
            //{
            //    throw new ArgumentNullException("subject");
            //}

            //if (predicate == null)
            //{
            //    throw new ArgumentNullException("predicate");
            //}

            //if (objectElement == null)
            //{
            //    throw new ArgumentNullException("objectElement");
            //}

            //if (subjectType == RdfType.Literal)
            //{
            //    throw new ArgumentOutOfRangeException("subjectType", "The type of subject cannot be literal.");
            //}

            //this.subjectElement = subject;
            //this.predicate = predicate;
            //this.objectElement = objectElement;

            //this.subjectType = subjectType;
            //this.objectType = objectType;
        }

        internal Triple(IntPtr nativeTriple)
        {
            nativePtr = nativeTriple;
        }
       

        internal Triple(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType, bool isDetached)
        {
            if (isDetached)
            {
                nativePtr = NativeMethods.sslog_new_triple_detached(subject, predicate, objectElement, (int)subjectType, (int)objectType);
            }
            else
            {
                nativePtr = NativeMethods.sslog_new_triple(subject, predicate, objectElement, (int)subjectType, (int)objectType);
            }

            if (nativePtr == null)
            {
                throw new ArgumentNullException("subject");
            }

            nativeTriple = (sslog_triple)Marshal.PtrToStructure(nativePtr, typeof(sslog_triple));

           
            return;
        }

        ~Triple()
        {
            // Gives error
            //NativeMethods.sslog_free_triple(nativePtr);
        }

        public static Triple Create(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType)
        {
            return new Triple(subject, predicate, objectElement, subjectType, objectType);
        }

        public static Triple CreateDetached(string subject, string predicate, string objectElement, RdfType subjectType, RdfType objectType)
        {
            return new Triple(subject, predicate, objectElement, subjectType, objectType, true);
        }

        internal static Triple Create(IntPtr nativeTriple)
        {
            return new Triple(nativeTriple);
        }




        /// <summary>
        /// Gets or sets the triple's subject.
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when the value is null.</exception>
        public virtual string Subject
        {
            get
            {
                return this.nativeTriple.subject;
            }

            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                this.subjectElement = value;
            }
        }

        public static string AnyUri
        {
            get
            {
                return Constant.SS_RDF_SIB_ANY;
            }
        }

        /// <summary>
        /// Gets or sets the triple's predicate.
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when the value is null.</exception>
        public virtual string Predicate
        {
            get
            {
                return this.predicate;
            }

            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                this.predicate = value;
            }
        }

        /// <summary>
        /// Gets or sets the triple's object.
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when the value is null.</exception>
        public virtual string ObjectElement
        {
            get
            {
                return NativeMethods.sslog_triple_get_object(nativePtr);
            }

            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                this.objectElement = value;
            }
        }

        /// <summary>
        /// Gets or sets the type of the object.
        /// </summary>
        public virtual RdfType ObjectType
        {
            get { return NativeMethods.sslog_triple_get_object_type(nativePtr); }
            set { this.objectType = value; }
        }

        /// <summary>
        /// Gets or sets the type of the subject.
        /// </summary>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when the value is literal.</exception>
        public virtual RdfType SubjectType
        {
            get
            {
                return NativeMethods.sslog_triple_get_subject_type(nativePtr);
            }

            set
            {
                if (value == RdfType.Literal)
                {
                    throw new ArgumentOutOfRangeException("value", "The type of the subject cannot be literal.");
                }

                this.subjectType = value;
            }
        }

        /// <summary>
        /// Set subject and it's type. Type for subject can be URI or blank node.
        /// </summary>
        /// <param name="value">Subject value of the triple.</param>
        /// <param name="type">Type of the subject.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the subject is null.</exception>
        /// <exception cref="System.ArgumentOutOfRangeException">Thrown when the value is literal.</exception>
        public virtual void SetSubject(string value, RdfType type)
        {
            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            if (type == RdfType.Literal)
            {
                throw new ArgumentOutOfRangeException("type", "The type of the subject cannot be literal.");
            }

            this.subjectElement = value;
            this.subjectType = type;
        }

        /// <summary>
        /// Set subject and it's type.
        /// </summary>
        /// <param name="value">Object value of the triple.</param>
        /// <param name="type">Type of the subject.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when the subject is null.</exception>
        public virtual void SetObject(string value, RdfType type)
        {
            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            this.objectElement = value;
        }

        /// <summary>
        /// Returns the System.String that represents the triple with types of 
        /// subject and object.
        /// </summary>
        /// <returns>Current state of the triple.</returns>
        public override string ToString()
        {
            return "Triple: " + this.Subject + " - " + this.Predicate + " - " + this.ObjectElement
                + " (types: " + this.SubjectType + ", " + this.ObjectType + ")";
        }

        /// <summary>
        /// Gets individual from triple.
        ///
        /// If the triple describes an individual (URI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - classURI)
        /// then the individual structure will be returned.
        ///
        /// If the triple describes a property value of some individuals (URI - propertyURI - URI/Data)
        /// then the owner (individual) of this property will be returned. In this way the individual
        /// can be returned if there is a rdf:type triple for the individual with 'URI'.
        /// </summary>
        /// <returns>individual on success or NULL otherwise (triple not describes individual, no rdf:type for individual)</returns>
        public Individual toIndividual()
        {
            IntPtr individualPtr = NativeMethods.sslog_individual_from_triple(this.nativePtr);

            if (individualPtr.ToInt32().Equals(0))
                return null;
            else
                return new Individual(individualPtr);
        }

        /// <summary>
        /// Gets class from triple.
        ///
        /// If the triple describes a class
        /// (ClassURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/2000/01/rdf-schema#Class)
        /// or individual (IndividualURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - classURI)
        /// then the property structure will be returned.
        /// </summary>
        /// <returns>class on success or NULL otherwise (triple not describes class)</returns>
        public OntologyClass toClass()
        {
            IntPtr classPtr = NativeMethods.sslog_property_from_triple(this.nativePtr);

            if (classPtr.ToInt32().Equals(0))
                return null;
            else
            {
                sslog_class e = (sslog_class)Marshal.PtrToStructure(classPtr, typeof(sslog_class));
                OntologyClass ontClass = new OntologyClass(e.uri); 
                return ontClass;
            }
        }

        /// <summary>
        /// Gets property from triple.
        ///
        /// If the triple describes a property
        /// (PropertyURI - http://www.w3.org/1999/02/22-rdf-syntax-ns#type - http://www.w3.org/1999/02/22-rdf-syntax-ns#Property)
        /// or property value (IndividualURI - PropertyURI - Object/Data)
        /// then the property structure will be returned.
        /// </summary>
        /// <returns>property on success or NULL otherwise (triple not describes property)</returns>
        public Property toProperty()
        {
            IntPtr propertyPtr = NativeMethods.sslog_property_from_triple(this.nativePtr);

            if (propertyPtr.ToInt32().Equals(0))
                return null;
            else
            {
                sslog_property_s e = (sslog_property_s)Marshal.PtrToStructure(propertyPtr, typeof(sslog_property_s));
                Property property = new Property(e.uri, NativeMethods.sslog_property_is_object(propertyPtr), null);
                return property;
            }
        }
    }
}
