//  ------------------------------------------------------------------------
//  <copyright file="Timo.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>22.06.2016 11:24</created>
//  <lastedit>05.07.2016 14:38</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Threading;

namespace PetrSU.SmartSlog.Demo.Drinkers
{
    /// <summary>
    /// This class is a part of a Drinkers example. In this example there are 2 KP, one(Timo) drinks beer and publish 
    /// liters, also it subscribes to sms from his Wife (second KP). Wife KP gets notification about liters 
    /// and checks that Timo did not drink much beer.
    /// </summary>
    public class Timo
    {
        /// <summary>
        /// Number of liters that can be drank by Timo.
        /// </summary>
        private const int MaxLiters = 5;

        /// <summary>
        /// Name of node. This name is used while joining to the smart space.
        /// </summary>
        private static string nodeName = "TimoPublisher";

        /// <summary>
        /// Name of the smart space. Now always 'X'.
        /// </summary>
        private static string smartSpaceName = "X";

        /// <summary>
        /// Address of the smart space, SIB.
        /// </summary>
        private static string address = "194.85.173.9"; // PetrSU 
        ////private static string address = "137.204.143.19";   // mml.arces.unibo.it 

        /// <summary>
        /// Port of the smart space, SIB.
        /// </summary>
        private static int port = 10011;

        /// <summary>
        /// Name of Timo.
        /// </summary>
        private static string name = "Timo";

        /// <summary>
        /// Surname of Timo.
        /// </summary>
        private static string surname = "Ericsson";

        /// <summary>
        /// Telephone number of Wife, it is Used as 'from' field for SMS.
        /// </summary>
        private static string wifeTelNumber = "732468";

        /// <summary>
        /// UUID of the sms individual. Individual with this UUID is used as sms. 
        /// </summary>
        private static string smsUUID = "SMSDrinkers";

        private static string defaultDrink = String.Empty;

        /// <summary>
        /// Main method starts KP. It finds Timo in the smart space or create a new individual,
        /// subscribes to SMS from wife and drinks beer.
        /// </summary>
        /// <param name="args">Programm arguments(not used)</param>
        public static void Main(string[] args)
        {
            Node node = new Node(nodeName, smartSpaceName, address, port);
            OntologyStructure.RegisterOntology(node);

            try
            {
                node.Join();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't connect to the smart space.\n" + e.StackTrace);
                Console.ReadLine();
                return;
            }

            Console.WriteLine("Timo joined to the smart space.");

            Collection<Individual> individuals = node.GetIndividuals(OntologyStructure.Man);

            Individual timo = null;

            // Try to find Timo in the list with individuals from the smart space.
            foreach (Individual individual in individuals)
            {
                node.Populate(individual);

                if (individual.GetDataProperty(OntologyStructure.Fname) == name
                        && individual.GetDataProperty(OntologyStructure.Lname) == surname)
                {
                    timo = individual;
                    break;
                }
            }
            
            // If there is no Timo in the smart space, then create a new and insert it to the smart space.
            if (timo == null)
            {
                timo = CreateMan(node, name, surname);
                node.Insert(timo);

                Console.WriteLine("Timo inserted.");
            }
            else
            {
                Console.WriteLine("Timo found.");
            }

            // Set Timo's properties with appropriate values.
            SetDefaultProperties(node, timo);

            // Create a SMS-individual and subscribe to SMS properties asynchronous.
            Individual sms = node.CreateIndividual(OntologyStructure.SMS, smsUUID);
            node.Remove(sms);
            Subscription subscription = CreateSubscription(node, sms);

            try
            {
                node.Subscribe(subscription);
            }
            catch(Exception ex)
            {
                Console.WriteLine("Can't subscribe: " + ex.ToString());
                LeaveSmartSpace(node);
                Console.ReadLine();
                return;
            }

            Console.WriteLine("Press enter to start drinking!!!");
            Console.ReadLine();

            // Start to drink Beer. Update this value in the smart space.
            node.UpdateProperty(timo, OntologyStructure.Drinks, defaultDrink, "Beer");

            int liters = 0;

            // Drink beer, update Timo's liters and check SMS from wife.
            while (true)
            {
                // SMS-ndividual updated automaticlly by asynchronous subscription 
                // and we only checks value of a property.
                if (sms.GetDataProperty(OntologyStructure.From) == wifeTelNumber)
                {
                    Console.WriteLine("\nIncoming SMS from " + wifeTelNumber + "\n");
                    Thread.Sleep(2000);

                    Console.WriteLine("It's time to go to the Hell...");
                    Thread.Sleep(2000);

                    break;
                }

                liters++;
                node.UpdateProperty(timo, OntologyStructure.Numberofdrinks, (liters - 1).ToString(), liters.ToString());

                Console.WriteLine("Timo drinks {0} liters.", timo.GetDataProperty(OntologyStructure.Numberofdrinks));

                // When Timo drunk too much he goes home.
                if (liters > MaxLiters)
                {
                    Console.WriteLine("La-la-la-tra-lala... ");
                    break;
                }

                Thread.Sleep(3000);
            }

            node.RemoveProperty(timo, OntologyStructure.Drinks, "Beer");
            node.Unsubscribe(subscription);

            LeaveSmartSpace(node);
            Console.ReadLine();
        }

        /// <summary>
        /// Creates new individual with class 'Man' and with given name and surname. 
        /// </summary>
        /// <param name="node">Node for creating individual.</param>
        /// <param name="name">Name of a man.</param>
        /// <param name="surname">Surname of a man.</param>
        /// <returns>New created individual.</returns>
        private static Individual CreateMan(Node node, string name, string surname)
        {
            Individual man = node.CreateIndividual(OntologyStructure.Man);

            man.InsertProperty(OntologyStructure.Fname, name);
            man.InsertProperty(OntologyStructure.Lname, surname);

            return man;
        }

        /// <summary>
        /// Sest default values for given man in the smart space. It sets to empty string 
        /// properties: name of drink, liters.
        /// </summary>
        /// <param name="node">Node to access smart space.</param>
        /// <param name="man">Individual to set properties.</param>
        private static void SetDefaultProperties(Node node, Individual man)
        {
            string drinkName = (string)node.GetProperty(man, OntologyStructure.Drinks);
            string liters = (string)node.GetProperty(man, OntologyStructure.Numberofdrinks);

            if (drinkName == null)
            {
                drinkName = defaultDrink;
            }

            if (liters == null)
            {
                liters = "0";
            }

            node.UpdateProperty(man, OntologyStructure.Drinks, drinkName, defaultDrink);
            node.UpdateProperty(man, OntologyStructure.Numberofdrinks, liters, "0");
        }

        /// <summary>
        /// Creates and prepare a subscription to SMS. It adds 'to' and 'from' telephones 
        /// properties to the subscription.
        /// </summary>
        /// <param name="node">Node to access smart space.</param>
        /// <param name="sms">Individual that will be added to a subscription.</param>
        /// <returns>Subscription that contains given individual and properties about SMS telephones. </returns>
        private static Subscription CreateSubscription(Node node, Individual sms)
        {
            Subscription subscription = node.CreateSubscription(isAsynchronous: true);

            List<Property> properties = new List<Property>()
            {
                OntologyStructure.To, 
                OntologyStructure.From
            };

            subscription.Add(sms, properties);

            return subscription;
        }

        /// <summary>
        /// Leaves the smart space.
        /// </summary>
        /// <param name="node">Node to leave.</param>
        private static void LeaveSmartSpace(Node node)
        {
            try
            {
                node.Leave();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't leave the smart space.\n" + e.StackTrace);
            }
        }
    }
}
