﻿//  ------------------------------------------------------------------------
//  <copyright file="Person.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>22.06.2016 11:24</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Linq;
using System.Threading;
using PetrSU.SmartSlog;

namespace PertSU.SmartSlog.Demo.Gps
{
    /// <summary>
    /// A part of a GPS example: one KP publishes inforamtion about a person and a gps-data. 
    /// Another gets persons using subscription to classes.
    /// This class inserts and removes the Person object while you are not pressed CTRL+C or 
    /// or the number of attempts is not exceeded.
    /// </summary>
    public class Person
    {
        /// <summary>
        /// It sets a maximum time for sleeping between inserting or removing next point.
        /// </summary>
        private const int MaxSleepingTime = 15;

        /// <summary>
        /// Name of the smart space, currently it is always 'X'.
        /// </summary>
        private const string SmartSpaceName = "X";

        /// <summary>
        /// Just randowm generator, it is used in a generating process of points.
        /// </summary>
        private static Random rnd = new Random();

        /// <summary>
        /// Name of the KP in the smart space.
        /// </summary>
        private string nodeName = "GPS_PersonKP";

        /// <summary>
        /// IP addres of the SIB in the smart space.
        /// </summary>
        private string address = "194.85.173.9"; // PetrSU 
        //// private string sibAddress = "137.204.143.19"; // mml.arces.unibo.it 

        /// <summary>
        /// Port of the SIB.
        /// </summary>
        private int sibPort = 10010;

        /// <summary>
        /// How many times this KP will be simulate activity:
        /// join -> publish -> remove-> leave.
        /// "-1" - infinitely, KP waits a CTRL+C.
        /// </summary>
        private int maxAttempts = -1;

        /// <summary>
        /// The flag, it stops generating process of points.
        /// </summary>
        private bool isCancelled = false;
        
        /// <summary>
        /// Individual that represents a persion in the smart space.
        /// </summary>
        private Individual individual = null;

        /// <summary>
        /// Node of the smart space.
        /// </summary>
        private Node node = null;

        /// <summary>
        /// Just starts the programm: create new Person and starts activity.
        /// Parameters:
        ///  --name (-n): name of the KP;
        ///  --address (-a): IP address of the SIB;
        ///  --port (-p): port of the SIB. 
        /// </summary>
        /// <param name="args">Parameters for the programm.</param>
        public static void Main(string[] args)
        {
            if (args.Count() % 2 != 0 || args.Count() > 7)
            {
                System.Console.WriteLine("Incorrect number of parameters.");
                System.Console.ReadLine();
                return;
            }

            Person person = InitPerson(args);

            System.Console.WriteLine("Person is created: " + person.individual.Uri);

            person.SimulateActivity();

            Console.WriteLine("\nPlease, give me a headshot by pressing Enter!");
            Console.ReadLine();
        }

        /// <summary>
        /// Creates and initializes person object. 
        /// There are some parameters that can be used:
        ///  --name (-n): name of the KP;
        ///  --address (-a): IP address of the SIB;
        ///  --port (-p): port of the SIB. 
        ///  It creates hadler for CTR+C (@see CancelHandler).
        /// </summary>
        /// <param name="args">Array with parameters.</param>
        /// <returns>New person object.</returns>
        private static Person InitPerson(string[] args)
        {
            Person person = new Person();

            for (int i = args.Count() - 1; i > 0; i -= 2)
            {
                string value = args[i];
                string key = args[i - 1];

                switch (key)
                {
                    case "--name":
                    case "-n":
                        person.nodeName = value;
                        break;
                    case "--address":
                    case "-a":
                        person.address = value;
                        break;
                    case "--port":
                    case "-p":
                        int port = 0;

                        if (int.TryParse(value, out port) == true)
                        {
                            person.sibPort = port;
                        }

                        break;
                }
            }

            // Turn off the default system behavior when CTRL+C is pressed. When 
            // Console.TreatControlCAsInput is false, CTRL+C is treated as an
            // interrupt instead of as input.
            Console.TreatControlCAsInput = false;

            // Establish an event handler to process key press events.
            Console.CancelKeyPress += new ConsoleCancelEventHandler(person.CancelHandler);

            person.node = new Node(person.nodeName, SmartSpaceName, person.address, person.sibPort);

            OntologyStructure.RegisterOntology(person.node.Session);

            // Create individuals and populate them.
            Individual point = person.node.CreateIndividual(OntologyStructure.Point);

            person.individual = person.node.CreateIndividual(OntologyStructure.Person);
            person.individual.InsertProperty(OntologyStructure.Location, point);

            PopulateLocally(person.individual);
            
            return person;
        }

        /// <summary>
        /// Populates a person using random values for properties.
        /// It sets a firts name and a surname properties of a person only once, 
        /// a location information is updated every time when you call this method.
        /// </summary>
        /// <param name="person">Person object that will be populated.</param>
        private static void PopulateLocally(Individual person)
        {
            if (person.GetDataProperty(OntologyStructure.FirstName) == null)
            {
                person.InsertProperty(OntologyStructure.FirstName, "FirstName_" + rnd.Next().ToString());
                person.InsertProperty(OntologyStructure.Surname, "Surname_" + rnd.Next().ToString());
            }

            Individual point = person.GetObjectProperty(OntologyStructure.Location);

            if (point == null)
                return;
            
            string altitude = rnd.Next().ToString();
           
            // For simplicity we remove property and set a new.
            point.RemoveProperties(OntologyStructure.Alt);
            point.InsertProperty(OntologyStructure.Alt, altitude);

            string latitude = rnd.Next().ToString();
            
            point.RemoveProperties(OntologyStructure.Lat);
            point.InsertProperty(OntologyStructure.Lat, latitude);

            string longitude = rnd.Next().ToString();
            
            point.RemoveProperties(OntologyStructure.Long);
            point.InsertProperty(OntologyStructure.Long, longitude);
        }

        /// <summary>
        /// Connects to the smart space, inserts ndividual, then sleeps for some time 
        /// (maximum MaxSleepingTime), then removes individul and leave the smart space.
        /// This function works while maxAttempts greater than zero (0) or while CTRL+C is
        /// not pressed by user.
        /// </summary>
        private void SimulateActivity()
        {
            while (isCancelled == false)
            {
                if (maxAttempts > 0)
                {
                    --maxAttempts;
                }
                else if (maxAttempts == 0)
                {
                    break;
                }

                bool result = JoinAndPublish();

                if (result == true)
                {
                    Thread.Sleep(rnd.Next(MaxSleepingTime) * 1000);
                    CleanAndLeave();
                }

                Thread.Sleep(rnd.Next(MaxSleepingTime) * 1000);
            }
        }

        /// <summary>
        /// Join to the smart space and insert individual.
        /// </summary>
        /// <returns>True on success or false otherwise.</returns>
        private bool JoinAndPublish()
        {
            try
            {
                node.Join();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't connect to the smart space.\n" + e.StackTrace);
                Console.ReadLine();
                return false;
            }

            node.GetIndividuals(OntologyStructure.Point);

            // Refresh point individuals, because they was deleted
            Individual point = node.CreateIndividual(OntologyStructure.Point);
            individual.InsertProperty(OntologyStructure.Location, point);

            // We populate the person's individual again to change gps-coordinates.
            PopulateLocally(individual);
          
            if(point != null)
                node.Insert(point);
            if(individual != null)
                node.Insert(individual);

            Console.WriteLine("Person is added... ({0})", individual.Uri);

            return true;
        }

        /// <summary>
        /// Remove individual and leave the smart space.
        /// </summary>
        /// <returns>True if it successfully leaves the smart space and false otherwise.</returns>
        private bool CleanAndLeave()
        {
            node.Remove(individual.GetObjectProperty(OntologyStructure.Location));
            node.Remove(individual);

            Console.WriteLine("Person is removed... ({0})", individual.Uri);

            try
            {
                node.Leave();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't leave the smart space.\n" + e.StackTrace);
                return false;
            }

            return true;
        }

        /// <summary>
        /// Handler for cancel event(CTRL+C)
        /// It changes isCancelled to true, this flag stops generating process.
        /// When you press CTRL+C, the generating process is interrupted and the 
        /// console cancel event handler, CancelHandler, is invoked. Upon entry 
        /// to the event handler, the Cancel property is false, which means 
        /// the current process will terminate when the event handler terminates. 
        /// However, the event handler sets the Cancel property to true, which 
        /// means the process will not terminate and the KP can properly leave the smart space.
        /// </summary>
        /// <param name="sender">Sender of the event, who has raised this event.</param>
        /// <param name="args">Parameters of the event.</param>
        private void CancelHandler(object sender, ConsoleCancelEventArgs args)
        {
            Console.TreatControlCAsInput = true;

            // Announce that the event handler has been invoked.
            Console.WriteLine("\nHo-ho, funny boy or girl you want to kill ME!!!");
            Console.WriteLine("\nOK, but give me some time to clean up...");

            // Set the Cancel property to true to prevent the process from terminating.
            // Console.WriteLine("Setting the Cancel property to true...");
            args.Cancel = true;

            // Set flag to cancel our simulate.
            isCancelled = true;
        }
    }
}

