﻿//  ------------------------------------------------------------------------
//  <copyright file="PersonRegistrator.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>22.06.2016 11:24</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using PetrSU.SmartSlog;

namespace PertSU.SmartSlog.Demo.Gps 
{
    /// <summary>
    /// A part of a GPS example: one KP publishes inforamtion about a person and a gps-data. 
    /// Another gets persons using subscription to classes.
    /// This class gets persons using subscription to classes and print information about 
    /// their gps-coordinates while you are not pressed CTRL+C.
    /// </summary>
    public class PersonRegistrator
    {
        /// <summary>
        /// Name of the KP in the smart space.
        /// </summary>
        private static string nodeName = "GPS_PersonKP_Registartor";

        /// <summary>
        /// Name of the smart space, currently it is always 'X'.
        /// </summary>
        private static string smartSpaceName = "X";

        /// <summary>
        /// IP addres of the SIB in the smart space.
        /// </summary>
        private static string address = "194.85.173.9"; // PetrSU 
        //// private string sibAddress = "137.204.143.19"; // mml.arces.unibo.it 

        ////private static string address = "192.168.0.222";

        /// <summary>
        /// Port of the SIB.
        /// </summary>
        private static int port = 10010;

        /// <summary>
        /// Container for a subscription. It contains newly inserted individuals and 
        /// individuals that were removed from the smart space.
        /// </summary>
        private static Subscription subscription = null;

        /// <summary>
        /// Node of the smart space.
        /// </summary>
        private static Node node = null;

        /// <summary>
        /// Flag to stops program, it is used with CTRL+C handler.
        /// </summary>
        private static bool isProgramStopped = false;

        /// <summary>
        /// It creates hadler for CTR+C (@see CancelHandler).
        /// Join to the smart space and tracks subscription.
        /// </summary>
        /// <param name="args">Parameters for programm, not used.</param>
        public static void Main(string[] args)
        {
            // Turn off the default system behavior when CTRL+C is pressed. When 
            // Console.TreatControlCAsInput is false, CTRL+C is treated as an
            // interrupt instead of as input.
            Console.TreatControlCAsInput = false;

            // Establish an event handler to process key press events.
            Console.CancelKeyPress += new ConsoleCancelEventHandler(CancelHandler);

            node = new Node(nodeName, smartSpaceName, address, port);
            OntologyStructure.RegisterOntology(node.Session);

            try
            {
                node.Join();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't connect to the smart space.\n" + e.StackTrace);
                Console.ReadLine();
                return;
            }

            Console.WriteLine("Registrator is ready!");

            subscription = node.CreateSubscription(isAsynchronous: false);

            // Add add class to the subscription, we well get indication evry time 
            // when individuals wiil be inserted to the smart space
            // or will be removed from it. 
            subscription.Add(OntologyStructure.Person);

            try
            {
                node.Subscribe(subscription);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Can't subscribe: " + ex.ToString());
                LeaveSmartSpace(node);
                return;
            }

            // Waiting information about individuals and pritn it.
            while (subscription.IsActive && !isProgramStopped)
            {
                node.WaitSubscribe(subscription);

                System.Console.WriteLine("\n\nNew individuals:");

                ISubscriptionChangesData changes = subscription.LastChanges;

                foreach (Individual ind in changes.GetIndividuals(ActionType.Insert))
                {
                    // We get an individual by subscription to a class, and using this individual we 
                    // get an object-property with gps-coordinates. Object property is anothe individual
                    // and we poplate it (receive all properties).
                    Individual point = (Individual)node.GetProperty(ind, OntologyStructure.Location);
                    if (point != null)
                    {
                        node.Populate(point);
                    }

                    PrintPerson(ind); 
                }

                System.Console.WriteLine("Removed individuals:");

                foreach (string uri in changes.GetRemovedIndividualsUri())
                {
                    System.Console.WriteLine(" - Person: " + uri + " (no location information)");
                }
            }

            node.Unsubscribe(subscription);

            LeaveSmartSpace(node);
            Console.ReadLine();
        }
        
        /// <summary>
        /// Only leaves the smart space and catches exception.  
        /// </summary>
        /// <param name="node">Node to leave.</param>
        private static void LeaveSmartSpace(Node node)
        {
            Console.WriteLine("I'm living...");
            try
            {
                node.Leave();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't leave the smart space.\n" + e.StackTrace);
            }
        }

        /// <summary>
        /// Stops current subscripiton.
        /// </summary>
        private static void StopSubscripitonChecking()
        {
            if (subscription == null)
            {
                return;
            }

            if (node != null)
            {
                node.StopWaiting(subscription);
            }
        }

        /// <summary>
        /// Prints onformation about a person and gps-coordinates.
        /// </summary>
        /// <param name="person">Person to print.</param>
        private static void PrintPerson(Individual person)
        {
            Individual point = person.GetObjectProperty(OntologyStructure.Location);

            if (point == null)
            {
                System.Console.WriteLine(" - Person: " + person.Uri + " (no location information)");
            }
            else
            {
                System.Console.WriteLine(
                    " - Person: {0} \n   (Alt: {1}; Lat: {2}; Long: {3})",
                    person.Uri,
                    point.GetDataProperty(OntologyStructure.Alt),
                    point.GetDataProperty(OntologyStructure.Lat),
                    point.GetDataProperty(OntologyStructure.Long));
            }
        }

        /// <summary>
        /// Handler for cancel event(CTRL+C)
        /// It unsubscribe the subscSession.
        /// When you press CTRL+C, the generating process is interrupted and the 
        /// console cancel event handler, CancelHandler, is invoked. Upon entry 
        /// to the event handler, the Cancel property is false, which means 
        /// the current process will terminate when the event handler terminates. 
        /// However, the event handler sets the Cancel property to true, which 
        /// means the process will not terminate and the KP can properly leave the smart space.
        /// </summary>
        /// <param name="sender">Sender of the event, who has raised this event.</param>
        /// <param name="args">Parameters of the event.</param>
        private static void CancelHandler(object sender, ConsoleCancelEventArgs args)
        {
            Console.TreatControlCAsInput = true;

            // Announce that the event handler has been invoked.
            Console.WriteLine("\nHo-ho, funny boy or girl you want to switch me off!!!");

            // Set the Cancel property to true to prevent the process from terminating.
            // Console.WriteLine("Setting the Cancel property to true...");
            args.Cancel = true;

            StopSubscripitonChecking();

            isProgramStopped = true;
        }
    }
}
