﻿//  ------------------------------------------------------------------------
//  <copyright file="Consumer.cs" company="PetrSU">
//  SmartSlog Library is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License as 
//  published by the Free Software Foundation; either version 2 of 
//  the License, or (at your option) any later version.
//  
//  SmartSlog Library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with SmartSlog Library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor,
//  Boston, MA  02110-1301  USA
//  </copyright>
//  
//  <author>SmartSlog Team (Aleksandr A. Lomov)</author>
//  <email>mailto:lomov@cs.karelia.ru</email>
//  <author>Andrey S. Vdovenko</author>
//  <email>mailto:vdovenko@cs.karelia.ru</email>
//  <created>16.03.2016 13:18</created>
//  <lastedit>05.07.2016 14:39</lastedit>
//  ------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using PetrSU.SmartSlog;

namespace PertSU.SmartSlog.Demo.HelloWorldSubscription
{
    /// <summary>
    /// Consumer of 'Hello World' example with subscription.
    /// </summary>
    /// <remarks>
    /// This KP make a subscription to triples using <KP - says - *> triple-pattern.
    /// It gets information about changes and outputs old and new values. 
    /// To stop the KP use CTRL+C. 
    /// </remarks>
    class Consumer : DemoKp
    {
        public Consumer() { }

        public Consumer(string[] args)
        {
            Init(args);
            node = new Node(nodeName, SmartSpaceName, address, port);
        }

        /// <summary>
        /// Just starts the programm: create new Person and starts activity.
        /// Parameters:
        ///  --name (-n): name of the KP;
        ///  --address (-a): IP address of the SIB;
        ///  --port (-p): port of the SIB. 
        /// </summary>
        /// <param name="args">Parameters for the programm.</param>
        static void Main(string[] args)
        {
            Consumer consumer = new Consumer(args);

            consumer.Start();

            Console.ReadLine();
        }

        private void Start()
        {
            try
            {
                node.Join();
            }
            catch (Exception e)
            {
                Console.WriteLine("Can't connect to the smart space.\n" + e.StackTrace);
                Console.ReadLine();
                return;
            }

            PerformQueries();

            node.Leave();
        }

        /// <summary>
        /// Performs different SparQL queries
        /// </summary>
        void PerformQueries()
        {
            while(true)
            {
                Console.WriteLine("Choose query type or press 4 to exit:\n1 - SPARQL CONSTRUCT\n2 - SPARQL ASK\n3 - SPARQL SELECT\n");

                int choose;

                if (!Int32.TryParse(Console.ReadLine(), out choose))
                {
                    Console.WriteLine("Command not recognized. Use predifened set.");
                    continue;
                }

                int[] defined = { 1, 2, 3, 4 };

                if (!defined.Contains(choose))
                {
                    Console.WriteLine("Command not recognized. Use predifened set.");
                    continue;
                }

                if (choose == 4)
                    return;

                Console.WriteLine("Enter query (keep empty to use standard):");

                string query = Console.ReadLine();

                switch (choose)
                {
                    case 1:
                        if (query == "")
                            query = @"PREFIX vCard: <http://www.w3.org/2001/vcard-rdf/3.0#>
                                PREFIX foaf: <http://xmlns.com/foaf/0.1/>
                                CONSTRUCT { 
                                  ?X vCard:FN ?name .
                                  ?X vCard:URL ?url .
                                  ?X vCard:TITLE ?title .
                                }
                                FROM <http://www.w3.org/People/Berners-Lee/card>
                                WHERE { 
                                  OPTIONAL { ?X foaf:name ?name . FILTER isLiteral(?name) . }
                                  OPTIONAL { ?X foaf:homepage ?url . FILTER isURI(?url) . }
                                  OPTIONAL { ?X foaf:title ?title . FILTER isLiteral(?title) . }
                                }";

                        List<ITriple> triples = node.SparqlConstruct(query);
                        if (triples == null)
                        {
                            Console.WriteLine("Error ocured while performing request");
                            break;
                        }
                        PrintTriples(triples);
                        break;
                    case 2:
                        if (query == "")
                            query = @"PREFIX foaf:    <http://xmlns.com/foaf/0.1/>
                                      ASK  { ?x foaf:name  ""Alice"" }";

                        bool result = node.SparqlAsk(query);
                        Console.WriteLine("Result: {0}", (result == true) ? "true" : "false");
                        break;
                    case 3:
                        if (query == "")
                            query = @"SELECT *
                                      WHERE { ?sub ?pred ?obj } LIMIT 100";
                        SparqlResult query_result = node.SparqlSelect(query);
                        PrintSparqlResult(query_result);
                        break;
                }
            }
        }

        void PrintTriples(List<ITriple> triples)
        {
            if (triples.Count == 0) 
            {
                Console.WriteLine("No triples to print.\n");
                return;
            }

            int counter = 1;

            foreach(ITriple triple in triples)
            {
                Console.WriteLine("{0} triple: '{1} - {2} - {3} (o-type: {4})'", ++counter, triple.Subject, triple.Predicate, triple.ObjectElement, triple.ObjectType);
            }
        }

        void PrintSparqlResult(SparqlResult result)
        {
            if (result == null)
            {
                Console.WriteLine("No results");
                return;
            }

            Console.WriteLine("Total result is " + result.rows_count);

            foreach (string name in result.names)
            {
                Console.Write(name + ", ");
            }

            Console.WriteLine("");

            foreach (SparqlResultRow row in result.rows)
            {
                foreach (string value in row.values)
                {
                    Console.Write(value + ", ");
                }

                Console.WriteLine("");
            }

            Console.WriteLine("");
        }
    }
}
